package benchmark.helpClasses;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.Vector;

/**
 * 
 */

public class LogFileWriter extends PrintStream
{
	private File file;

	private OutputStreamWriter stream;

	/**
	 * Only Sys.out
	 * 
	 * @throws IOException
	 */
	public LogFileWriter()
	{
		super(System.out);
	}

	public LogFileWriter(File file, boolean append) throws FileNotFoundException
	{
		super(new FileOutputStream(file, append));
		this.file = file;
	}

	public LogFileWriter(String filePath, boolean append) throws FileNotFoundException
	{
		this(new File(filePath), append);
	}

	public LogFileWriter(String string, boolean append, boolean writeSysOut) throws FileNotFoundException
	{
		this(string, append);
		if ( writeSysOut )
		{
			this.stream = new OutputStreamWriter(System.out);
		}
	}

	@Override
	public void close()
	{
		super.flush();
		super.close();
		if ( this.stream != null )
		{
			try
			{
				this.stream.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
	}

	public void delLastLine() throws IOException
	{
		final Vector<String> lines = new Vector<String>();
		if ( this.file != null )
		{
			try
			{
				final LineNumberReader reader = new LineNumberReader(new FileReader(this.file));
				String str1;
				while ((str1 = reader.readLine()) != null)
				{
					lines.add(str1);
				}
				reader.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
		lines.remove(lines.size() - 1);
		final FileWriter writer = new FileWriter(this.file);
		for (final String str : lines)
		{
			writer.write(str + "\n");
		}
		writer.close();

	}

	@Override
	public void flush()
	{
		super.flush();
		if ( this.stream != null )
		{
			try
			{
				this.stream.flush();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
	}

	public String[] getAlreadyTreatedFiles()
	{
		final Vector<String> treatedInstances = new Vector<String>();
		if ( this.file != null )
		{
			try
			{
				final LineNumberReader reader = new LineNumberReader(new FileReader(this.file));
				String str1;
				while ((str1 = reader.readLine()) != null)
				{
					if ( str1.indexOf("dat") >= 0 )
					{
						treatedInstances.add(str1.substring(0, str1.indexOf("dat") + 3));
					}
				}
				reader.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
		return treatedInstances.toArray(new String[0]);
	}

	public String getLastEntry()
	{
		String out = null;
		if ( this.file != null )
		{
			try
			{
				final LineNumberReader reader = new LineNumberReader(new FileReader(this.file));
				String str1;
				while ((str1 = reader.readLine()) != null)
				{
					out = str1;
				}
				reader.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
		return out;
	}

	@Override
	public void print(boolean b)
	{
		this.write(String.valueOf(b));
	}

	@Override
	public void print(char c)
	{
		this.write(String.valueOf(c));
	}

	@Override
	public void print(char[] s)
	{
		for (final char st : s)
		{
			this.write(st);
		}
	}

	@Override
	public void print(double d)
	{
		this.write(String.valueOf(d));
	}

	@Override
	public void print(float f)
	{
		this.write(String.valueOf(f));
	}

	@Override
	public void print(int i)
	{
		this.write(String.valueOf(i));
	}

	@Override
	public void print(long l)
	{
		this.write(String.valueOf(l));
	}

	@Override
	public void print(Object obj)
	{
		this.write(String.valueOf(obj));
	}

	@Override
	public void print(String s)
	{
		this.write(s);
	}

	@Override
	public void println()
	{
		this.println("");
	}

	@Override
	public void println(boolean x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(char x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(char[] x)
	{
		for (final char xs : x)
		{
			this.write(String.valueOf(xs));
		}
		this.writeln();
	}

	@Override
	public void println(double x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(float x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(int x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(long x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(Object x)
	{
		this.writeln(String.valueOf(x));
	}

	@Override
	public void println(String x)
	{
		this.writeln(String.valueOf(x));
	}

	public void write(double str)
	{
		this.write(Double.toString(str));
	}

	public void write(String str)
	{
		super.print(str);
		if ( this.stream != null )
		{
			try
			{
				this.stream.write(str);
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}
	}

	public void writeln()
	{
		this.writeln("");

	}

	public void writeln(String string)
	{
		this.write(string + "\n");

	}
}