package benchmarks;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Random;
import java.util.Set;

import benchmarks.TestforKolja.Heuristic;
import benchmarks.TestforKolja.NodeDegree;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.IntTupel;

class FKDA
{
	private NodeDegree[] v;

	private int n, trials;

	private AdjacentGraph graph;

	private Set<IntTupel> result;

	public FKDA()
	{
	}

	private void add_edge(int i, int j)
	{
		this.result.add(new IntTupel(this.v[i].vertex, this.v[j].vertex));
		this.v[i].degree++;
		this.v[j].degree++;
	}

	private int backward_edge(int i, int j)
	{
		int l = this.n;
		for (; l > j + 1; l--)
		{
			if ( !this.edge_exists(this.v[j].vertex, this.v[l].vertex) && (this.v[l].degree < this.v[i].degree) )
				return l;
		}
		return -1;
	}

	private int edge_creation(int i, int na, Heuristic heur)
	{
		int j, l;

		for (j = i + 1; j <= i + na - 1; j++)
		{
			while (this.v[j].degree < this.v[i].degree)
			{
				switch(heur) {
					case BACKWARD: {
						l = this.backward_edge(i, j); break;
					}
					case RANDOM: {
						l = this.random_edge(i, j); break;
					}
					default: {
						l = this.forward_edge(i, j); break;
					}
				}

				if ( l == -1 )
					return j;
				this.add_edge(j, l);
			}
		}
		Arrays.sort(this.v);
		return j;
	}

	private boolean edge_exists(int node1, int node2)
	{ // takes vertices, not indices!
		return this.graph.isAdjacent(node1, node2) || this.result.contains(new IntTupel(node1, node2));
	}

	public Set<IntTupel> fkda(AdjacentGraph graph, int k, Heuristic heur)
	{
		this.init(graph);
		int i = 0;
		int na, l, j;
		while (i < this.n)
		{
			na = this.greedy_examination(i, k);
			j = this.edge_creation(i, na, heur);
			if ( j == i + na )
			{
				i = i + na;
			}
			else
			{
				l = this.relaxed_edge_creation(i, j);
				if ( l < i )
				{
					i = 0;
				}
			}
		}

//		System.out.println("is k-anon? " + this.isKAnon(k));
//		if ( !this.isKAnon(k) )
//		{
//			for (int u = 0; u <= this.n; u++)
//			{
//				System.out.println(this.v[u]);
//			}
//		}
		return this.result;
	}

	private int forward_edge(int i, int j)
	{
		for (int l = j + 1; l < this.n; l++)
		{
			if ( !this.edge_exists(this.v[j].vertex, this.v[l].vertex) && this.v[l].degree < this.v[i].degree )
				return l;
		}
		return -1;
	}

	private int greedy_examination(int i, int k)
	{
		int j = i + 1;
		for (; j < this.n; j++)
		{
			if ( this.v[j].degree < this.v[i].degree )
			{
				if ( i > 0 && this.v[i].degree == this.v[i - 1].degree )
				{
					if ( !(this.n - j + 1 < k) )
						return j - i;
				}
				else
				{
					if ( !(this.n - i + 1 < 2 * k || this.n - j + 1 < k) )
						return Math.max(k, j - i);
				}
				break;
			}
		}
		return this.n - i + 1;
	}

	private void init(AdjacentGraph graph)
	{
		this.graph = graph;
		this.set_v();
		this.n = graph.getAmountOfVertices() - 1;
		this.result = new HashSet<IntTupel>();
		this.trials = n;
	}
	
	public void set_trials(int t) {
		if(t > 0) {
			this.trials = t;
		}
	}

	private boolean isKAnon(int k)
	{
		Arrays.sort(this.v);
		int curr;
		int streak = 1;
		curr = this.v[0].degree;
		for (int i = 1; i <= this.n; i++)
		{
			if ( curr == this.v[i].degree )
			{
				streak++;
			}
			else
			{
				if ( streak < k )
					return false;
				streak = 1;
				curr = this.v[i].degree;
			}
		}
		if ( streak < k )
			return false;

		return true;
	}

	private int random_edge(int i, int j)
	{
		final Random rand = new Random();
		int l;
		int k = 0;
		do
		{
			l = rand.nextInt(this.n+1); // TODO: wie oft?
			k++;
		} while (k < trials && (l == j || this.edge_exists(this.v[j].vertex, this.v[l].vertex) || this.v[l].degree >= this.v[i].degree));
		if (l != j && !this.edge_exists(this.v[j].vertex, this.v[l].vertex) && (this.v[l].degree < this.v[i].degree) )
			return l;
		return -1;
	}

	private int relaxed_edge_creation(int i, int j)
	{
		int l;
		int dvl, dvi;
		for (l = this.n; l > 0; l--)
		{
			if ( j == l )
			{
				continue; // avoid self loops
			}
			if ( !this.edge_exists(this.v[j].vertex, this.v[l].vertex) )
			{
				this.add_edge(j, l);
				dvl = this.v[l].degree;
				dvi = this.v[i].degree;
				if ( this.v[j].degree == this.v[i].degree )
				{
					Arrays.sort(this.v);
					if ( dvl > dvi )
						return -1; // TODO: korrekt? Der Check in fkda sollte nicht reichen,
					// weil l der alte Index ist
					else
						return l;
				}
			}
		}
		return l;
	}

	private void set_v()
	{
		this.v = new NodeDegree[this.graph.getAmountOfVertices()];
		for (int i = 1; i <= this.graph.getAmountOfVertices(); i++)
		{
			this.v[i - 1] = new NodeDegree(i, this.graph.getDegree(i));
		}
		Arrays.sort(this.v);
	}
}