package benchmarks;

import java.io.File;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

import seppImpl.CompactDegreeSequence;
import seppImpl.DegreeAnonymizer2;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.data.IntTupel;
import de.vibee.kanon.tools.GreedyGraphConstructor;
import de.vibee.kanon.tools.Reader;

public class MainTest
{
	public static int logLevel = 3;

	public static long timeStamp = 0;

	static boolean a = false;

	static boolean b = false;

	static boolean c = false;

	public static int[] getRandomDegreeSequence(final int n, final int k)
	{
		final int i[] = new int[n + 1];
		for (int t = 0; t < i.length; t++)
		{
			if ( Math.random() > 0.5f )
			{
				i[t] = (int) Math.round(Math.random() * Math.pow(1.1f, i.length - t));
			}
			else
			{
				i[t] = (int) Math.round(Math.random() * k);
			}
		}
		return i;
	}

	public static int[] getRandomkAnonDegreeSequence(final int n, final int k)
	{
		final int i[] = new int[n + 1];
		final boolean b[] = new boolean[n + 1];
		for (int t = 0; t < i.length; t++)
		{
			b[t] = Math.random() >= 0.5f ? true : false;
		}

		final int c = (int) Math.round(Math.random() * 2 * k * n);
		while (c > 0)
		{
			int l;
			int j;
			do
			{
				l = (int) Math.round(Math.random() * (i.length - 1));
				j = (int) Math.round(Math.random() * (i.length - 1));
			} while (b[l] == false || b[j] == false);
			i[l]++;
			i[j]++;
		}
		return i;
	}

	/**
	 * @param args
	 * @throws ExecutionException
	 * @throws InterruptedException
	 */
	public static void main(final String[] args) throws InterruptedException, ExecutionException
	{
		MainTest.timeStamp = System.currentTimeMillis();

		final Options options = new Options();
		final Option help = new Option("help", "Print this help");
		OptionBuilder.withArgName("k");
		OptionBuilder.hasArg();
		OptionBuilder.withDescription("Specify k (anonymization level)");
		final Option kOption = OptionBuilder.create("k");
		final Option enabledataReductionRules = new Option("drr", "data-reduction-rules", false, "Enable data reduction rules");
		final Option disabledataReductionRules = new Option("nodrr", "disable-data-reduction-rules", false, "Disable data reduction rules (Default)");
		final Option verbose = new Option("verbose", "Be louder");
		OptionBuilder.withArgName("selection-mode");
		OptionBuilder.withLongOpt("selection-mode");
		OptionBuilder.hasArg();
		OptionBuilder
				.withDescription("specify vertex selection heuristik (0 = Minimum common neighbors, 1 = minimum sum of neighbor degrees, 2 = random (200 iteration))");
		final Option selectionMode = OptionBuilder.create("sm");
		OptionBuilder.withArgName("order-mode");
		OptionBuilder.withLongOpt("order-mode");
		OptionBuilder.hasArg();
		OptionBuilder
				.withDescription("specify vertex ordering heuristik for graph construction  (0 = vertices with largest difference first, 1 = vertices with fewest common neighbors in set first, 3 = random (500 iterations))");
		// new Option("sel", "selection-mode", true,
		// "Specifies heuristic for selecting vertices based on block sequence");
		final Option orderMode = OptionBuilder.create("om");
		OptionBuilder.withArgName("if");
		OptionBuilder.withLongOpt("input-file");
		OptionBuilder.hasArg();
		OptionBuilder.withDescription("specify input file");
		final Option inputFile = OptionBuilder.create("if");

		options.addOption(help);
		options.addOption(kOption);
		options.addOption(inputFile);
		options.addOption(enabledataReductionRules);
		options.addOption(disabledataReductionRules);
		options.addOption(orderMode);
		options.addOption(selectionMode);
		options.addOption(verbose);

		final CommandLineParser parser = new GnuParser();
		CommandLine cmd = null;

		try
		{
			cmd = parser.parse(options, args);
		} catch (final ParseException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		if ( cmd.hasOption("help") )
		{
			final HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("kanonymizer", options);
			System.exit(0);
		}

		final boolean datareduction;

		Reader reader = null;
		if ( cmd.hasOption("input-file") )
		{
			System.out.println("#Input file: " + cmd.getOptionValue("input-file"));
			if ( !new File(cmd.getOptionValue("input-file")).exists() )
			{
				System.out.println("Input file does not exist. Terminating.");
				System.exit(1);
			}
			reader = new Reader(cmd.getOptionValue("input-file"));
		}
		else
		{
			final HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("kanonymizer", options);
			System.out.println("Input file not specified. Terminating.");
			System.exit(1);
		}

		int k = 0;
		if ( cmd.hasOption("k") )
		{
			try
			{
				k = Integer.parseInt(cmd.getOptionValue("k"));
				System.out.println("#k: " + k);
			} catch (final NumberFormatException ie)
			{
				System.out.println("k must be specified as a positive Integer. Terminating.");
				System.exit(1);
			}
		}
		else
		{
			final HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("kanonymizer", options);
			System.out.println("k not specified. Terminating.");
			System.exit(1);
		}
		if ( cmd.hasOption("drr") )
		{
			datareduction = true;
			System.out.println("#Data reduction rule enabled");
		}
		else
		{
			datareduction = false;
		}

		int sm = -1;
		if ( cmd.hasOption("sm") )
		{
			try
			{
				sm = Integer.parseInt(cmd.getOptionValue("sm"));
				if ( sm < 0 || sm > 3 )
					throw new NumberFormatException();
			} catch (final NumberFormatException ie)
			{
				System.out.println("Selection mode must be specified as an integer [0-2]. Terminating.");
				System.exit(1);
			}
		}

		int om = -1;

		if ( cmd.hasOption("om") )
		{
			try
			{
				om = Integer.parseInt(cmd.getOptionValue("om"));
				if ( om < 0 || om > 2 )
					throw new NumberFormatException();
			} catch (final NumberFormatException ie)
			{
				System.out.println("Ordering mode must be specified as an integer [0-2]. Terminating.");
				System.exit(1);
			}
		}

		if ( cmd.hasOption("verbose") )
		{
			MainTest.logLevel = 1;
			// System.out.println("#Data reduction rule enabled");
		}

		long time = System.currentTimeMillis();

		final AdjacentGraph graph = reader.readInputGraphAlternateFormat();
		final CompactDegreeSequence initialDegreeSequence = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));

		final long graphBuildTime = System.currentTimeMillis() - time;

		System.out.println("#Building graph and initial sequence done in " + graphBuildTime);

		System.out.println("#Initial block degree sequence: " + initialDegreeSequence);

		System.out.println("#Upperbound:" + initialDegreeSequence.getUpperBoundForKAnon(k));

		time = System.currentTimeMillis();

		final Set<CompactDegreeSequence> solutions = new DegreeAnonymizer2(graph, initialDegreeSequence, k, true, datareduction)
				.getAllkAnonRealSequences();

		// if (executor.isTerminated()){
		// Set<CompactDegreeSequence> i = solutions.get();
		final CompactDegreeSequence difference = solutions.iterator().next().difference(initialDegreeSequence);
		// System.out.println("Vertex raise list: " + difference);
		System.out.println("Anonymizing runtime " + (System.currentTimeMillis() - time) / 1000d + "sec. Anonymization cost is "
				+ difference.getTotalDegreeSum());
		System.out.println("Computed " + solutions.size() + " " + k + "-anonymous degree sequences");
		Set<IntTupel> edgeInsertionSet = null;

		if ( sm != -1 && om != -1 )
		{
			for (final CompactDegreeSequence solution : solutions)
			{
				edgeInsertionSet = new GreedyGraphConstructor(graph).constructGraph(initialDegreeSequence, solution, true, om, sm);
				System.out.println("Selecting method " + sm + ", Building method " + om + " " + (edgeInsertionSet == null ? "not " : "")
						+ "successful");
				System.out.println("Graph is " + graph.getAnonymity() + "-anonym");
				System.out.println("adding edges...");
				final AdjacentGraph graph2 = new Reader(cmd.getOptionValue("input-file")).readInputGraphAlternateFormat();
				for (final IntTupel edge : edgeInsertionSet)
				{
					System.out.println(graph2.addEdge(edge.getFirst(), edge.getSecond()));
				}
				System.out.println("Graph is " + graph2.getAnonymity() + "-anonym");
				// System.exit(0);
			}
		}

		else if ( sm != -1 && om == -1 )
		{
			for (final CompactDegreeSequence solution : solutions)
			{
				for (int buildingOrderMethod = 0; buildingOrderMethod <= 2; buildingOrderMethod++)
				{

					edgeInsertionSet = new GreedyGraphConstructor(graph).constructGraph(initialDegreeSequence, solution, true, buildingOrderMethod,
							sm);
					if ( buildingOrderMethod == 0 )
					{
						System.out.println("Building method 0: Degree difference:               " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}
					if ( buildingOrderMethod == 1 )
					{
						System.out.println("Building method 1: Fewest Common Neighbors in set:  " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}
					if ( buildingOrderMethod == 2 )
					{
						System.out.println("Building method 2: Random order:                    " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}

				}
			}
		}

		else if ( sm == -1 && om != -1 )
		{
			for (final CompactDegreeSequence solution : solutions)
			{
				for (int selectionMethod = 0; selectionMethod <= 3; selectionMethod++)
				{
					edgeInsertionSet = new GreedyGraphConstructor(graph).constructGraph(initialDegreeSequence, solution, true, om, selectionMethod);

					if ( selectionMethod == 0 )
					{
						System.out.println("Selection method 0: Fewest common neighbors:            " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}
					if ( selectionMethod == 1 )
					{
						System.out.println("Selection method 1: Minimum sum of neighbor degrees:     " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}
					if ( selectionMethod == 2 )
					{
						System.out.println("Selection method 2: Random selection:                    " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}
					if ( selectionMethod == 2 )
					{
						System.out.println("Selection method 3: Minimum common neighbors (improved): " + (edgeInsertionSet == null ? "Not " : "")
								+ "successful");
					}

				}
			}
		}

		else
		{
			for (int selectionMethod = 0; selectionMethod <= 3; selectionMethod++)
			{

				if ( selectionMethod == 0 )
				{
					System.out.println("--- \nSelection Method 0: Fewest Common Neighbors \n ---");
				}
				else if ( selectionMethod == 1 )
				{
					System.out.println("--- \nSelection Method 1: Lowest Neighbour Degree \n ---");
				}
				else if ( selectionMethod == 2 )
				{
					System.out.println("--- \nSelection Method 2: Random \n ---");
				}
				else if ( selectionMethod == 3 )
				{
					System.out.println("--- \nSelection Method 3: Fewest Common Neighbor (improved) \n ---");
				}

				for (int buildingOrderMethod = 0; buildingOrderMethod <= 2; buildingOrderMethod++)
				{
					long successfulCounter = 0;
					for (final CompactDegreeSequence solution : solutions)
					{
						edgeInsertionSet = new GreedyGraphConstructor(graph).constructGraph(initialDegreeSequence, solution, true,
								buildingOrderMethod, selectionMethod);
						if ( edgeInsertionSet != null && edgeInsertionSet.size() > 0 )
						{
							successfulCounter++;
						}
						// break;

					}
					if ( buildingOrderMethod == 0 )
					{
						System.out.println("Building method 0: Degree difference: " + successfulCounter + " from " + solutions.size()
								+ " successfully constructed.");
					}
					if ( buildingOrderMethod == 1 )
					{
						System.out.println("Building method 1: Common Neighbors:  " + successfulCounter + " from " + solutions.size()
								+ " successfully constructed.");
					}
					if ( buildingOrderMethod == 2 )
					{
						System.out.println("Building method 2: Random order:      " + successfulCounter + " from " + solutions.size()
								+ " successfully constructed.");
					}

				}

			}
		}

	}

	public static int sum(final int[] remainingCost)
	{
		int s = 0;
		for (int i = 0; i < remainingCost.length; i++)
		{
			s += remainingCost[i];
		}
		return s;
	}

}
