package benchmarks;

import java.io.File;
import java.util.Arrays;

import benchmarks.TestforKolja.Heuristic;

import seppImpl.CompactDegreeSequence;
import seppImpl.DegreeAnonymizer2;
import seppImpl.GraphRealizer;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.tools.Reader;

public class SeppsMain
{
	public final static int logLevel = 1;

	public static int[] getRandomDegreeSequence(final int n, final int k)
	{
		final int i[] = new int[n + 1];
		for (int t = 0; t < i.length; t++)
		{
			if ( Math.random() > 0.5f )
			{
				i[t] = (int) Math.round(Math.random() * Math.pow(1.1f, i.length - t));
			}
			else
			{
				i[t] = (int) Math.round(Math.random() * k);
			}
		}
		return i;
	}

	public static int[] getRandomkAnonDegreeSequence(final int n, final int k)
	{
		final int i[] = new int[n + 1];
		final boolean b[] = new boolean[n + 1];
		for (int t = 0; t < i.length; t++)
		{
			b[t] = Math.random() >= 0.5f ? true : false;
		}

		final int c = (int) Math.round(Math.random() * 2 * k * n);
		while (c > 0)
		{
			int l;
			int j;
			do
			{
				l = (int) Math.round(Math.random() * (i.length - 1));
				j = (int) Math.round(Math.random() * (i.length - 1));
			} while (b[l] == false || b[j] == false);
			i[l]++;
			i[j]++;
		}
		return i;
	}

	/**
	 * @param args
	 */
	@SuppressWarnings("deprecation")
	public static void main(final String[] args)
	{
		System.out.println(Arrays.toString(args));
		MainTest.logLevel = 0;
		DegreeAnonymizer2.logDR = 1;
		GraphRealizer.log = 0;

		final boolean[] b = { true };
		for (final boolean applyDataReduction : b)
		{
			System.out.println("\napplyRules=" + applyDataReduction);
			final int[] kValues = { 2, 3, 5, 10, 15, 20, 30, 50, 100, 150, 200, 4, 7, 9, 25, 40, 75 };
			// for (final int k : kValues)
			final int parseInt = Integer.parseInt(args[0]);
			final int parseInt2 = Integer.parseInt(args[1]);
			for (int kVal = parseInt; kVal < parseInt2; kVal++)
			{
				int k = kValues[kVal];
				System.out.println("k = " + k + "\t");
				final String[] paths = { "graphs/graphConference/", "graphs/10thDimacs/", "graphs/stanford/" };
//				 final String[] paths = { "syntheticKAnon/" };
				for (final String path : paths)
				{
					final String[] list = new File(path).list();
					Arrays.sort(list);

					for (final String filePath : list)
					{
						// if (filePath.contains("Cite"))
						{
							System.out.print(filePath + "\t");
							final Reader reader = new Reader(path + filePath);
							final AdjacentGraph graph = reader.readInputGraphAlternateFormat();
							// final CompactDegreeSequence initialDegreeSequence = new
							// CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));

							MainTest.timeStamp = System.currentTimeMillis();

							int tmpLog = MainTest.logLevel;
							MainTest.logLevel = 0;
							int tmpLogDR = DegreeAnonymizer2.logDR;
							DegreeAnonymizer2.logDR = 0;

							TestHeuristic heuristicTester = new TestHeuristic(graph, k);
							heuristicTester.start();

							try
							{
								heuristicTester.join(1000 * 60 * 60);
							} catch (InterruptedException e1)
							{
							}
							if ( heuristicTester.isAlive() )
							{
								System.out.print("\t\tUBheuristic aborted");
								heuristicTester.interrupt();
								heuristicTester.stop();
							}
							System.gc();

							MainTest.logLevel = tmpLog;
							DegreeAnonymizer2.logDR = tmpLogDR;

							final TestAnonymizerCaller s = new TestAnonymizerCaller(graph, k, applyDataReduction);
							s.start();
							try
							{
								s.join(1000 * 60 * 60);

							} catch (final InterruptedException e)
							{

							}
							System.out.print(s.output);
							if ( s.isAlive() )
							{
								for (int tmp = s.output.split("\t").length; tmp <= 8; tmp++)
									System.out.print("\t");
								s.interrupt();
								s.stop();

								System.out.print("aborted,\t#testRealSol:" + s.testedToRealizeSolutions);
								if ( s.anonymizer != null )
								{
									System.out.print("\t lastCostValue:" + (s.anonymizer.lastExaminedCost + s.anonymizer.preprocessCost));
									System.out.print("\t, lastUpperBoundInSolutionIterator:"
											+ s.anonymizer.solutionDegreeIterator.getCurrentUpperbound());
								}
							}
							System.out.println();
							System.out.flush();
							System.gc();
						}
					}
				}
			}
		}
	}

	public static int getFKDAHeuristicResult(final int k, final AdjacentGraph graph)
	{
		FKDA fkda = new FKDA();
		final int size = Math.min(Math.min(fkda.fkda(graph, k, Heuristic.RANDOM).size(), fkda.fkda(graph, k, Heuristic.BACKWARD).size()),
				fkda.fkda(graph, k, Heuristic.FORWARD).size());
		return size * 2;
	}

	public static int sum(final int[] remainingCost)
	{
		int s = 0;
		for (int i = 0; i < remainingCost.length; i++)
		{
			s += remainingCost[i];
		}
		return s;
	}

}
