package benchmarks;

import java.util.Set;
import java.util.concurrent.Callable;

import seppImpl.CannotRealize;
import seppImpl.CompactDegreeSequence;
import seppImpl.DegreeAnonymizer2;
import seppImpl.GraphRealizer;

import dataStructures.Store;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.data.IntTupel;

public class TestAnonymizerCaller extends Thread
{
	private AdjacentGraph g;

	public int testedToRealizeSolutions = 0;

	@Override
	public void run()
	{
		applyTest(applyDataReduction, this.g, this.k);
	}

	private int k;

	private boolean applyDataReduction;

	public TestAnonymizerCaller(AdjacentGraph graph, int k, boolean applyDR)
	{
		this.g = graph;
		this.k = k;
		applyDataReduction = applyDR;

	}

	public DegreeAnonymizer2 anonymizer;

	public String output = "";

	boolean applyTest(boolean applyDataReduction, final AdjacentGraph graph, final int k)
	{

		long time = System.currentTimeMillis();
		CompactDegreeSequence initialDegreeSequence = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));
		anonymizer = new DegreeAnonymizer2(graph, initialDegreeSequence, k, true, applyDataReduction);
		final Set<CompactDegreeSequence> solutions = anonymizer.getAllkAnonRealSequences();

		final double neededTime = (System.currentTimeMillis() - time) / 1000d;

		final int numberOfSolutions = solutions.size();

		final int degreeAnon2 = solutions.iterator().next().difference(initialDegreeSequence).getTotalDegreeSum();

		output += ",\t time:" + neededTime + "s,\t OPT=" + degreeAnon2 + ",\t noSolutions:" + numberOfSolutions;

		output += "\t, (hopeforSuccussRate: "
				+ 100f
				* ((float) anonymizer.getLastSolutionDegreeIterator().getNoOfAEGHopeDestroyed() / anonymizer.getLastSolutionDegreeIterator()
						.getNoOfAEGHopeTests()) + "%,\t total: " + anonymizer.getLastSolutionDegreeIterator().getNoOfAEGHopeTests() + ")";

		for (final CompactDegreeSequence sol : solutions)
		{
			try
			{
				final GraphRealizer graphRealizer = new GraphRealizer(graph, sol);
				final Set<IntTupel> s = graphRealizer.constructGraph();
				this.testedToRealizeSolutions++;
				output += "\t, real";
				output += ",\t#testSol:" + testedToRealizeSolutions;
				return true;

			} catch (final CannotRealize e)
			{
				this.testedToRealizeSolutions++;
			}
		}
		output += "\t, Nreal";
		output += ",\t#testSol:" + testedToRealizeSolutions;
		return false;
	}

}
