package de.vibee.kanon;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Set;

import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.data.IntTupel;
import de.vibee.kanon.data.Node;
import de.vibee.kanon.tools.DegreeAnonymizer;
import de.vibee.kanon.tools.GreedyGraphConstructor;
import de.vibee.kanon.tools.Reader;

public class Main
{

	public static void main(final String[] args)
	{
		// TODO Auto-generated method stub

		// TODO Auto-generated method stub
		System.out.println("#Graph k-anonymization tool by Clemens Hoffmann");
		if ( args.length != 3 )
		{
			System.out.println("#Invalid arguments. usage: kanonymize inputfile k outputpath");
			return;
		}

		final File inputFile = new File(args[0]);
		if ( !inputFile.exists() )
		{
			System.out.println("#File " + args[0] + " doesn't exist.");
			return;
		}
		else
		{
			System.out.println("#Input file: " + inputFile.getName());
		}

		int k;

		try
		{
			k = Integer.valueOf(args[1]);
		} catch (final NumberFormatException e)
		{
			System.out.println("#Invalid arguments. usage: kanonymize inputfile k outputpath");
			return;
		}

		if ( k <= 0 )
		{
			System.out.println("#Argument k must be valid integers.");
			return;
		}
		else
		{
			System.out.println("#k = " + k);
		}

		final File outputPath = new File(args[2]);
		if ( !outputPath.exists() || !outputPath.isDirectory() )
		{
			if ( !outputPath.mkdirs() )
			{
				System.out.println("#Can't create output directory. Aborting");
				System.exit(0);
			}
		}
		System.out.println("#Saving result to " + outputPath.getAbsolutePath() + "/" + inputFile.getName() + ".log");
		final long time = System.currentTimeMillis();
		final Reader reader = new Reader(args[0]);
		final AdjacentGraph graph = reader.readInputGraphAlternateFormat();
		final Node initialDegreeSequence[] = DegreeSequence.calcDegreeSequence(graph);
		final long graphBuildTime = System.currentTimeMillis() - time;
		System.out.println("#Building graph and initial sequence done in " + graphBuildTime);
		final int graphSize = graph.getAmountOfVertices();
		final int edgeSetSize = graph.getAmountOfEdges();
		System.out.println("#Size of graph = " + graphSize);
		System.out.println("#Amount of edges = " + edgeSetSize);
		final int delta = DegreeSequence.getDelta(initialDegreeSequence);
		System.out.println("#Delta = " + delta);
		final int anonymity = DegreeSequence.getAnonymity(initialDegreeSequence);
		System.out.println("#Sequence is " + anonymity + "-anonym");

		final Node solutionSequence[] = new DegreeAnonymizer(graph, initialDegreeSequence, delta).startSequenceCalculation(k, false);
		final long anonymizationTime = System.currentTimeMillis() - time - graphBuildTime;

		System.out.println("#DP-Algorithmus finished in " + anonymizationTime + " seconds");

		Set<IntTupel> edgeSet = null;
		if ( solutionSequence == null )
		{
			System.out.println("Could not find any realizable solution.");
		}
		else
		{
			edgeSet = new GreedyGraphConstructor(graph).constructGraph(initialDegreeSequence, solutionSequence, false);
			final Node finalSequence[] = DegreeSequence.calcDegreeSequence(graph);
			System.out.println("Graph is " + DegreeSequence.getAnonymity(finalSequence) + "-anonym");
		}

		final long overallTime = System.currentTimeMillis() - time;
		try
		{
			final File outputFile = new File(outputPath + "/" + inputFile.getName() + ".log");
			final FileOutputStream fos = new FileOutputStream(outputFile);
			fos.write(("#Input file: " + inputFile.getName() + "\n").getBytes());
			fos.write(("#Amount od nodes = " + graphSize + "\n").getBytes());
			fos.write(("#Amount of edge = " + edgeSetSize + "\n").getBytes());
			fos.write(("#k = " + k + "\n").getBytes());
			fos.write(("#Delta = " + delta + "\n").getBytes());
			fos.write(("#Anonymity = " + anonymity + "\n").getBytes());
			fos.write(("#Processing time = " + (double) overallTime / 1000 + " seconds \n").getBytes());
			if ( solutionSequence == null )
			{
				fos.write("#No solution found.\n".getBytes());
			}
			else
			{
				fos.write("#Solution:\n".getBytes());
				for (final IntTupel edge : edgeSet)
				{
					fos.write((edge.getFirst() + " " + edge.getSecond() + "\n").getBytes());
				}
				fos.write("\n".getBytes());

			}
			fos.close();
		} catch (final IOException e)
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

}
