package de.vibee.kanon.data;

import java.util.Arrays;
import java.util.Comparator;

public class DegreeSequence
{

	// private AdjacentGraph graph = null;

	// public DegreeSequence(AdjacentGraph graph){
	// this.graph = graph;
	// }

	// public DegreeSequence(){
	//
	// }

	/**
	 * Calculates the original degree sequence of WeightedGraph g
	 * 
	 * @return sorted degree Sequence of WeightedGraph g
	 */
	public static Node[] calcDegreeSequence(final AdjacentGraph graph)
	{
		final Node degreeSequence[] = new Node[graph.getAmountOfVertices() + 1];
		for (int i = 1; i < graph.getAmountOfVertices() + 1; i++)
		{
			degreeSequence[i] = new Node(i, graph.getDegree(i));
		}
		DegreeSequence.sortSequenceByDegree(degreeSequence);
		for (int i = 1; i < degreeSequence.length; i++)
		{
			degreeSequence[i].setSortedPosition(i);
		}
		return degreeSequence;
	}

	public static int getAnonymity(final int degSeq[])
	{

		int anonymity = Integer.MAX_VALUE;
		int gap = 0;
		for (int i = 1; i < degSeq.length; i += gap - 1)
		{
			gap = 0;
			for (int j = i; j < degSeq.length; j++)
			{
				gap++;
				if ( degSeq[i] != degSeq[j] )
				{
					gap = j - i;
					anonymity = Math.min(anonymity, gap);
					break;
				}
			}
			anonymity = Math.min(gap, anonymity);
		}
		return anonymity;

	}

	/**
	 * Calculates anonymity of a degree sequence
	 * 
	 * @param degSeq Degree sequence, whose anonymity is being calculated
	 * @return Anonymity of degree sequence
	 */
	public static int getAnonymity(final Node degSeq[])
	{

		final int seq[] = new int[degSeq.length];
		for (int i = 1; i < degSeq.length; i++)
		{
			seq[i] = degSeq[i].getDegree();
		}

		return DegreeSequence.getAnonymity(seq);

	}

	public static int getDelta(final int degSeq[])
	{

		int maxDegree = Integer.MIN_VALUE;
		for (int i = 1; i < degSeq.length; i++)
		{
			maxDegree = Math.max(maxDegree, degSeq[i]);
		}

		return maxDegree;
	}

	/**
	 * Calculates Delta (maximum degree)
	 * 
	 * @return Delta
	 */
	public static int getDelta(final Node degSeq[])
	{

		final int seq[] = new int[degSeq.length];
		for (int i = 1; i < degSeq.length; i++)
		{
			seq[i] = degSeq[i].getDegree();
		}

		return DegreeSequence.getDelta(seq);
	}

	/**
	 * 
	 * @param nodeSeq Sorted degree sequence by degree
	 */
	public static void sortSequenceByDegree(final Node nodeSeq[])
	{
		Arrays.sort(nodeSeq, 1, nodeSeq.length, new Comparator<Node>() {

			@Override
			public int compare(final Node o1, final Node o2)
			{
				return o1.getDegree() - o2.getDegree();
			}

		});
	}

	/**
	 * 
	 * @param nodeSeq Sorted degree sequence by node index
	 */
	public static void sortSequenceByIndex(final Node nodeSeq[])
	{
		Arrays.sort(nodeSeq, 1, nodeSeq.length, new Comparator<Node>() {

			@Override
			public int compare(final Node o1, final Node o2)
			{
				return o1.getPosition() - o2.getPosition();
			}
		});
	}

}
