package de.vibee.kanon.tools;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.hash.TIntIntHashMap;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.Node;

public class DegreeAnonymizer
{

	public static boolean simpleErdosGallaiConstraint(final int degreeSequence[])
	{
		final int n = degreeSequence.length - 1;
		if ( n == 1 )
			return false;
		for (int r = 1; r < n; r++)
		{

			int leftSum = 0;

			for (int i = 1; i <= r; i++)
			{
				leftSum += degreeSequence[i];
			}

			int rightSum = 0;

			for (int i = r + 1; i <= n; i++)
			{
				rightSum += Math.min(r, degreeSequence[i]);
			}

			if ( leftSum > (long) r * (r - 1) + rightSum )
				return false;
			System.out.println("simple: " + leftSum + "; " + (rightSum + r * (r - 1)));

		}

		return true;
	}

	public static boolean simpleErdosGallaiConstraint(final int solutionSeq[], final int originalSeq[])
	{

		int count = 0;
		for (int i = 1; i < solutionSeq.length; i++)
		{
			if ( i >= originalSeq.length || solutionSeq[i] != originalSeq[i] )
			{
				count++;
			}
		}

		final Integer solutionSequenceArray[] = new Integer[count];

		for (int i = 1; i < solutionSeq.length; i++)
		{
			if ( i >= originalSeq.length )
			{
				solutionSequenceArray[i] = solutionSeq[i];
			}
			else if ( solutionSeq[i] > originalSeq[i] )
			{
				solutionSequenceArray[i] = solutionSeq[i] - originalSeq[i];
			}
		}
		Arrays.sort(solutionSequenceArray, new Comparator<Integer>() {
			@Override
			public int compare(final Integer o1, final Integer o2)
			{
				return o2 - o1;
			}
		});

		final int degreeSequence[] = new int[solutionSequenceArray.length];

		for (int i = 1; i < solutionSequenceArray.length; i++)
		{
			degreeSequence[i] = solutionSequenceArray[i];
		}

		return DegreeAnonymizer.simpleErdosGallaiConstraint(degreeSequence);
		// Collections.sort(solutionSequenceList, Collections.reverseOrder());
		// // sort descending
		// solutionSequenceList.add(0,0); // add stub at position 0
		// Integer solutionSequenceArray[] = (Integer[])
		// solutionSequenceList.toArray(new
		// Integer[solutionSequenceList.size()]);

	}

	// List<int[]> test;
	int solutions = 0;

	AdjacentGraph graph;

	Node originalDegSeq[];

	int delta;

	boolean failedBuilding = false;

	public DegreeAnonymizer(final AdjacentGraph graph, final Node degSeq[], final int delta)
	{
		this.graph = graph;
		this.originalDegSeq = degSeq;
		this.delta = delta;
	}

	// fehlerhaft!
	private boolean advancedErdosGallaiConstraint(final int solutionSeq[], final Node degSeq[])
	{

		final List<Integer> solutionSequenceList = new LinkedList<>(); // only
																		// contains
																		// affected
																		// nodes
																		// (solution
																		// degree
																		// -
																		// original
																		// degree)
		final int degrees[] = new int[solutionSeq[solutionSeq.length - 1] + 1];
		for (int i = 1; i < solutionSeq.length; i++)
		{
			if ( solutionSeq[i] != degSeq[i].getDegree() )
			{
				solutionSequenceList.add(solutionSeq[i] - degSeq[i].getDegree());
				degrees[solutionSeq[i] - degSeq[i].getDegree()]++;
			}
		}

		Collections.sort(solutionSequenceList, Collections.reverseOrder()); // sort
																			// descending
		solutionSequenceList.add(0, 0); // add stub at position 0
		final Integer solutionSequenceArray[] = solutionSequenceList.toArray(new Integer[solutionSequenceList.size()]);

		long sum = 0; // the left-hand sum

		for (int r = 1; r < solutionSequenceArray.length - 1; r++)
		{

			sum += solutionSequenceArray[r]; // dynamically save left-hand sum
												// instead of
												// re-calculating each iteration

			// magically calculate right-hand sum in O(delta)

			long sum2 = 0; // right-hand sum
			int tempsum = 0;
			final int i = r + 1;
			final int degOfI = solutionSequenceArray[i]; // degree of node i =
															// r+1
			int j = 0; // iterates over the degrees in the sequence
			while (tempsum < i)
			{
				tempsum += degrees[j]; // count vertices as long as we have not
										// reached vertice i
				j++;
			}
			if ( r <= degOfI )
			{
				sum2 += (tempsum - i + 1) * r;
			}
			else
			{
				sum2 += (tempsum - i + 1) * degOfI;
			}

			while (j < degrees.length)
			{
				final int amountOfJDegree = degrees[j];
				if ( r <= j )
				{
					sum2 += r * amountOfJDegree;
				}
				else
				{
					sum2 += j * amountOfJDegree;
				}
				j++;
			}

			if ( sum > (long) r * (r - 1) + sum2 )
				return false;

		}

		return true;

	}

	private int affectedNodes(final Node degSeq[], final int a, final int b, final int d)
	{

		// if (d > degSeq[b].getDegree()) return b-a;

		int affectedNodes = 0;
		for (int t = a; t <= b; t++)
		{
			if ( degSeq[t].getDegree() != d )
			{
				affectedNodes++;
			}
		}
		return affectedNodes;
	}

	private int[] buildSolutionSequence(final List<TIntArrayList[]> table, final Node degSeq[], final int k, final int pos)
	{
		int i = degSeq.length - 1;
		int j = table.size() - 1;
		int d = table.get(j)[i].get(pos + 2);
		int l = table.get(j)[i].get(pos + 3);
		int pre = table.get(j)[i].get(pos + 4);
		final int solutionSeq[] = new int[degSeq.length];

		while (i >= 2 * k)
		{

			final int a = i - l;

			final int b = j - this.cost(degSeq, i - l + 1, i, d);
			if ( b >= 0 && table.get(b)[i - l] != null )
			{

				this.raiseSeq(solutionSeq, a + 1, i, d);
				final TIntList pretable = table.get(b)[i - l].subList(pre, pre + 5);

				i = a;
				j = b;

				if ( i >= 2 * k )
				{
					pre = pretable.get(4);
				}
				d = pretable.get(2);
				l = pretable.get(3);

			}

		}

		this.raiseSeq(solutionSeq, 1, i, d);

		return solutionSeq;

	}

	public int[] calcAffectedSequence(final int k, final boolean greedy)
	{

		final TIntIntHashMap backup = new TIntIntHashMap();
		final TIntArrayList adjustedNodes = new TIntArrayList();
		int initCost = 0;
		int initAffectedNodes = 0;
		int initMaxRaise = 0;
		for (int t = this.originalDegSeq.length - k; t < this.originalDegSeq.length - 1; t++)
		{
			final int actCost = this.delta - this.originalDegSeq[t].getDegree();
			if ( actCost > 0 )
			{
				initMaxRaise = Math.max(initMaxRaise, actCost);
				initAffectedNodes++;
				initCost += actCost;
				backup.put(t, this.originalDegSeq[t].getDegree());
				adjustedNodes.add(t);
				this.originalDegSeq[t].setDegree(this.originalDegSeq[this.originalDegSeq.length - 1].getDegree());
			}

		}

		for (int t = 1; t < k; t++)
		{
			final int actCost = this.originalDegSeq[k].getDegree() - this.originalDegSeq[t].getDegree();
			if ( actCost > 0 )
			{
				initAffectedNodes++;
				initMaxRaise = Math.max(initMaxRaise, actCost);
				initCost += actCost;
				backup.put(t, this.originalDegSeq[t].getDegree());
				adjustedNodes.add(t);
				this.originalDegSeq[t].setDegree(this.originalDegSeq[k].getDegree());
			}

		}

		System.out.println("#Saved " + initCost + " iteration by adjusting last " + k + " nodes.");

		final List<TIntArrayList[]> table = new ArrayList<>(300);
		int j = 0;
		int highestI = 0;
		while (true)
		{

			table.add(new TIntArrayList[this.originalDegSeq.length]);

			for (int i = k; i < 2 * k && i < this.originalDegSeq.length; i++)
			{

				for (int d = this.originalDegSeq[i].getDegree(); d <= this.delta + 2; d++)
				{
					final int cost = this.cost(this.originalDegSeq, 1, i, d);
					final int affNodes = this.affectedNodes(this.originalDegSeq, 1, i, d);
					final int maxRaise = this.maxCost(this.originalDegSeq, 1, i, d);

					if ( j == cost )
					{

						if ( table.get(j)[i] == null )
						{
							table.get(j)[i] = new TIntArrayList(5);
						}

						boolean add = true;
						for (int pos = 0; pos < table.get(j)[i].size(); pos += 5)
						{
							if ( table.get(j)[i].get(pos) >= affNodes && table.get(j)[i].get(pos + 1) <= maxRaise )
							{
								add = false;
								break;
							}
						}
						if ( add )
						{ // don't make an entry if there already is an entry
							// equal or better than
							// the actual one
							// a better entry has equal or more affected nodes
							// and equal or less max
							// raise
							final int t[] = new int[] { affNodes, maxRaise, d, 1, 0 };
							table.get(j)[i].addAll(t);
							highestI = Math.max(highestI, i);
						}

					}

					else if ( j < cost )
					{ // cost will be even higher with higher d
						break;
					}
				}

			}

			for (int i = 2 * k; i < this.originalDegSeq.length; i++)
			{
				if ( i - 2 * k > highestI )
				{
					break; // if there has not been an entry made yet at any
							// T[i,j], i > highestI +
							// 2k, break (we only look 2k-1 back in the table)
				}
				LBREAK: for (int l = k; l <= 2 * k - 1; l++)
				{
					for (int d = this.originalDegSeq[i].getDegree(); d <= this.delta + 2; d++)
					{
						final int subSeqCost = j - this.cost(this.originalDegSeq, i - l + 1, i, d);
						final int affNodes = this.affectedNodes(this.originalDegSeq, i - l + 1, i, d);
						final int maxRaise = this.maxCost(this.originalDegSeq, i - l + 1, i, d);

						if ( subSeqCost >= 0 )
						{
							// System.out.println(subSeqCost);
							final TIntArrayList pre = table.get(subSeqCost)[i - l];
							if ( pre != null )
							{

								for (int o = 0; o < table.get(subSeqCost)[i - l].size(); o += 5)
								{
									final TIntList entry = table.get(subSeqCost)[i - l].subList(o, o + 5);
									final int preAffectedNodes = entry.get(0);
									final int preMaxRaise = entry.get(1);
									if ( table.get(j)[i] == null )
									{
										table.get(j)[i] = new TIntArrayList(5);
									}
									boolean add = true;
									boolean crap = false;
									final List<Integer> removeCrap = new LinkedList<>();

									for (int pos = 0; pos < table.get(j)[i].size(); pos += 5)
									{

										if ( table.get(j)[i].get(pos) >= affNodes + preAffectedNodes
												&& table.get(j)[i].get(pos + 1) <= Math.max(maxRaise, preMaxRaise) )
										{
											add = false;
											break;
										}
										else if ( table.get(j)[i].get(pos) < affNodes + preAffectedNodes
												&& table.get(j)[i].get(pos + 1) > Math.max(maxRaise, preMaxRaise) )
										{
											crap = true;
											removeCrap.add(pos);
										}
									}

									if ( add )
									{
										final int t[] = new int[] { affNodes + preAffectedNodes, Math.max(maxRaise, preMaxRaise), d, l, o };
										highestI = Math.max(highestI, i);
										table.get(j)[i].addAll(t);
									}

									for (int t = removeCrap.size() - 1; t >= 0; t--)
									{
										final int pos = removeCrap.get(t);
										table.get(j)[i].removeAt(pos);
										table.get(j)[i].removeAt(pos);
										table.get(j)[i].removeAt(pos);
										table.get(j)[i].removeAt(pos);
										table.get(j)[i].removeAt(pos);
									}

								}

							}

						}
						else
						{
							if ( d == this.originalDegSeq[i].getDegree() )
							{
								break LBREAK; // subsequence cost will only
												// raise with increasing l
							}
							else
							{
								break;
							}
						}
					}

				}

			}

			if ( (j + initCost) % 2 == 0 && table.get(j)[this.originalDegSeq.length - 1] != null )
			{
				final TIntList tupelList = table.get(j)[this.originalDegSeq.length - 1];
				boolean returnNull = false;
				for (int pos = 0; pos < tupelList.size(); pos += 5)
				{
					if ( tupelList.get(pos) + initAffectedNodes > Math.max(tupelList.get(pos + 1), initMaxRaise) || tupelList.get(pos) == 0
							&& tupelList.get(pos + 1) == 0 )
					{
						System.out.println("Sequence is " + k + "-anonymized by " + (j + initCost) / 2 + " added edges ( " + (j + initCost)
								+ " ) modification)");
						returnNull = true;

						final int solutionSeq[] = this.buildSolutionSequence(table, this.originalDegSeq, k, pos);

						// restore the degree sequence without pre affected
						// nodes for graph building
						// (magic)
						final TIntIntHashMap backupModifiedDegrees = new TIntIntHashMap();
						TIntIterator it = adjustedNodes.iterator();
						while (it.hasNext())
						{
							final int index = it.next();
							backupModifiedDegrees.put(index, this.originalDegSeq[index].getDegree());
							this.originalDegSeq[index].setDegree(backup.get(index));
						}

						if ( this.simpleErdosGallaiConstraint(solutionSeq, this.originalDegSeq) == true )
						{
							System.out.println("Erdos-Gallai check sais yes");
						}
						else
						{
							System.out.println("Erdos-Gallai sais no");
						}

						if ( new GreedyGraphConstructor(this.graph).constructGraph(this.originalDegSeq, solutionSeq, true) != null )
							return solutionSeq;

						it = adjustedNodes.iterator();
						while (it.hasNext())
						{
							final int index = it.next();
							this.originalDegSeq[index].setDegree(backupModifiedDegrees.get(index));
						}

					}
				}
				if ( returnNull && !greedy )
					return null;
			}

			System.out.println("#Progress: Anonymized  " + (highestI - 1) + " of " + this.graph.getAmountOfVertices() + " nodes with cost = "
					+ (j + initCost));
			++j;

		}

	}

	public Node[] calcSequence(final Node reducedDegSeq[], final int s, final int k, final boolean greedy)
	{

		// int initCost = 0;

		final List<BitSet> table = new ArrayList<>(300);

		for (int j = 0; j <= 2 * s; j++)
		{
			System.out.println("j = " + j);

			table.add(new BitSet(reducedDegSeq.length));

			for (int i = k; i < 2 * k && i < reducedDegSeq.length; i++)
			{

				for (int d = reducedDegSeq[i].getDegree(); d <= this.delta + 2; d++)
				{
					final int cost = this.cost(reducedDegSeq, 1, i, d);

					if ( j == cost )
					{

						table.get(j).set(i);
						break;

					}

					else if ( j < cost )
					{ // cost will be even higher with higher d
						break;
					}
				}

			}

			for (int i = 2 * k; i < reducedDegSeq.length; i++)
			{
				// if ((i - 2*k) > highestI) break; //if there has not been an
				// entry made yet at any
				// T[i,j], i > highestI + 2k, break (we only look 2k-1 back in
				// the table)
				LBREAK: for (int l = k; l <= 2 * k - 1; l++)
				{
					for (int d = reducedDegSeq[i].getDegree(); d <= this.delta + 2; d++)
					{
						final int subSeqCost = j - this.cost(reducedDegSeq, i - l + 1, i, d);

						if ( subSeqCost >= 0 )
						{

							if ( table.get(subSeqCost).get(i - l) )
							{
								table.get(j).set(i);
								break LBREAK;

							}

						}
						else
						{
							if ( d == reducedDegSeq[i].getDegree() )
							{
								break LBREAK; // subsequence cost will only
												// raise with increasing l
							}
							else
							{
								break;
							}
						}
					}

				}

			}

			if ( j % 2 == 0 && table.get(j).get(reducedDegSeq.length - 1) )
			{

				final int solution[] = new int[reducedDegSeq.length];
				for (int i = 1; i < reducedDegSeq.length; i++)
				{
					solution[i] = reducedDegSeq[i].getDegree();
				}
				this.failedBuilding = false;
				final int sol[] = this.getPossibleSolutionSequences(reducedDegSeq, table, k, reducedDegSeq.length - 1, j, solution);
				if ( sol != null )
				{

					final Node red2[] = new Node[sol.length];
					for (int i = 1; i < sol.length; i++)
					{
						final Node n = new Node(reducedDegSeq[i].getPosition(), sol[i]);
						n.setSortedPosition(reducedDegSeq[i].getSortedPosition());
						red2[i] = n;
					}
					return red2;
				}
				else if ( sol == null && this.failedBuilding )
					return null;

			}
		}

		return null;

	}

	private int cost(final Node degSeq[], final int a, final int b, final int d)
	{

		if ( degSeq[a].getDegree() == d )
			return 0;

		if ( b >= degSeq.length )
			return Integer.MAX_VALUE;
		int cost = 0;
		for (int t = a; t <= b; t++)
		{
			cost += d - degSeq[t].getDegree();
		}
		return cost;
	}

	private int[] getPossibleSolutionSequences(final Node degSeq[], final List<BitSet> table, final int k, final int i, final int j,
			final int[] solution)
	{

		if ( i >= 2 * k )
		{

			boolean split = false;

			DBREAK: for (int d = degSeq[i].getDegree(); d <= this.delta + 2; d++)
			{
				for (int l = k; l < 2 * k; l++)
				{
					final int cost = this.cost(degSeq, i - l + 1, i, d);
					if ( j - cost >= 0 )
					{
						if ( table.get(j - cost).get(i - l) )
						{
							if ( !split )
							{
								final int backup[] = Arrays.copyOfRange(solution, i - l + 1, i + 1);
								this.raiseSeq(solution, i - l + 1, i, d);
								split = true;

								if ( this.getPossibleSolutionSequences(degSeq, table, k, i - l, j - cost, solution) != null )
									return solution;
								for (int m = l - 1; m >= 0; m--)
								{
									solution[i - m] = backup[l - 1 - m];
								}
							}
							else
							{
								this.solutions++;
								if ( this.solutions % 100 == 0 )
								{
									System.out.println(this.solutions + " solutions found");
								}

								final int backup[] = Arrays.copyOfRange(solution, i - l + 1, i + 1);
								this.raiseSeq(solution, i - l + 1, i, d);

								if ( this.getPossibleSolutionSequences(degSeq, table, k, i - l, j - cost, solution) != null )
									return solution;
								for (int m = l - 1; m >= 0; m--)
								{
									solution[i - m] = backup[l - 1 - m];
								}

							}
						}
					}
					else
					{
						break DBREAK; // cost too high; higher d or l will
										// increase cost even more
					}

				}
			}
			return null;

		}
		if ( i < 2 * k && j > 0 )
		{
			for (int d = degSeq[i].getDegree(); d <= this.delta + 2; d++)
			{
				if ( this.cost(degSeq, 1, i, d) == j )
				{
					this.raiseSeq(solution, 1, i, d);
				}
			}
		}

		if ( this.simpleErdosGallaiConstraint(solution, degSeq) )
		{

			final int solutionSequence[] = new int[this.originalDegSeq.length];

			for (int m = 1; m < solutionSequence.length; m++)
			{
				solutionSequence[m] = this.originalDegSeq[m].getDegree();

			}

			for (int m = 1; m < solution.length; m++)
			{
				solutionSequence[degSeq[m].getSortedPosition()] = solution[m];
			}

			if ( new GreedyGraphConstructor(this.graph).constructGraph(this.originalDegSeq, solutionSequence, true) != null )
				return solution;
			else
			{
				this.failedBuilding = true;
			}

		}

		return null;

	}

	private int maxCost(final Node degSeq[], final int a, final int b, final int d)
	{
		return d - degSeq[a].getDegree();
	}

	private void raiseSeq(final int solutionSeq[], final int a, final int b, final int d)
	{
		Arrays.fill(solutionSeq, a, b + 1, d);
		// for (int t = a; t <= b; t++){
		// solutionSeq[t] = d;
		// }
	}

	private Node[] reductionRuleOne(final Node degSeq[], final int s, final int k)
	{

		final List<Node>[] boxes = new LinkedList[this.delta + 1];

		for (int i = 0; i < boxes.length; i++)
		{
			boxes[i] = new LinkedList();
		}

		for (final Node node : degSeq)
		{
			if ( node != null )
			{
				boxes[node.getDegree()].add(node);
			}
		}

		boolean foundLargeBlock = false;

		int amountOfRemovedNodes = 0;
		int largebox = 0;
		for (int i = boxes.length - 1; i >= 0; i--)
		{
			if ( !foundLargeBlock && boxes[i].size() > 2 * s + k )
			{
				foundLargeBlock = true;
				largebox = i;
				while (boxes[i].size() > 2 * s + k)
				{
					boxes[i].remove(0);
					amountOfRemovedNodes++;
				}
				break;
			}
		}

		if ( foundLargeBlock )
		{
			for (int i = boxes.length - 1; i >= 0; i--)
			{
				if ( i != largebox )
				{
					if ( foundLargeBlock && boxes[i].size() > 3 * k )
					{
						while (boxes[i].size() > 3 * k)
						{
							boxes[i].remove(0);
							amountOfRemovedNodes++;
						}

					}
				}

			}
		}

		System.out.println("Removed " + amountOfRemovedNodes + " nodes.");
		final Node reducedDegSeq[] = new Node[degSeq.length - amountOfRemovedNodes];

		int index = 1;
		for (final List<Node> box : boxes)
		{
			for (final Node node : box)
			{
				reducedDegSeq[index++] = node;
			}
		}

		return reducedDegSeq;

	}

	// private boolean getTableValue(List<List<BitSet>> table, int j, int x, int
	// i){
	//
	// if (j >= 0 && j < table.size() && x >= 0 && x < table.get(j).size() && i
	// >= 0 && i <
	// table.get(j).get(x).size()){
	// return table.get(j).get(x).get(i);
	// }
	// return false;
	// }

	public boolean simpleErdosGallaiConstraint(final int solutionSeq[], final Node degSeq[])
	{

		final int originalSeq[] = new int[degSeq.length];

		for (int i = 1; i < degSeq.length; i++)
		{
			originalSeq[i] = degSeq[i].getDegree();
		}

		return DegreeAnonymizer.simpleErdosGallaiConstraint(solutionSeq, originalSeq);

	}

	public Node[] startSequenceCalculation(final int k, final boolean greedy)
	{

		for (int s = 0; s < Math.pow(this.originalDegSeq.length, 2); s++)
		{
			final Node reducedSeq[] = this.reductionRuleOne(this.originalDegSeq, s, k);
			final Node reducedSolutionSeq[] = this.calcSequence(reducedSeq, s, k, greedy);
			if ( reducedSolutionSeq != null )
				return reducedSolutionSeq;
			else if ( reducedSolutionSeq == null && this.failedBuilding )
				return null;
		}
		return null;
	}

}
