package de.vibee.kanon.tools;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import seppImpl.CompactDegreeSequence;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.tools.IntTripel;

public class ErdosGallaiAdvanced implements ErdosGallaiInterface
{

	public long time1 = 0, time2 = 0, noOfEGCalls = 0, maxRFail = 0;

	private AdjacentGraph graph;

	private CompactDegreeSequence init;

	public ErdosGallaiAdvanced(final AdjacentGraph graph, final CompactDegreeSequence init)
	{
		this.setGraph(graph, init);
	}

	/**
	 * 
	 * @param init initial compact degree sequence
	 * @param sol solution compact degree sequence
	 * @param fixAll fix all vertices, even the ones with degree raise = 1
	 * @return Realizability of solution degree sequence
	 */
	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll)
	{
		// System.out.println("here");
		this.noOfEGCalls++;
		// return 5;
		final List<IntTripel> degreeRaiseList = new ArrayList<>(1000); // For
																		// each
																		// vertex
																		// whose
		// degree has to
		// be raised: Store original vertex
		// degree (first) and solution vertex
		// degree (second), ordered descending
		// final int[] noOfIncreasesPerDegree = new int[init.getMaxValue() + 1];
		degreeRaiseList.add(null);
		int sumIncreases = 0;
		final TIntSet solutionDegreeSet = new TIntHashSet();
		final List<IntTripel> degreeBlockRaiseList = new BlockRaiseListComputer().generateBlockRaiseListBottomUp(this.init, sol);
		for (final IntTripel t : degreeBlockRaiseList)
		{

			for (int i = 0; i < t.getThird(); i++)
			{
				// if (graph.getAllNodesWithDegree(t.getFirst()).size() >=
				// t.getThird())
				if ( solutionDegreeSet.contains(t.getFirst()) )
				{
					degreeRaiseList.add(new IntTripel(t.getFirst(), t.getSecond(), 0));
				}
				else
				{
					degreeRaiseList.add(new IntTripel(t.getFirst(), t.getSecond(), 1));
				}
				solutionDegreeSet.add(t.getSecond());
				// else degreeRaiseList.add(new IntTripel(t.getFirst(),
				// t.getSecond(), 0));
				// noOfIncreasesPerDegree[t.getFirst()]++;
				sumIncreases += t.getSecond() - t.getFirst();
				// if (graph.getAllNodesWithDegree(t.getFirst()).size() >=
				// t.getThird()){
				//
				// }
				// else{
				//
				// }
			}

		}

		Collections.sort(degreeRaiseList, new Comparator<IntTripel>() {
			@Override
			public int compare(final IntTripel o1, final IntTripel o2)
			{
				if ( o1 == null )
					return -1;
				if ( o2 == null )
					return 1;
				if ( o2.getSecond() - o2.getFirst() != o1.getSecond() - o1.getFirst() )
					return o1.getSecond() - o1.getFirst() - (o2.getSecond() - o2.getFirst());
				else
					return o2.getSecond() - o1.getSecond();
			}

		});

		int sum1 = 0;
		// System.out.println("RaiseList: " + degreeRaiseList);
		int minimalSumSoFar = 0;

		for (int r = 1; r < degreeRaiseList.size() - 1; r++)
		{ // r = 1,...,n-1
			// System.out.println(r);
			long tmp = System.currentTimeMillis();

			// calculate left sum
			final IntTripel t = degreeRaiseList.get(r);
			sum1 += t.getSecond() - t.getFirst();

			final int amountOfNodesWithDegree = this.graph.getAmountOfNodesWithDegree(t.getFirst());
			// final int noOfIncreases = noOfIncreasesPerDegree[t.getFirst()];

			// calculate (r*(r-1)) first. Then iterate over first r entries in
			// degree raise list
			// (index j), for each j iterate over entries j+1,...,r entries. If
			// for a pair of
			// entries t1, t2 there
			// is only one vertex with vertex degree stored in entry t1 and only
			// one vertex with
			// vertex degree stored in entry t2 AND these vertices are adjacent,
			// substract 2 from
			// the product.
			// In order to fit to the data reduction rules from Degree
			// Anonymization, we only do
			// this if a vertex degree has to be raised by more than 1.
			int mul = r * (r - 1);
			if ( mul > 0 )
			{
				int j = r;
				// while (j < r && mul > 0)
				{
					final IntTripel t1 = degreeRaiseList.get(j);
					// if (t1.getThird() == 1){
					if ( (fixAll || t1.getSecond() - t1.getFirst() != 1) && t1.getThird() == 1 )
					// graph.getAmountOfNodesWithDegree(t1.getFirst()) <
					// sumIncreases &&
					{
						int minimalSum = Integer.MAX_VALUE;
						final TIntIterator nodeOneDegreeIterator = this.graph.getAllNodesWithDegree(t1.getFirst()).iterator();
						while (nodeOneDegreeIterator.hasNext() && minimalSum > 0)
						{
							final int nodet1 = nodeOneDegreeIterator.next();
							int sum = 0;
							final TIntSet alreadyTaken = new TIntHashSet();
							int l = 1;
							while (l < j)
							{
								final IntTripel t2 = degreeRaiseList.get(l);
								// if (t2.getThird() == 1){
								if ( (fixAll || t2.getSecond() - t2.getFirst() != 1) && t2.getThird() == 1 )
								{
									final TIntIterator nodeTwoDegreeIterator = this.graph.getAllNodesWithDegree(t2.getFirst()).iterator();
									boolean substract = true;
									while (nodeTwoDegreeIterator.hasNext())
									{
										final int nodet2 = nodeTwoDegreeIterator.next();
										if ( !this.graph.isAdjacent(nodet1, nodet2) )
										{
											if ( alreadyTaken.add(nodet2) )
											{
												substract = false;
												break;
											}

										}
									}
									if ( substract )
									{
										sum += 2;
									}

								}
								++l;
								// }
							}
							minimalSum = Math.min(minimalSum, sum);
						}
						if ( minimalSum != Integer.MAX_VALUE )
						{

							minimalSumSoFar += minimalSum;
							mul -= minimalSumSoFar;
						}

					}
					// }

					++j;
				}
			}
			this.time1 += System.currentTimeMillis() - tmp;
			tmp = System.currentTimeMillis();
			// Calculation of the right side sum. This code contains a lot of
			// magic. If you feel
			// desperate
			// cause you do not understand anything, and you think that I suck
			// because I did not
			// comment the code,
			// take a look at some pictures of the Hydrochoerus hydrochaeris.
			// You will feel more
			// comfortable then.
			int sum2 = 0;
			int i = r + 1;
			boolean decreased = false;
			while (i < degreeRaiseList.size())
			{

				final IntTripel t1 = degreeRaiseList.get(i);
				sum2 += Math.min(r, t1.getSecond() - t1.getFirst()); // just
																		// what's
																		// the
																		// algorithm
																		// of
				// erdos & gallai

				if ( (fixAll || t1.getSecond() - t1.getFirst() != 1) && t1.getThird() == 1 )
				// graph.getAmountOfNodesWithDegree(t1.getFirst()) <
				// sumIncreases &&
				{ // cause of the data reduction rules in DegAnon..

					final TIntIterator nodeOneDegreeIterator = this.graph.getAllNodesWithDegree(t1.getFirst()).iterator();
					int maxNonAdjacentNodes = 0; // Here we save how much we can
													// substract in the
					// end for the i-th vertex

					while (nodeOneDegreeIterator.hasNext() && maxNonAdjacentNodes <= Math.min(r, t1.getSecond() - t1.getFirst()))
					{ // inspect all vertices which come into consideration for
						// the i-th raise
						final int nodet1 = nodeOneDegreeIterator.next();

						int nonAdjacentNodes = 0;

						for (int x = 1; x <= r; x++)
						{ // inspect the first r vertices ...
							final IntTripel t2 = degreeRaiseList.get(x);
							if ( t2.getThird() == 0 )
							{
								nonAdjacentNodes = Integer.MAX_VALUE;
								break;
							}
							else
							{
								if ( fixAll || t2.getSecond() - t2.getFirst() != 1 )
								// graph.getAmountOfNodesWithDegree(t2.getFirst())
								// < sumIncreases &&
								{
									final TIntIterator nodeTwoDegreeIterator = this.graph.getAllNodesWithDegree(t2.getFirst()).iterator();

									while (nodeTwoDegreeIterator.hasNext())
									{ // .. where for each vertex, we look at
										// all that come into
										// consideration
										final int nodet2 = nodeTwoDegreeIterator.next();

										if ( !this.graph.isAdjacent(nodet1, nodet2) )
										{ // in this case, we can connect node i
											// to a node x with
											// specific vertex degree
											nonAdjacentNodes++;
											break;
										}

									}

								}
							}

						}

						maxNonAdjacentNodes = Math.max(maxNonAdjacentNodes, nonAdjacentNodes);

					}

					if ( r <= t1.getSecond() - t1.getFirst() )
					{
						if ( r - maxNonAdjacentNodes > 0 )
						{
							decreased = true;
							sum2 -= r - maxNonAdjacentNodes;
						}
					}
					else
					{
						if ( t1.getSecond() - t1.getFirst() - maxNonAdjacentNodes > 0 )
						{
							decreased = true;
							sum2 -= t1.getSecond() - t1.getFirst() - maxNonAdjacentNodes;
						}
					}

				}

				++i;

			}
			this.time2 += System.currentTimeMillis() - tmp;

			// System.out.println(r);
			if ( sum1 > mul + sum2 )
			{
				// if ( SeppsMain.logLevel > 0 )
				// System.out.println("erdos gallai: fail for r = " + r +
				// ": sum left side: " + sum1
				// + ", sum right side: " + (mul + sum2));
				// System.out.println("diff: "+(sum1 - mul - sum2) + "\t r: " +
				// r + "\t liste: " +
				// degreeRaiseList.toString());
				this.maxRFail = Math.max(this.maxRFail, r);
				return sum1 - mul - sum2;
			}

		}
		// System.out.println("\t liste: " + degreeRaiseList.toString());
		return 0;
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff)
	{
		// TODO Auto-generated method stub
		return this.computeFeasibility(sol, fixAll);
	}

	@Override
	public long getMaxRFail()
	{
		// TODO Auto-generated method stub
		return this.maxRFail;
	}

	@Override
	public long getNoOfEGCalls()
	{
		// TODO Auto-generated method stub
		return this.noOfEGCalls;
	}

	@Override
	public long getTime1()
	{
		// TODO Auto-generated method stub
		return this.time1;
	}

	@Override
	public long getTime2()
	{
		// TODO Auto-generated method stub
		return this.time2;
	}

	@Override
	public long getTime3()
	{
		return 0;
	}

	@Override
	public void setGraph(final AdjacentGraph graph, final CompactDegreeSequence init)
	{
		// TODO Auto-generated method stub
		this.graph = graph;
		this.init = init;
	}
}
