package de.vibee.kanon.tools;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.hash.TObjectIntHashMap;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

import seppImpl.CompactDegreeSequence;
import dataStructures.Store;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreesChangeList;

public class FastErdosGallai implements ErdosGallaiInterface
{

	private class MyIntArrayKey
	{
		private final int[] data;

		private final int hash;

		public MyIntArrayKey(final int[] values)
		{
			this.data = values;
			this.hash = Arrays.hashCode(this.data);
		}

		@Override
		public boolean equals(final Object obj)
		{
			if ( obj != null && obj instanceof MyIntArrayKey )
			{
				final MyIntArrayKey s = (MyIntArrayKey) obj;
				return Arrays.equals(s.data, this.data);
			}
			return false;
		}

		@Override
		public int hashCode()
		{
			return this.hash;
		}

		@Override
		public String toString()
		{
			return "MyIntArrayKey [data=" + Arrays.toString(this.data) + "]";
		}

	}

	private long noOfCountingArgumentSufficient = 0, noOfLookUps = 0, noOfComputations = 0;

	private long noOfSolDependingLookUps = 0, noOfSolDependingFullLookUps = 0, noOfSolDependingComputations = 0, maxSolDepHashMapSize = 0;

	private final long timeforNoOfEdges = 0, timeforNoOfEdgesComputingOnly = 0;

	private long time1 = 0, time2 = 0, time3 = 0, noOfEGCalls = 0, maxRFail = 0;

	private DegreesChangeList degreeRaiseList;

	private AdjacentGraph graph;

	private CompactDegreeSequence init;

	// private HashMap<MyKey,Integer> noOfEdges;
	private final TObjectIntHashMap<MyIntArrayKey> noOfEdgesBetweenTwoBlocks;

	private final TObjectIntHashMap<MyIntArrayKey> SolDependingResults;

	public FastErdosGallai(final AdjacentGraph graph, final CompactDegreeSequence init)
	{
		this.noOfEdgesBetweenTwoBlocks = new TObjectIntHashMap<MyIntArrayKey>(10,0.5f,-1);
		this.SolDependingResults = new TObjectIntHashMap<MyIntArrayKey>();
		this.setGraph(graph, init);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll)
	{
		return this.computeFeasibility(sol, fixAll, true);
	}

	@Override
	public int computeFeasibility(final CompactDegreeSequence sol, final boolean fixAll, final boolean getMaxDiff)
	{
		this.noOfEGCalls++;
		if ( this.graph == null )
			return 0;

		final long tmp = System.currentTimeMillis();
		int maxDiff = 0;
		this.SolDependingResults.clear();
		// SolDependingResults = new TObjectIntHashMap<MyIntArrayKey>();

		// List of changes
		this.degreeRaiseList = sol.differenceWithDegrees(this.init);
		final TreeMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> degreeRaiseTree = this.degreeRaiseList.getTreeMap();

		long leftsum = 0, rightsum = 0;
		int r = 0; // the index from the Erdös Gallai Inequalities

		int edgesWithinTheCliquePart = 0, edgesBetweenParts = 0;
		int lastFromIncrease = -1;

		for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> entry : degreeRaiseTree.descendingMap().entrySet())
		{ // iterating over all increases (of vertex degrees) in descending
			// order
			final int increase = entry.getKey();
			final TreeMap<Integer, Integer> value = entry.getValue().getS2();
			final int amountByIncrease = entry.getValue().getS1();

			SortedMap<Integer, Integer> tmpSmallMap;
			SortedMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> tmpMap, tmpMap2;
			r += amountByIncrease;
			edgesBetweenParts = 0; // do not reset edgesWithinTheCliquePart!
			// System.out.println("r: "+r);

			leftsum += increase * amountByIncrease;
			rightsum = r * (r - 1);

			long tmp2 = System.currentTimeMillis();
			if ( r > 1 )
			{ // count the number of edges within the "clique-part"
				for (final Entry<Integer, Integer> innerEntry : value.descendingMap().entrySet())
				{// iterating over vertex degrees (in ascending order) getting
					// raised by increase
					final int origDegree = innerEntry.getKey();
					if ( !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && (fixAll || increase > 1) )
					{
						final int amountOfVertices = innerEntry.getValue();

						tmpMap = degreeRaiseTree.tailMap(increase + 1);
						edgesWithinTheCliquePart += this.edgesBetweenBigMap(increase, origDegree, amountOfVertices, tmpMap, fixAll);

						tmpSmallMap = value.tailMap(origDegree);
						edgesWithinTheCliquePart += this.edgesBetweenSmallMap(increase, origDegree, amountOfVertices, tmpSmallMap, increase, fixAll);
					}
				}
			}
			this.time2 += System.currentTimeMillis() - tmp2;

			tmpMap = degreeRaiseTree.headMap(increase);
			tmpMap2 = degreeRaiseTree.tailMap(increase);
			final int fromIncrease = tmpMap2.firstKey();
			final int toIncrease = tmpMap2.lastKey();
			for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> entry2 : tmpMap.entrySet())
			{ // count the number of edges between the two parts
				final int increase2 = entry2.getKey();
				final int amountByIncrease2 = entry2.getValue().getS1();
				rightsum += amountByIncrease2 * Math.min(r, increase2);
				for (final Entry<Integer, Integer> entry3 : entry2.getValue().getS2().entrySet())
				{
					final int origDegree2 = entry3.getKey();
					if ( !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree2) && (fixAll || increase2 > 1))
					{
						final int amountOfVertices2 = entry3.getValue();

						tmp2 = System.currentTimeMillis();

						int tmpEdgesBetweenParts = 0;

						final int[] val = { increase2, origDegree2, amountOfVertices2, fromIncrease, toIncrease };
						final MyIntArrayKey key = new MyIntArrayKey(val);

						final int[] val2 = { increase2, origDegree2, amountOfVertices2, lastFromIncrease, toIncrease };
						final MyIntArrayKey key2 = new MyIntArrayKey(val2);
						if ( this.SolDependingResults.containsKey(key2) )
						{
							this.noOfSolDependingLookUps++;
							tmpEdgesBetweenParts = this.SolDependingResults.get(key2);
							tmpEdgesBetweenParts += this.edgesBetweenSmallMap(increase2, origDegree2, amountOfVertices2, tmpMap2.get(fromIncrease)
									.getS2(), fromIncrease, fixAll);
							this.SolDependingResults.put(key, tmpEdgesBetweenParts);
						}
						else
						{
							this.noOfSolDependingComputations++;
							tmpEdgesBetweenParts = this.edgesBetweenBigMap(increase2, origDegree2, amountOfVertices2, tmpMap2, fixAll);
							this.SolDependingResults.put(key, tmpEdgesBetweenParts);
						}
						// if (tmpEdgesBetweenParts !=
						// edgesBetweenBigMap(increase2, origDegree2,
						// amountOfVertices2, tmpMap2))
						// {
						// System.err.println("ups");
						// System.exit(1);
						// }
						this.time3 += System.currentTimeMillis() - tmp2;
						if ( r * amountOfVertices2 < tmpEdgesBetweenParts + amountOfVertices2 * Math.min(r, increase2) )
						{
							edgesBetweenParts += tmpEdgesBetweenParts + amountOfVertices2 * Math.min(r, increase2) - r * amountOfVertices2;
						}
					}
				}
			}
			lastFromIncrease = fromIncrease;

			final long diff = leftsum + edgesWithinTheCliquePart + edgesBetweenParts - rightsum;
			if ( diff > 0 )
			{
				// System.out.println("diff: "+diff
				// + "\t left: " + leftsum
				// + "\t edgesWithinTheCliquePart: " + edgesWithinTheCliquePart
				// + "\t edgesBetweenParts: " + edgesBetweenParts
				// + "\t right: " + rightsum
				// + "\t r: " + r + "\t liste: " + degreeRaiseTree.toString());
				this.maxRFail = Math.max(this.maxRFail, r);
				if ( getMaxDiff )
				{
					maxDiff = Math.max(maxDiff, (int) diff);
				}
				else
				{
					this.maxSolDepHashMapSize = Math.max(this.maxSolDepHashMapSize, this.SolDependingResults.size());
					this.time1 += System.currentTimeMillis() - tmp;
					return (int) diff;
				}
			}

			if ( r * (r - 1) > 2 * leftsum )
			{ // the r * (r-1) part is large enough to satisfy the inequality...
				this.maxSolDepHashMapSize = Math.max(this.maxSolDepHashMapSize, this.SolDependingResults.size());
				this.time1 += System.currentTimeMillis() - tmp;
				return maxDiff;
			}
		}
		this.maxSolDepHashMapSize = Math.max(this.maxSolDepHashMapSize, this.SolDependingResults.size());
		this.time1 += System.currentTimeMillis() - tmp;
		return maxDiff;
	}

	private int edgesBetweenBigMap(final int increase, final int origDegree, final int amountOfVertices,
			final SortedMap<Integer, Store<Integer, TreeMap<Integer, Integer>>> tailMap, boolean fixAll)
	{
		int sum = 0;
		for (final Entry<Integer, Store<Integer, TreeMap<Integer, Integer>>> amountEntry : tailMap.entrySet())
		{
			final int increase2 = amountEntry.getKey();
			if ( !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && (fixAll || increase2 > 1))
			{
				final TreeMap<Integer, Integer> value2 = amountEntry.getValue().getS2();
				sum += this.edgesBetweenSmallMap(increase, origDegree, amountOfVertices, value2, increase2, fixAll);
			}
		}
		// time2 += (System.currentTimeMillis() - tmp);
		return sum;
	}

	private int edgesBetweenSmallMap(final int increase, final int origDegree, final int amountOfVertices, final SortedMap<Integer, Integer> value2,
			final int increase2, boolean fixAll)
	{
		int sum = 0;
		// final long tmp = System.currentTimeMillis();
		for (final Entry<Integer, Integer> innerEntry2 : value2.entrySet())
		{
			final int origDegree2 = innerEntry2.getKey();
			final int amountOfVertices2 = innerEntry2.getValue();
			if ( !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree) && !this.degreeRaiseList.getDegreeRaisedTo().contains(origDegree2)
					 && (fixAll || (increase2 > 1 && increase > 1)))
			{
				sum += this.getAmountOfEdgesBetweenDegreeBlocks(origDegree, origDegree2, amountOfVertices, amountOfVertices2);
			}
		}
		// time2 += (System.currentTimeMillis() - tmp);
		return sum;
	}

	public int getAmountOfEdgesBetweenDegreeBlocks(final int degree1, final int degree2, final int noOfDeg1Vertices, final int noOfDeg2Vertices)
	{
		// final long tmp = System.currentTimeMillis();
		if ( this.graph.getAmountOfNodesWithDegree(degree1) - noOfDeg1Vertices >= degree2 * noOfDeg2Vertices
				|| this.graph.getAmountOfNodesWithDegree(degree2) - noOfDeg2Vertices >= degree1 * noOfDeg1Vertices )
		{ // just a simple counting argument
			this.noOfCountingArgumentSufficient++;
			// this.time += (System.currentTimeMillis() - tmp);
			// this.timeforNoOfEdges += (System.currentTimeMillis() - tmp);
			return 0;
		}

		// long tmp2 = System.currentTimeMillis();
		final int[] myArray = { degree1, degree2, noOfDeg1Vertices, noOfDeg2Vertices };
		final MyIntArrayKey key = new MyIntArrayKey(myArray);
		// String key =
		// Integer.toString(degree1)+";"+Integer.toString(degree2)+";"+Integer.toString(noOfDeg1Vertices)
		// +";"+Integer.toString(noOfDeg2Vertices)+";"+Integer.toString(requiredEdgesPerVertex);
		// this.time += (System.currentTimeMillis() - tmp);

		// long tmp2 = System.currentTimeMillis();
//		final boolean containsKey = this.noOfEdgesBetweenTwoBlocks.containsKey(key);
		// this.time2 += (System.currentTimeMillis() - tmp2);
		final int noEdges;
		if ( (noEdges = this.noOfEdgesBetweenTwoBlocks.get(key)) != -1 )
//		if ( containsKey )
		{
			this.noOfLookUps++;
			// this.timeforNoOfEdges += (System.currentTimeMillis() - tmp);
			return noEdges;
//			return this.noOfEdgesBetweenTwoBlocks.get(key);
		}
		else
		{
			this.noOfComputations++;
			int minNoOfEdges = 0;

			final List<Integer> noOfBadNeighbors = new ArrayList<Integer>(this.graph.getAmountOfNodesWithDegree(degree1));

			final TIntIterator nodeIterator = this.graph.getAllNodesWithDegree(degree1).iterator();
			while (nodeIterator.hasNext())
			{
				final int node1 = nodeIterator.next();
				int noOfAdjacentVertices = 0;

				TIntIterator neighborIterator;
				if ( this.graph.getAdjacentNodes(node1).size() < this.graph.getAmountOfNodesWithDegree(degree2) )
				{
					neighborIterator = this.graph.getAdjacentNodes(node1).iterator();
				}
				else
				{
					neighborIterator = this.graph.getAllNodesWithDegree(degree2).iterator();
				}
				while (neighborIterator.hasNext())
				{
					final int node2 = neighborIterator.next();
					if ( this.graph.getAdjacentNodes(node1).size() < this.graph.getAmountOfNodesWithDegree(degree2)
							&& this.graph.getDegree(node2) == degree2
							|| this.graph.getAdjacentNodes(node1).size() >= this.graph.getAmountOfNodesWithDegree(degree2)
							&& this.graph.isAdjacent(node1, node2) )
					{
						noOfAdjacentVertices++;
					}
				}
				noOfBadNeighbors.add(noOfAdjacentVertices);
			}
			Collections.sort(noOfBadNeighbors);

			if ( this.graph.getAmountOfNodesWithDegree(degree1) < noOfDeg1Vertices )
			{
				System.err.println("ohje; deg: " + degree1 + ", " + this.graph.getAmountOfNodesWithDegree(degree1) + "\tno: " + noOfDeg1Vertices);
				// return 10000;
			}

			final Iterator<Integer> iterator = noOfBadNeighbors.iterator();
			for (int i = 1; i <= noOfDeg1Vertices; i++)
			{
				final int no = iterator.next();
				if ( no > 0 && this.graph.getAmountOfNodesWithDegree(degree2) - noOfDeg2Vertices < no )
				{
					// problems during realization process only occur if a
					// vertex with degree1 is
					// adjacent to more than
					// (this.getAmountOfNodesWithDegree(degree2) - tmp)
					// vertices of degree2
					minNoOfEdges += no - (this.graph.getAmountOfNodesWithDegree(degree2) - noOfDeg2Vertices);
				}
			}
			// if (minNoOfEdges > 0)
			// {
			// System.out.println("minNoOfEdges: " + minNoOfEdges
			// + "\tnodes1: " + this.getAllNodesWithDegree(degree1)
			// + "\tnodes2: " + this.getAllNodesWithDegree(degree2)
			// + "\tkey: " + key);
			// }
			this.noOfEdgesBetweenTwoBlocks.put(key, minNoOfEdges);
			// this.timeforNoOfEdgesComputingOnly += (System.currentTimeMillis()
			// - tmp);
			// this.timeforNoOfEdges += (System.currentTimeMillis() - tmp);
			return minNoOfEdges;
		}
	}

	public long getEdgesBetweenCalls()
	{
		return this.noOfComputations + this.noOfLookUps + this.noOfCountingArgumentSufficient;
	}

	public int getHashMapSize()
	{
		return this.noOfEdgesBetweenTwoBlocks.size();
	}

	@Override
	public long getMaxRFail()
	{
		return this.maxRFail;
	}

	public long getMaxSolDepHashMapSize()
	{
		return this.maxSolDepHashMapSize;
	}

	public long getNoOfComputations()
	{
		return this.noOfComputations;
	}

	public long getNoOfCountingArgumentSufficient()
	{
		return this.noOfCountingArgumentSufficient;
	}

	@Override
	public long getNoOfEGCalls()
	{
		return this.noOfEGCalls;
	}

	public long getNoOfLookUps()
	{
		return this.noOfLookUps;
	}

	public long getNoOfSolDependingComputations()
	{
		return this.noOfSolDependingComputations;
	}

	public long getNoOfSolDependingFullLookUps()
	{
		return this.noOfSolDependingFullLookUps;
	}

	public long getNoOfSolDependingLookUps()
	{
		return this.noOfSolDependingLookUps;
	}

	public long getSolDependingChecks()
	{
		return this.noOfSolDependingFullLookUps + this.noOfSolDependingComputations + this.noOfSolDependingLookUps;
	}

	@Override
	public long getTime1()
	{
		return this.time1;
	}

	@Override
	public long getTime2()
	{
		return this.time2;
	}

	@Override
	public long getTime3()
	{
		return this.time3;
	}

	public long getTimeforNoOfEdges()
	{
		return this.timeforNoOfEdges;
	}

	public long getTimeforNoOfEdgesComputingOnly()
	{
		return this.timeforNoOfEdgesComputingOnly;
	}

	@Override
	public void setGraph(final AdjacentGraph graph, final CompactDegreeSequence init)
	{
		this.graph = graph;
		this.init = init;
		this.noOfEdgesBetweenTwoBlocks.clear();
		this.SolDependingResults.clear();
		// noOfEdges = new HashMap<MyKey,Integer>();
		this.noOfCountingArgumentSufficient = 0;
		this.noOfLookUps = 0;
		this.noOfComputations = 0;
		this.noOfSolDependingFullLookUps = 0;
		this.noOfSolDependingComputations = 0;
		this.noOfSolDependingLookUps = 0;
	}
}
