package de.vibee.kanon.tools;

import java.util.Set;

import seppImpl.CompactDegreeSequence;
import benchmarks.MainTest;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.IntTupel;
import de.vibee.kanon.data.ModifyListEntry;
import de.vibee.kanon.data.Node;

public class GreedyGraphConstructor
{

	private final AdjacentGraph graph;

	public GreedyGraphConstructor(final AdjacentGraph graph)
	{
		this.graph = graph;
	}

	public Set<IntTupel> constructGraph(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq, final boolean simulate,
			final int orderingMethod, final int selectionMethod)
	{

		final ModifyList modifyList = new ModifyList(this.graph);

		if ( selectionMethod == 2 )
		{
			for (int j = 0; j < 500; j++)
			{
				modifyList.computeAndGetModifyList(initialSeq, solutionSeq, selectionMethod);
				Set<IntTupel> backupList = null;
				if ( orderingMethod == 0 )
				{
					modifyList.sortModifyListByDegreeDifference();
					if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
						return backupList;

				}

				// By common Neighbors (Descending)
				if ( orderingMethod == 1 )
				{
					modifyList.sortModifyListByCommonNeighbors();
					if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
						// System.out.println("#Building graph with modifyList sorted descending by amount of common neighbors successful.");
						return backupList;
				}

				// Random
				if ( orderingMethod == 2 )
				{
					for (int i = 0; i < 100; i++)
					{
						modifyList.randomizeModifyListPositions();

						if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
							return backupList;
					}
				}
			}
			return null;
		}

		modifyList.computeAndGetModifyList(initialSeq, solutionSeq, selectionMethod);

		Set<IntTupel> backupList = null;

		// By amount of edges to add (Descending)
		if ( orderingMethod == 0 )
		{
			modifyList.sortModifyListByDegreeDifference();
			if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
				return backupList;
			else
				return null;
		}

		// By common Neighbors (Descending)
		if ( orderingMethod == 1 )
		{
			// System.out.println("yo");
			modifyList.sortModifyListByCommonNeighbors();
			// modifyListComputer.sortModifyListByDegreeDifference();
			if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
				// System.out.println("#Building graph with modifyList sorted descending by amount of common neighbors successful.");
				return backupList;
			else
				return null;
			// System.out.println("#Building graph with modifyList sorted descending by amount of common neighbors NOT successful.");
		}

		// Random
		if ( orderingMethod == 2 )
		{
			for (int i = 0; i < 200; i++)
			{
				modifyList.randomizeModifyListPositions();

				if ( (backupList = new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList)) != null )
					return backupList;
			}
			return null;

		}

		return null;

	}

	public Set<IntTupel> constructGraph(final Node initialDegreeSequence[], final int solutionSequence[], final boolean simulate)
	{

		final ModifyList modifyList = new ModifyList(this.graph);

		int maxDifference = 0;
		int diff;
		for (int i = 1; i < initialDegreeSequence.length; i++)
		{
			final Node n = initialDegreeSequence[i];
			if ( n != null && (diff = solutionSequence[i] - n.getDegree()) != 0 )
			{
				modifyList.add(new ModifyListEntry(n.getPosition(), n.getDegree(), diff, false));
				maxDifference = Math.max(diff, maxDifference);
			}
		}

		if ( maxDifference > modifyList.size() + 1 )
		{
			if ( MainTest.logLevel > 0 )
			{
				System.out.println("Can't construct graph! Can't adjust " + modifyList.size() + " nodes when on node needs to be increased by "
						+ maxDifference);
			}
			return null;
		}

		return new ModifyListTransferer(this.graph).transferModifylistToGraph(simulate, modifyList);

	}

	public Set<IntTupel> constructGraph(final Node initialDegreeSequence[], final Node reducedSolutionSequence[], final boolean simulate)
	{

		final int solutionSequence[] = new int[initialDegreeSequence.length];

		for (int i = 1; i < solutionSequence.length; i++)
		{
			solutionSequence[i] = initialDegreeSequence[i].getDegree();

		}

		for (final Node node : reducedSolutionSequence)
		{
			if ( node != null )
			{
				solutionSequence[node.getSortedPosition()] = node.getDegree();
			}
		}

		return this.constructGraph(initialDegreeSequence, solutionSequence, simulate);

	}

}
