package de.vibee.kanon.tools;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;
import gnu.trove.map.TIntIntMap;
import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import seppImpl.CompactDegreeSequence;
import benchmarks.MainTest;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.IntTupel;
import de.vibee.kanon.data.ModifyListEntry;
import de.vibee.tools.IntTripel;

public class ModifyList implements Cloneable
{

	private List<ModifyListEntry> modifyList;

	AdjacentGraph graph;

	public ModifyList(final AdjacentGraph graph)
	{
		this.graph = graph;
		this.modifyList = null;
	}

	public boolean add(final ModifyListEntry entry)
	{
		this.modifyList.add(entry);
		return true;
	}

	private boolean alreadyContains(final int node)
	{
		for (final ModifyListEntry m : this.modifyList)
		{
			if ( m.getNode() == node )
				return true;
		}
		return false;
	}

	private int amountOfAdjacentVerticesInList(final int node)
	{
		int sum = 0;
		for (final ModifyListEntry mod : this.modifyList)
		{
			if ( mod.getNode() != node && this.graph.isAdjacent(node, mod.getNode()) )
			{
				sum++;
			}
		}
		return sum;
	}

	@Override
	public ModifyList clone()
	{
		final ModifyList clone = new ModifyList(this.graph);
		clone.modifyList = new ArrayList<>();
		for (final ModifyListEntry m : this.modifyList)
		{
			clone.modifyList.add(m.clone());
		}
		return clone;
	}

	public void computeAndGetModifyList(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq, final int selectionMethod)
	{

		final List<IntTripel> degreeRaiseList = new BlockRaiseListComputer().generateBlockRaiseListBottomUp(initialSeq, solutionSeq);

		this.modifyList = new ArrayList<>();
		final Map<Integer, TIntSet> increasedNodes = new HashMap<>(); // solution
																		// degree
																		// mapped
																		// to a
																		// set
																		// of
																		// vertex
																		// indexes
																		// which
																		// shall
																		// be
																		// raise
																		// to
																		// that
																		// solution
																		// degree
		final TIntSet modifyListNodes = new TIntHashSet();

		for (int i = 0; i < degreeRaiseList.size(); i++)
		{ // for each vertex degree, where we have vertices which must be
			// modified...

			final IntTripel entry = degreeRaiseList.get(i);
			final int initDegree = entry.getFirst();
			final int solDegree = entry.getSecond();
			int amountOfNodes = entry.getThird();

			final TIntSet nodesWithThisDegree = this.graph.getAllNodesWithDegree(initDegree);

			if ( nodesWithThisDegree != null && nodesWithThisDegree.size() == amountOfNodes )
			{ // we must raise all the nodes in our graph with degree initDegree
				final TIntIterator nodesIterator = nodesWithThisDegree.iterator();
				while (nodesIterator.hasNext())
				{
					final int node = nodesIterator.next();
					this.modifyList.add(new ModifyListEntry(node, initDegree, solDegree - initDegree, true));
					modifyListNodes.add(node);
					if ( increasedNodes.get(solDegree) == null )
					{
						increasedNodes.put(solDegree, new TIntHashSet());
					}
					increasedNodes.get(solDegree).add(node);
					--amountOfNodes;
				}
			}

			else if ( nodesWithThisDegree == null || amountOfNodes > nodesWithThisDegree.size() )
			{ // there are less nodes with initDegree than we must raise ->
				// re-raise some nodes
				// which we have risen before
				final TIntSet alreadyRisenNodes = increasedNodes.get(initDegree);
				final TIntIterator it = alreadyRisenNodes.iterator();

				// while (alreadyRisenNodes.size() > 0 && 0 < amountOfNodes){
				while (amountOfNodes > 0 && (nodesWithThisDegree == null || nodesWithThisDegree.size() < amountOfNodes))
				{
					if ( !it.hasNext() )
					{
						System.out.println("JHHHH");
					}
					final int node = it.next();
					it.remove();
					if ( increasedNodes.get(solDegree) == null )
					{
						increasedNodes.put(solDegree, new TIntHashSet());
					}
					increasedNodes.get(solDegree).add(node);

					final Iterator<ModifyListEntry> it2 = this.modifyList.iterator();
					while (it2.hasNext())
					{
						final ModifyListEntry modd = it2.next();
						if ( modd.getNode() == node )
						{
							it2.remove();
							this.modifyList.add(new ModifyListEntry(node, initDegree, solDegree - initDegree + modd.getDegreeDifference(), true));

							break;
						}
					}

					--amountOfNodes;
				}
			}

			if ( amountOfNodes > 0 )
			{ // When there are more nodes availible than we need --> here come
				// the strategies

				final List<IntTupel> raiseCandidates = new ArrayList<>(); // first:
																			// node
																			// identifier,
																			// second:
																			// amount
																			// of
																			// common
																			// neighbors
				final TIntIterator nodesIterator = nodesWithThisDegree.iterator();

				if ( selectionMethod == 0 || selectionMethod == 3 )
				{ // sorted ascending by the amount of common neighbors in
					// modifyList

					// Collect information about shared neighbors to nodes that
					// have already been
					// raised.. choose the best one then..
					while (nodesIterator.hasNext())
					{
						final int node = nodesIterator.next();
						final TIntSet neighbors = this.graph.getAdjacentNodes(node);
						int sumOfCommonNeighborsInSelection = 0;
						final TIntIterator neighborIterator = neighbors.iterator();
						while (neighborIterator.hasNext())
						{
							final int neighbor = neighborIterator.next();
							if ( modifyListNodes.contains(neighbor) )
							{
								sumOfCommonNeighborsInSelection++;
							}
						}
						raiseCandidates.add(new IntTupel(node, sumOfCommonNeighborsInSelection));
					}
					Collections.sort(raiseCandidates, new Comparator<IntTupel>() {
						@Override
						public int compare(final IntTupel o1, final IntTupel o2)
						{
							return o1.getSecond() - o2.getSecond();
						}
					});

				}

				else if ( selectionMethod == 1 )
				{ // sorted ascending by neighbor degree sum

					// Collect information about shared neighbors to nodes that
					// have already been
					// raised.. choose the best one then..
					while (nodesIterator.hasNext())
					{
						final int node = nodesIterator.next();
						final TIntSet neighbors = this.graph.getAdjacentNodes(node);
						int sumOfNeighborsDegree = 0;
						final TIntIterator neighborIterator = neighbors.iterator();
						while (neighborIterator.hasNext())
						{
							final int neighbor = neighborIterator.next();
							sumOfNeighborsDegree += this.graph.getAdjacentNodes(neighbor).size();
						}
						raiseCandidates.add(new IntTupel(node, sumOfNeighborsDegree));
					}

					Collections.sort(raiseCandidates, new Comparator<IntTupel>() {
						@Override
						public int compare(final IntTupel o1, final IntTupel o2)
						{
							return o1.getSecond() - o2.getSecond();
						}
					});

				}
				if ( selectionMethod == 0 || selectionMethod == 1 || selectionMethod == 3 )
				{
					for (int x = 0; x < amountOfNodes; x++)
					{
						final IntTupel raiseThisOne = raiseCandidates.get(x);
						modifyListNodes.add(raiseThisOne.getFirst());
						this.modifyList.add(new ModifyListEntry(raiseThisOne.getFirst(), initDegree, solDegree - initDegree, false));
						if ( increasedNodes.get(solDegree) == null )
						{
							increasedNodes.put(solDegree, new TIntHashSet());
						}
						increasedNodes.get(solDegree).add(raiseThisOne.getFirst());
					}
				}

				if ( selectionMethod == 2 )
				{ // random method

					final TIntSet selectionSet = new TIntHashSet();
					final TIntList nodesWithThisDegreeList = new TIntArrayList(nodesWithThisDegree);

					while (amountOfNodes != selectionSet.size())
					{
						final int pos = new Random().nextInt(nodesWithThisDegree.size());
						selectionSet.add(pos);
					}
					final TIntIterator it = selectionSet.iterator();

					while (it.hasNext())
					{
						final int node = nodesWithThisDegreeList.get(it.next());
						modifyListNodes.add(node);
						this.modifyList.add(new ModifyListEntry(node, initDegree, solDegree - initDegree, false));
						if ( increasedNodes.get(solDegree) == null )
						{
							increasedNodes.put(solDegree, new TIntHashSet());
						}
						increasedNodes.get(solDegree).add(node);
					}
				}
			}

		}

		if ( selectionMethod == 3 )
		{ // we improve the computed solution!
			this.improveSolutionByCommonNeighborsSort();
		}

	}

	public ModifyListEntry get(final int i)
	{
		if ( this.modifyList != null && this.modifyList.size() > i )
			return this.modifyList.get(i);
		else
			return null;
	}

	private void improveSolutionByCommonNeighborsSort()
	{
		boolean doItAgainSam = true;
		while (doItAgainSam)
		{
			doItAgainSam = false;
			for (int i = this.modifyList.size() - 1; i >= 0; i--)
			{
				if ( !this.modifyList.get(i).isFixed() )
				{
					final int node = this.modifyList.get(i).getNode();
					int actualSum = this.amountOfAdjacentVerticesInList(node);
					final TIntSet nodesWithThisDegree = this.graph.getAllNodesWithDegree(this.graph.getDegree(node));

					final TIntIterator nodesIterator = nodesWithThisDegree.iterator();
					while (nodesIterator.hasNext() && actualSum > 0)
					{
						final int newNode = nodesIterator.next();
						if ( newNode != node && !this.alreadyContains(newNode) )
						{
							final int newSum = this.amountOfAdjacentVerticesInList(newNode);
							if ( newSum < actualSum )
							{
								if ( MainTest.logLevel > 0 )
								{
									System.out.println("Swapped node #" + node + ",amount of common neighbors: " + actualSum + " with node #"
											+ newNode + ",amount of common neighbors: " + newSum);
								}
								doItAgainSam = true;
								actualSum = newSum;
								final int diff = this.modifyList.get(i).getDegreeDifference();
								this.modifyList.remove(i);
								this.modifyList.add(i, new ModifyListEntry(newNode, this.graph.getDegree(node), diff, false));

							}
						}

					}
				}

			}
		}
	}

	public Iterator<ModifyListEntry> iterator()
	{
		return this.modifyList.iterator();
	}

	public void randomizeModifyListPositions()
	{

		final Random posRandom = new Random();
		int rand;
		final ModifyListEntry randomArray[] = new ModifyListEntry[this.modifyList.size()];
		// Arrays.fill(randomArray, -1);
		for (int i = 0; i < randomArray.length; i++)
		{
			while (randomArray[rand = posRandom.nextInt(randomArray.length)] != null)
			{
				;
			}
			randomArray[rand] = this.modifyList.get(i);
		}
		this.modifyList.clear();
		for (final ModifyListEntry t : randomArray)
		{
			this.modifyList.add(t);
		}
	}

	public int size()
	{
		if ( this.modifyList != null )
			return this.modifyList.size();
		else
			return 0;
	}

	public void sortModifyListByCommonNeighbors()
	{
		final TIntIntMap neighborList = new TIntIntHashMap();

		for (final ModifyListEntry modElement : this.modifyList)
		{
			final int node1 = modElement.getNode();
			int sumOfNeigborsInList = 0;
			for (final ModifyListEntry modElement2 : this.modifyList)
			{
				final int node2 = modElement2.getNode();
				if ( node1 != node2 )
				{
					if ( this.graph.isAdjacent(node1, node2) )
					{
						sumOfNeigborsInList++;
					}
				}
			}
			neighborList.put(node1, sumOfNeigborsInList);
		}

		Collections.sort(this.modifyList, new Comparator<ModifyListEntry>() {

			@Override
			public int compare(final ModifyListEntry o1, final ModifyListEntry o2)
			{
				if ( neighborList.get(o2.getNode()) == neighborList.get(o1.getNode()) )
					// sysout
					return o2.getDegreeDifference() - o1.getDegreeDifference();
				return neighborList.get(o2.getNode()) - neighborList.get(o1.getNode());
			}

		});

	}

	public void sortModifyListByDegreeDifference()
	{

		Collections.sort(this.modifyList, new Comparator<ModifyListEntry>() {

			@Override
			public int compare(final ModifyListEntry o1, final ModifyListEntry o2)
			{
				if ( o2.getDegreeDifference() != o1.getDegreeDifference() )
					return o2.getDegreeDifference() - o1.getDegreeDifference();
				else
					return ModifyList.this.graph.getDegree(o2.getNode()) - ModifyList.this.graph.getDegree(o1.getNode());
			}

		});

	}

	/**
	 * Calculates the total amount of degree increases (for debugging purposes)
	 * 
	 * @param modifyList
	 * @return
	 */
	private int sumOfAllRaises(final List<ModifyListEntry> modifyList)
	{
		int sum = 0;
		for (final ModifyListEntry mod : modifyList)
		{
			sum += mod.getDegreeDifference();
		}
		return sum;
	}

}
