package de.vibee.kanon.tools;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import benchmarks.MainTest;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.IntTupel;
import de.vibee.kanon.data.ModifyListEntry;

public class ModifyListTransferer
{

	private final AdjacentGraph graph;

	public ModifyListTransferer(final AdjacentGraph graph)
	{
		this.graph = graph;
	}

	/**
	 * Checks whether all nodes have the solution degree and the gaph construction is done.
	 * 
	 * @param modifyList List of nodes and the amount of degree raises that remains.
	 * @return
	 */
	private boolean finished(final ModifyList modifyList)
	{
		final Iterator<ModifyListEntry> it = modifyList.iterator();
		while (it.hasNext())
		{
			final ModifyListEntry m = it.next();
			if ( m.getDegreeDifference() > 0 )
				return false;
		}
		return true;
	}

	/**
	 * For each pair of vertices that could not be connected by an edge, this method tries to find a
	 * different solution by a local search algorithm. It removes added edges and tries to connect
	 * the endpoints of this edge to other vertices.
	 * 
	 * @param backupList List of already added edges. Each entry consists of the the node indices of
	 *            both nodes.
	 * @param modifyList List of vertices whose degree has to be raised.
	 * @param changedEdges Stores all added egdes, so that the indices match to the matrix for quick
	 *            access.
	 * @param matrix Data structure which stores IntTupel for each combination of nodes which are
	 *            affected.
	 */
	private void reconstructGraph(final ModifyList modifyList, final Set<IntTupel> changedEdges, final IntTupel[][] matrix)
	{
		if ( !this.finished(modifyList) )
		{
			for (int i = 0; i < modifyList.size(); i++)
			{
				final ModifyListEntry m = modifyList.get(i);
				if ( m.getDegreeDifference() > 0 )
				{
					for (int j = i + 1; j < modifyList.size(); j++)
					{
						final ModifyListEntry n = modifyList.get(j);
						if ( m.getDegreeDifference() > 0 && n.getDegreeDifference() > 0 )
						{
							final IntTupel mnTupel = matrix[i][j];
							if ( !changedEdges.contains(mnTupel) && this.graph.isAdjacent(m.getNode(), n.getNode()) )
							{

								for (int x = 0; x < modifyList.size(); x++)
								{
									if ( x != i && x != j )
									{
										final ModifyListEntry s = modifyList.get(x);

										for (int y = x + 1; y < modifyList.size(); y++)
										{
											if ( y != i && y != j && y != x && m.getDegreeDifference() > 0 && n.getDegreeDifference() > 0 )
											{
												final ModifyListEntry t = modifyList.get(y);
												final IntTupel stTupel = matrix[x][y];
												if ( changedEdges.contains(stTupel) )
												{

													if ( !this.graph.isAdjacent(m.getNode(), s.getNode())
															&& !this.graph.isAdjacent(n.getNode(), t.getNode()) )
													{

														changedEdges.remove(stTupel);
														changedEdges.add(matrix[i][x]);
														changedEdges.add(matrix[j][y]);

														this.graph.removeEdge(s.getNode(), t.getNode());
														this.graph.addEdge(m.getNode(), s.getNode());
														this.graph.addEdge(n.getNode(), t.getNode());
														IntTupel theonetodelete = null;
														for (final IntTupel backup : changedEdges)
														{
															if ( backup.getFirst() == s.getNode() && backup.getSecond() == t.getNode()
																	|| backup.getFirst() == t.getNode() && backup.getSecond() == s.getNode() )
															{
																theonetodelete = backup;
																break;
															}
														}

														changedEdges.remove(theonetodelete);

														m.reduceDegreeDifference(1);
														n.reduceDegreeDifference(1);
														changedEdges.add(new IntTupel(m.getNode(), s.getNode()));
														changedEdges.add(new IntTupel(n.getNode(), t.getNode()));
														this.reconstructGraph(modifyList, changedEdges, matrix);
														return;
													}
													else if ( !this.graph.isAdjacent(m.getNode(), t.getNode())
															&& !this.graph.isAdjacent(n.getNode(), s.getNode()) )
													{

														changedEdges.remove(stTupel);
														changedEdges.add(matrix[i][y]);
														changedEdges.add(matrix[j][x]);
														this.graph.removeEdge(s.getNode(), t.getNode());
														IntTupel theonetodelete = null;
														for (final IntTupel backup : changedEdges)
														{
															if ( backup.getFirst() == s.getNode() && backup.getSecond() == t.getNode()
																	|| backup.getFirst() == t.getNode() && backup.getSecond() == s.getNode() )
															{
																theonetodelete = backup;
																break;
															}
														}
														changedEdges.remove(theonetodelete);
														this.graph.addEdge(m.getNode(), t.getNode());
														this.graph.addEdge(n.getNode(), s.getNode());
														m.reduceDegreeDifference(1);
														n.reduceDegreeDifference(1);
														changedEdges.add(new IntTupel(m.getNode(), t.getNode()));
														changedEdges.add(new IntTupel(n.getNode(), s.getNode()));
														this.reconstructGraph(modifyList, changedEdges, matrix);
														return;
													}
												}

											}
										}
									}

								}

							}
						}

					}
				}

			}
		}
	}

	/**
	 * Performs the graph construction, i.e. adds edges and reconnects edges if needed.
	 * 
	 * @param simulate If true, all changes are undoed after graph constructing.
	 * @param modifyList List of all vertices whose degree must be raised. The first value of each
	 *            IntTupel is the node, the second value is the difference between original degree
	 *            and solution degree.
	 * @return A List of all edges, that have been added.
	 */
	protected Set<IntTupel> transferModifylistToGraph(final boolean simulate, ModifyList modifyList)
	{

		// List<Edge> addedEdges = new ArrayList<>();

		modifyList = modifyList.clone(); // So that we don't touch the original
											// modify list, we are
											// going to need it to be unmodified
											// in random mode for
											// the next run

		final Set<IntTupel> changedEdges = new HashSet<>();

		final IntTupel matrix[][] = new IntTupel[modifyList.size()][modifyList.size()];
		for (int i = 0; i < matrix.length; i++)
		{
			for (int j = i + 1; j < matrix[i].length; j++)
			{
				final IntTupel ijtupel = new IntTupel(modifyList.get(i).getNode(), modifyList.get(j).getNode());
				matrix[i][j] = ijtupel;
				matrix[j][i] = ijtupel;
			}
		}

		for (int i = 0; i < modifyList.size(); i++)
		{
			for (int j = i + 1; j < modifyList.size(); j++)
			{
				final ModifyListEntry n = modifyList.get(i);
				final ModifyListEntry m = modifyList.get(j);
				if ( n.getDegreeDifference() > 0 && n != m && m.getDegreeDifference() > 0 && !this.graph.isAdjacent(n.getNode(), m.getNode()) )
				{
					this.graph.addEdge(n.getNode(), m.getNode());
					n.reduceDegreeDifference(1);
					m.reduceDegreeDifference(1);
					changedEdges.add(matrix[i][j]);
				}
			}
		}

		this.reconstructGraph(modifyList, changedEdges, matrix);

		boolean alreadyUndone = false;
		boolean failed = false;

		// if (!finished(modifyList)){
		// for (int i = 0; i < modifyList.size(); i++){
		// ModifyListEntry n = modifyList.get(i);
		// if (n.getDegreeDifference() != 0){
		//
		// }
		// }
		// }

		for (int i = 0; i < modifyList.size(); i++)
		{
			final ModifyListEntry n = modifyList.get(i);
			if ( n.getDegreeDifference() != 0 )
			{
				if ( !alreadyUndone )
				{
					this.undoAddEdges(changedEdges);
					alreadyUndone = true;
				}
				failed = true;
				if ( MainTest.logLevel > 3 )
				{
					System.out.println("Edge insertion failure: Node " + n.getNode() + " requires to be connected to " + n.getDegreeDifference()
							+ " more edges, which is not possible. Original degree: " + this.graph.getDegree(n.getNode()) + ", Solution degree: "
							+ (this.graph.getDegree(n.getNode()) + modifyList.get(i).getDegreeDifference()));
				}
			}
		}

		if ( !failed && MainTest.logLevel > 0 )
		{

			System.out.println("Anonymity: " + this.graph.getAnonymity() + ", cost: " + changedEdges.size());
		}

		if ( failed )
			return null;
		else
		{
			if ( MainTest.logLevel > 0 )
			{
				System.out.println("Graph Construction " + (simulate ? "(Simulation)" : "") + " was succesful!");
			}
			// System.out.println("Graph is " + graph.getAnonymity() +
			// "-anonymous by adding " +
			// addedEdges.size() + " edges.");
			if ( simulate )
			{
				this.undoAddEdges(changedEdges);
			}
			return changedEdges;
		}

	}

	/**
	 * Based on the input edge list, all changes are undone, so that each edge that has been added
	 * while graph constructing is removed again.
	 * 
	 * @param backupList List of edges that has to be removed
	 * @return True, if all edges could be removed.
	 */
	private boolean undoAddEdges(final Set<IntTupel> backupList)
	{
		boolean successful = true;
		for (final IntTupel t : backupList)
		{
			successful &= this.graph.removeEdge(t.getFirst(), t.getSecond());
		}
		return successful;
	}

}
