package de.vibee.kanon.tools;

import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.collections15.map.HashedMap;

import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.IntTupel;

public class Reader
{

	File file;

	InputStream is = null;

	private String filename;

	public Reader(final InputStream is)
	{
		this.is = is;
	}

	public Reader(final String filename)
	{
		this.file = new File(filename);
	}

	public Reader(final URL filestream)
	{
		this.file = new File(filestream.getFile());
	}

	public String getFilename()
	{
		return this.filename;
	}

	private AdjacentGraph readInputGraph()
	{
//		System.out.println("#Reading input file in dimacs way");
		List<TIntSet> tGraphSet = null;

		BufferedReader in = null;
		try
		{
			if ( this.is == null )
			{
				in = new BufferedReader(new FileReader(this.file));
			}
			else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			// Include header information (|V|, |E|)
			String zeile = null;
			do
			{
				zeile = in.readLine();
				if ( zeile == null )
				{
					System.out.println("Error processing file. Exiting");
					System.exit(-1);
				}
			} while (!Pattern.matches("\\d+\\s+\\d+", zeile.trim()));

			final String header[] = zeile.split(" ");
			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(header[0]));
			tGraphSet.add(null);

			while ((zeile = in.readLine()) != null)
			{
				zeile = zeile.trim();
				if ( Pattern.matches("[\\d\\s]*", zeile) )
				{
					final TIntHashSet tNodeSet = new TIntHashSet();
					if ( !zeile.isEmpty() )
					{
						zeile = zeile.trim();
						final String nodes[] = zeile.split("\\s+");
						for (final String node : nodes)
						{
							final Integer nodeInteger = Integer.valueOf(node);
							if ( nodeInteger != tGraphSet.size() )
							{ // Exclude reflexive edges
								tNodeSet.add(Integer.valueOf(node));
							}
						}
					}
					tGraphSet.add(tNodeSet);
				}

			}
		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return new AdjacentGraph(tGraphSet);
	}

	
	private AdjacentGraph readTripleFormat(){
		List<TIntSet> tGraphSet = null;
		final List<IntTupel> edges = new LinkedList<>();
		BufferedReader in = null;
		Map<Integer,String> indexToName = new HashedMap<Integer, String>();
		try
		{
			if ( this.is == null )
			{
				in = new BufferedReader(new FileReader(this.file));
			}
			else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			String zeile = null;

			int maxVerticeNumber = -1;
			Map<String,Integer> nameToIndex = new HashedMap<String,Integer>();
//			Map<Integer,String> indexToName = new HashedMap<>();
			int index = 1;
			while ((zeile = in.readLine()) != null)
			{
				if ( Pattern.matches("\\d+\\s+\\d+\\s+-*\\d+", zeile) )
				{
					final String vertices[] = zeile.split("\\s");
					int a;
					int b;
					if (nameToIndex.containsKey(vertices[0])){
						a = nameToIndex.get(vertices[0]);
					}
					else{
						a = index++;
						nameToIndex.put(vertices[0], a);
						indexToName.put(a, vertices[0]);
					}
					if (nameToIndex.containsKey(vertices[1])){
						b = nameToIndex.get(vertices[1]);
					}
					else{
						b = index++;
						nameToIndex.put(vertices[1], b);
						indexToName.put(b, vertices[1]);
					}
					
					if ( a != b )
					{
						edges.add(new IntTupel(a, b));
						maxVerticeNumber = Math.max(maxVerticeNumber, a);
						maxVerticeNumber = Math.max(maxVerticeNumber, b);
					}

				}
				else if ( !Pattern.matches("%|#.*", zeile) )
					return null;

			}

			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(maxVerticeNumber + 1));
			tGraphSet.add(null);
			for (int i = 1; i <= maxVerticeNumber; i++)
			{
				tGraphSet.add(new TIntHashSet());
			}

			for (final IntTupel e : edges)
			{

				tGraphSet.get(e.getFirst()).add(e.getSecond());

				tGraphSet.get(e.getSecond()).add(e.getFirst());
			}

		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return new AdjacentGraph(tGraphSet, indexToName);
	}
	
	public AdjacentGraph readInputGraphAlternateFormat()
	{
		AdjacentGraph tripleFormatGraph;
		if ((tripleFormatGraph = readTripleFormat()) != null) return tripleFormatGraph;
		
		List<TIntSet> tGraphSet = null;
		final List<IntTupel> edges = new LinkedList<>();
		BufferedReader in = null;
		Map<Integer,String> indexToName = new HashedMap<Integer, String>();
		try
		{
			if ( this.is == null )
			{
				in = new BufferedReader(new FileReader(this.file));
			}
			else
			{
				in = new BufferedReader(new InputStreamReader(this.is));
			}

			String zeile = null;

			int maxVerticeNumber = -1;
			Map<String,Integer> nameToIndex = new HashedMap<String,Integer>();
//			Map<Integer,String> indexToName = new HashedMap<>();
			int index = 1;
			while ((zeile = in.readLine()) != null)
			{
				if ( Pattern.matches("\\d+\\s+\\d+", zeile) )
				{
					final String vertices[] = zeile.split("\\s");
					int a;
					int b;
					if (nameToIndex.containsKey(vertices[0])){
						a = nameToIndex.get(vertices[0]);
					}
					else{
						a = index++;
						nameToIndex.put(vertices[0], a);
						indexToName.put(a, vertices[0]);
					}
					if (nameToIndex.containsKey(vertices[1])){
						b = nameToIndex.get(vertices[1]);
					}
					else{
						b = index++;
						nameToIndex.put(vertices[1], b);
						indexToName.put(b, vertices[1]);
					}
					
					if ( a != b )
					{
						edges.add(new IntTupel(a, b));
						maxVerticeNumber = Math.max(maxVerticeNumber, a);
						maxVerticeNumber = Math.max(maxVerticeNumber, b);
					}

				}
				else if ( !Pattern.matches("%|#.*", zeile) )
					return this.readInputGraph();

			}

			tGraphSet = new ArrayList<TIntSet>(Integer.valueOf(maxVerticeNumber + 1));
			tGraphSet.add(null);
			for (int i = 1; i <= maxVerticeNumber; i++)
			{
				tGraphSet.add(new TIntHashSet());
			}

			for (final IntTupel e : edges)
			{

				tGraphSet.get(e.getFirst()).add(e.getSecond());

				tGraphSet.get(e.getSecond()).add(e.getFirst());
			}

		} catch (final Exception e)
		{
			System.out.println("#Error occured while reading input graph.");
			e.printStackTrace();
			return null;
		} finally
		{
			try
			{
				in.close();
			} catch (final IOException e)
			{
				e.printStackTrace();
			}
		}

		return new AdjacentGraph(tGraphSet, indexToName);
	}

}
