package de.vibee.test;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;

import seppImpl.CompactDegreeSequence;
import seppImpl.DegreeAnonymizer2;
import agape.generators.RandGenerator;
import agape.tutorials.UndirectedGraphFactoryForStringInteger;
import benchmarks.MainTest;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.tools.ErdosGallaiAdvanced;
import de.vibee.kanon.tools.ErdosGallaiInterface;
import de.vibee.kanon.tools.FastErdosGallai;
import de.vibee.kanon.tools.GreedyGraphConstructor;
import edu.uci.ics.jung.graph.Graph;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

public class ErdosGallaiTest
{

	@Test
	public void testErdosGallaiOnBarabasiAlbertGraphs() throws Exception
	{
		MainTest.logLevel = 0;
		int counterAll = 0, errorCounter = 0;
		long time1 = 0, time2 = 0, tmpTime;
		final long tmpTime2 = System.currentTimeMillis();

		for (int n = 4; n < 400; n += 2)
		{
			if ( n % 10 == 0 )
			{
				System.out.println("n: " + n + "\tEG-time: " + time1 / 1000d + "\telapsed time: " + (System.currentTimeMillis() - tmpTime2) / 1000d);
			}
			for (int t = 0; t < 50; t++)
			{
				counterAll++;

				final int steps = (int) Math.ceil(Math.random() * 20) + 5;
				final int m = (int) (Math.random() * 0.9 * n);

				final UndirectedGraphFactoryForStringInteger factory = new UndirectedGraphFactoryForStringInteger();
				final Graph<String, Integer> g5 = RandGenerator.generateBarabasiAlbertGraph(factory, factory.vertexFactory, factory.edgeFactory, n,
						m, steps);

				final HashMap<String, Integer> vertexNames = new HashMap<String, Integer>();
				final List<TIntSet> adjacentList = new ArrayList<>();
				final List<TIntSet> adjacentList2 = new ArrayList<>();

				int index = 1;
				adjacentList.add(null);
				adjacentList2.add(null);
				for (final String v : g5.getVertices())
				{
					vertexNames.put(v, index);
					adjacentList.add(new TIntHashSet());
					adjacentList2.add(new TIntHashSet());
					index++;
				}

				for (final String v : g5.getVertices())
				{
					for (final String v2 : g5.getNeighbors(v))
					{
						adjacentList.get(vertexNames.get(v)).add(vertexNames.get(v2));
						adjacentList.get(vertexNames.get(v2)).add(vertexNames.get(v));
						adjacentList2.get(vertexNames.get(v)).add(vertexNames.get(v2));
						adjacentList2.get(vertexNames.get(v2)).add(vertexNames.get(v));
					}
				}
				final AdjacentGraph graph = new AdjacentGraph(adjacentList);
				final CompactDegreeSequence initialDegreeSequence = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));

				final int maxTests = 5;
				for (int j = 1; j < maxTests; j++)
				{
					if ( j == maxTests - 1 )
					{
						for (int i = 1; i <= g5.getVertexCount(); i++)
						{
							for (int x = i + 1; x <= g5.getVertexCount(); x++)
							{
								adjacentList2.get(i).add(x);
								adjacentList2.get(x).add(i);
							}
						}
					}
					else if ( j == 1 )
					{
						int pos1 = (int) (Math.random() * (n + steps)) + 1;
						int pos2 = (int) (Math.random() * (n + steps)) + 1;
						if ( pos1 > pos2 )
						{
							final int tmp = pos1;
							pos1 = pos2;
							pos2 = tmp;
						}
						for (int i = pos1; i <= pos2; i++)
						{
							for (int x = i + 1; x <= pos2; x++)
							{
								adjacentList2.get(i).add(x);
								adjacentList2.get(x).add(i);
							}
						}
					}
					else
					{
						int noOfAddedEdges = (int) (Math.random() * n) + steps;
						while (noOfAddedEdges > 0)
						{
							final int v1 = (int) (Math.random() * graph.getAmountOfVertices()) + 1;
							final int v2 = (int) (Math.random() * graph.getAmountOfVertices()) + 1;
							if ( v1 != v2 )
							{
								adjacentList2.get(v1).add(v2);
								adjacentList2.get(v2).add(v1);
							}
							noOfAddedEdges--;
						}
					}

					final AdjacentGraph graph2 = new AdjacentGraph(adjacentList2);
					final CompactDegreeSequence finalDegreeSequence = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph2));
					tmpTime = System.currentTimeMillis();
					final ErdosGallaiInterface egInstance1 = new FastErdosGallai(graph, initialDegreeSequence);
					final int missingEdges = egInstance1.computeFeasibility(finalDegreeSequence, true);
					time1 += System.currentTimeMillis() - tmpTime;
					tmpTime = System.currentTimeMillis();
					final ErdosGallaiInterface egInstance2 = new ErdosGallaiAdvanced(graph, initialDegreeSequence);
					final int missingEdges2 = egInstance2.computeFeasibility(finalDegreeSequence, true);

					time2 += System.currentTimeMillis() - tmpTime;
					if ( missingEdges > 0 || missingEdges2 > 0 )
					{
						errorCounter++;
						System.out.println("André: " + missingEdges + " \t Clemens: " + missingEdges2);
						System.out.println(initialDegreeSequence);
						System.out.println(finalDegreeSequence);
						for (int i = 1; i < graph.getAmountOfVertices() + 1; i++)
						{
							final TIntSet set = adjacentList.get(i);
							System.out.println(i + ": \t" + set);
						}
						System.out.println("===============================================================================");
					}
					Assert.assertTrue(missingEdges == 0);
				}
			}
		}
		System.out.println("Time1: " + time1 / 1000d + "\tTime2: " + time2 / 1000d + "\t\t\tInstances tested: " + counterAll
				+ "\t\tInstances failed: " + errorCounter);
	}

	@Test
	public void testIfErdosIsOkWhenBuildingIsOK()
	{
		MainTest.logLevel = 0;
		int sizeOfGraph = 3;
		final int randomIterations = 20000;

		final int anonymityIncrease = 2;

		int count = 0;
		while (count++ < randomIterations)
		{
			final Random boolRand = new Random();
			final List<TIntSet> adjacentList = new ArrayList<>();
			adjacentList.add(null);
			for (int i = 1; i <= sizeOfGraph; i++)
			{
				adjacentList.add(new TIntHashSet());
			}

			for (int i = 1; i <= sizeOfGraph; i++)
			{
				// if (i == 0) adjacentList.add(null);

				// adjacentList.add(new TIntHashSet());
				for (int j = 1; j <= sizeOfGraph; j++)
				{
					if ( i != j )
					{
						if ( boolRand.nextBoolean() )
						{
							adjacentList.get(i).add(j);
							adjacentList.get(j).add(i);
						}
					}
				}

			}

			// for (int i = 1; i <= 10; i++){
			// TIntSet set = adjacentList.get(i);
			// System.out.println(set);
			// }

			// Reader r = new Reader("/home/clemens/Desktop/test.graph");
			//
			// AdjacentGraph graph = r.readInputGraph();
			final AdjacentGraph graph = new AdjacentGraph(adjacentList);
			final int initialAnonymity = graph.getAnonymity();
			final CompactDegreeSequence init = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));
			// CompactDegreeSequence init = new CompactDegreeSequence(new
			// int[]{0,0,0,1,0,1,4,3,0,1});

			final Set<CompactDegreeSequence> solutions = new DegreeAnonymizer2(graph, init, Math.min(initialAnonymity + anonymityIncrease,
					graph.getAmountOfVertices())).getAllkAnonRealSequences();

			for (final CompactDegreeSequence sol : solutions)
			{
				// int erdosGallaiAdvanced = new
				// ErdosGallaiAdvanced().computeFeasibility(graph, init,
				// sol,true);
				final int fastErdosGallai = new FastErdosGallai(graph, init).computeFeasibility(sol, true);
				// if (erdosGallaiAdvanced != fastErdosGallai)
				// {
				// System.out.println("A: " + fastErdosGallai + " vs. C: " +
				// erdosGallaiAdvanced);
				// System.out.println("init: " + init);
				// System.out.println("sol  :" + sol);
				// System.out.println("diff  :" + sol.difference(init));
				// System.out.println("Graph: (size: "+(graph.getSize()-1)+")");
				// for (int i = 1; i <= sizeOfGraph; i++)
				// {
				// final TIntSet set = adjacentList.get(i);
				// System.out.println(i+": "+set);
				// }
				// }
				final boolean bFastErdosGallai = fastErdosGallai == 0;
				final boolean bErdosGallaiAdvanced = true; // (erdosGallaiAdvanced
															// == 0);
				// final boolean erdosGallai =
				// (ErdosGallai.erdosGallaiAdvanced(graph, init, sol, true) ==
				// 0);
				final boolean graphBuildOK = new GreedyGraphConstructor(graph).constructGraph(init, sol, true, 0, 0) != null;
				if ( graphBuildOK && !bFastErdosGallai && !bErdosGallaiAdvanced )
				{
					System.out.println("Test fails");
					System.out.println("init: " + init);
					System.out.println("sol  :" + sol);
					System.out.println("Graph:");
					for (int i = 1; i <= sizeOfGraph; i++)
					{
						final TIntSet set = adjacentList.get(i);
						System.out.println(set);
					}
				}
				// System.out.println("graph is " + graph.getAnonymity());
				// System.out.println("...");

				Assert.assertFalse(graphBuildOK && !bFastErdosGallai && !bErdosGallaiAdvanced);
			}
			if ( count % (randomIterations / 4) == 0 )
			{
				System.out.println("Durchlauf nr: " + count);
			}
			if ( count % (randomIterations / 7) == 0 )
			{
				sizeOfGraph++;
			}

			// if (Math.random() > 0.2)
			// {
			// sizeOfGraph++;
			// }

		}

	}

}
