package imp;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.LineNumberReader;
import java.util.HashMap;

import org.jgrapht.graph.SimpleGraph;

import dataStructures.graph.Edge;
import dataStructures.graph.ExtendedSimpleGraphAjazenzList;
import dataStructures.graph.ExtendedUndirectedGraph;

/**
 * 
 * This class provides an importer, which converts a given DIMACS-graph-file into a
 * {@link SimpleGraph}. Actually, this For more information about the DIMACS-format:
 * ftp://dimacs.rutgers.edu/pub/challenge/graph/doc/ccformat.tex
 * 
 * @author SeppHar
 * 
 */
public abstract class DIMACS_GraphImporter
{

	public static ExtendedUndirectedGraph<Integer, Edge<Integer>> readDIMACS(String filePath) throws IOException
	{
		final LineNumberReader reader = new LineNumberReader(new FileReader(new File(filePath)));
		String line;
		final ExtendedSimpleGraphAjazenzList<Integer> graph = new ExtendedSimpleGraphAjazenzList<Integer>();

		final HashMap<Integer, Integer> vertexMap = new HashMap<Integer, Integer>();

		while ((line = reader.readLine()) != null)
		{
			// ignore comment lines
			if ( !line.startsWith("c ") )
			{
				// ignore problem lines and node description (only required for weighted clique
				// problem)
				if ( !line.startsWith("p ") && !line.startsWith("n ") )
				{
					if ( line.startsWith("e ") )
					{
						final String[] splits = line.split("[ ]+");
						if ( splits.length == 3 )
						{
							final Integer v1 = Integer.parseInt(splits[1]);
							final Integer v2 = Integer.parseInt(splits[2]);
							if ( !graph.containsVertex(v1) )
							{
								graph.addVertex(v1);
								vertexMap.put(v1, v1);
							}
							if ( !graph.containsVertex(v2) )
							{
								graph.addVertex(v2);
								vertexMap.put(v2, v2);
							}
							graph.addEdge(vertexMap.get(v1), vertexMap.get(v2));
						}
					}
				}
			}
		}
		reader.close();
		return graph;
	}
}
