package seppImpl;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import benchmarks.MainTest;
import dataStructures.Store;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.data.DegreesChangeList;
import de.vibee.kanon.data.Node;
import de.vibee.tools.IntTripel;

public class CompactDegreeSequence implements Serializable
{

	private static class DegreeChange
	{
		private boolean allVerticesFixed = false;

		private final int degreeIncrease;

		private final Set<IntTripel> s = new HashSet<IntTripel>();

		public DegreeChange(final int deg)
		{
			this.degreeIncrease = deg;
		}

		public void add(final IntTripel intTripel)
		{
			this.s.add(intTripel);
		}

		public void checkWhetherAllVerticesFixed(final CompactDegreeSequence origCompactDegreeSequence)
		{
			boolean t = true;
			for (final IntTripel tripel : this.s)
			{
				t = t && tripel.getThird() == origCompactDegreeSequence.get(tripel.getFirst());
				// System.out.println("test:" + tripel + " " +
				// origCompactDegreeSequence.get(tripel.getFirst()));
			}
			if ( t == true && this.s.size() > 0 )
			{
				System.out.println("fixed " + this);
			}
			this.allVerticesFixed = t;
		}

		public Set<IntTripel> getIntTriples()
		{
			return this.s;

		}

		public int getMaximumIntersectionWithFixedVertices(final AdjacentGraph graph, final Set<Integer> fixedVertices)
		{
			return MainTest.sum(this.getMinimumNumberOfEdgesWithFixedVertices(graph, fixedVertices));
		}

		public int getMinimumIntersectionWithFixedVertices(final AdjacentGraph graph, final Set<Integer> fixedVertices)
		{
			if ( this.isFixed() )
				return this.getMaximumIntersectionWithFixedVertices(graph, fixedVertices);
			int min = Integer.MAX_VALUE;
			for (final int a : this.getMinimumNumberOfEdgesWithFixedVertices(graph, fixedVertices))
			{
				min = Math.min(min, a);
			}
			return min;
		}

		private int[] getMinimumNumberOfEdgesWithFixedVertices(final AdjacentGraph graph, final Set<Integer> fixedVertices)
		{
			final int[] arr = new int[this.s.size()];
			Arrays.fill(arr, Integer.MAX_VALUE);

			int i = 0;
			for (final IntTripel tripel : this.s)
			{

				for (final Integer bagVertices : CompactDegreeSequence.getAllVerticesFromBag(graph, tripel.getFirst()))
				{
					int c = 0;
					for (final Integer fixedV : fixedVertices)
					{
						c += graph.isAdjacent(fixedV, bagVertices) ? 1 : 0;
					}
					arr[i] = Math.min(arr[i], c);
					if ( arr[i] == 0 )
					{
						if ( this.degreeIncrease > 1 )
						{
							System.out.println("take " + bagVertices + " for " + tripel + " \t(" + fixedVertices + ")");
						}
						break;
					}
				}
				i++;
			}
			return arr;
		}

		public boolean isFixed()
		{
			return this.allVerticesFixed;
		}

		@Override
		public String toString()
		{
			return "DegreeChange [allVerticesFixed=" + this.allVerticesFixed + ", degreeIncrease=" + this.degreeIncrease + ", s="
					+ Arrays.toString(this.s.toArray()) + "]";
		}

	}

	private static class ladfjl implements Comparator<IntTripel>
	{

		@Override
		public int compare(final IntTripel o1, final IntTripel o2)
		{
			return o2.getThird() - o1.getThird();
		}

	}

	/**
	 * 
	 */
	private static final long serialVersionUID = 8241107201328591868l;

	private static int addDegrees(final AdjacentGraph graph, final Set<Integer> fixedVertices, final int vertex)
	{
		int noOfEdgesBetweenFixedVertices = 0;
		if ( vertex > 0 )
		{
			for (final Integer vert : fixedVertices)
				if ( graph.isAdjacent(vert, vertex) )
				{
					noOfEdgesBetweenFixedVertices += 2;
				}
		}
		return noOfEdgesBetweenFixedVertices;
	}

	private static DegreeChange[] afgetAllChanges(final CompactDegreeSequence orginalSequence, final CompactDegreeSequence anonSequence)
	{
		final DegreeChange[] degreeChanges = new DegreeChange[anonSequence.degrees.length];
		for (int i = 0; i < anonSequence.degrees.length; i++)
		{
			degreeChanges[i] = new DegreeChange(i);
		}

		{
			final int[] newSeq = Arrays.copyOf(anonSequence.degrees, anonSequence.degrees.length);

			for (int i = 0; i < orginalSequence.degrees.length; i++)
			{
				int c = orginalSequence.get(i);
				for (int j = i; j < anonSequence.degrees.length; j++)
				{
					if ( c <= 0 )
					{
						break;
					}
					final int r = Math.min(c, newSeq[j]);
					newSeq[j] -= r;
					c -= r;
					if ( j > i && r > 0 )
					{
						final IntTripel intTripel = new IntTripel(i, j, r);
						System.out.println(intTripel);
						degreeChanges[j - i].add(intTripel);
					}
				}
			}
		}
		for (int i = 0; i < anonSequence.degrees.length; i++)
		{
			degreeChanges[i].checkWhetherAllVerticesFixed(orginalSequence);
		}

		return degreeChanges;
	}

	private static int findFixedVertexInBag(final AdjacentGraph graph, final DegreeChange[] l, final Set<Integer> fixedVertices, final int bagIndex,
			final CompactDegreeSequence orginalSequence, final CompactDegreeSequence anonSequence)
	{
		final DegreeChange degreeChange = l[bagIndex];
		if ( !degreeChange.isFixed() )
			return -1;

		final Set<Integer> s = CompactDegreeSequence.getAllVerticesFromBag(graph, degreeChange);
		int min = Integer.MAX_VALUE;
		int minVert = -1;

		for (final Integer vert : s)
		{
			final int addDegrees = CompactDegreeSequence.addDegrees(graph, fixedVertices, vert);
			if ( addDegrees < min )
			{
				minVert = vert;
				min = addDegrees;
			}
		}
		return minVert;
	}

	private static Set<Integer> getAllVerticesFromBag(final AdjacentGraph graph, final DegreeChange l)
	{
		final SortedSet<Integer> bagIndices = new TreeSet<Integer>();
		for (final IntTripel t : l.getIntTriples())
		{
			bagIndices.add(t.getFirst());
		}

		final HashSet<Integer> s = new HashSet<Integer>();
		for (int j = 1; j < graph.getAmountOfVertices() + 1; j++)
		{
			// if(graph.getDegree(j)>50)
			// System.out.print(j+":"+graph.getDegree(j)+"\t");
			if ( bagIndices.contains(graph.getDegree(j)) )
			{
				s.add(j);
			}
		}
		return s;
	}

	// private static ArrayList<IntTripel> getAllChanges(CompactDegreeSequence
	// orginalSequence,
	// CompactDegreeSequence anonSequence)
	// {
	// final ArrayList<IntTripel> l = new ArrayList<IntTripel>();
	// {
	// final int[] newSeq = Arrays.copyOf(anonSequence.degrees,
	// anonSequence.degrees.length);
	//
	// // final int[] result = new int[this.degrees.length];
	// // Arrays.fill(result, 0);
	//
	// for (int i = 0; i < orginalSequence.degrees.length; i++)
	// {
	// int c = orginalSequence.get(i);
	// for (int j = i; j < anonSequence.degrees.length; j++)
	// {
	// if ( c <= 0 )
	// {
	// break;
	// }
	// final int r = Math.min(c, newSeq[j]);
	// newSeq[j] -= r;
	// c -= r;
	// if ( j > i && r > 0 )
	// {
	// final IntTripel e = new IntTripel(i, j, r);
	// l.add(e);
	// }
	// }
	// }
	// }
	// return l;
	// }
	private static Set<Integer> getAllVerticesFromBag(final AdjacentGraph graph, final int first)
	{
		final HashSet<Integer> s = new HashSet<Integer>();

		for (int j = 1; j < graph.getAmountOfVertices() + 1; j++)
		{
			if ( graph.getDegree(j) == first )
			{
				s.add(j);
			}
		}
		return s;
	}

	public static boolean isExtendedRealizable(final AdjacentGraph graph, final CompactDegreeSequence orginalSequence,
			final CompactDegreeSequence anonSequence)
	{
		System.out.println(orginalSequence);
		System.out.println(anonSequence);
		System.out.println("diff:" + anonSequence.difference(orginalSequence));

		final boolean[] dif = new boolean[orginalSequence.degrees.length];
		for (int i = 0; i < orginalSequence.degrees.length; i++)
		{
			dif[i] = orginalSequence.getNumberOfVerticesInBags(0, i) - anonSequence.getNumberOfVerticesInBags(0, i) >= orginalSequence.get(i);
		}

		final CompactDegreeSequence diffSequence = anonSequence.difference(orginalSequence);
		if ( diffSequence.getTotalDegreeSum() % 2 == 1 )
			return false;

		int entryCounter = 0;
		int degreeCounter = 0;

		final Set<Integer> fixedVertices = new HashSet<Integer>();
		int noOfEdgesBetweenFixedVertices = 0;

		final DegreeChange[] l = CompactDegreeSequence.afgetAllChanges(orginalSequence, anonSequence);

		for (int i = diffSequence.degrees.length - 1; i >= 1; i--)
		{
			if ( diffSequence.degrees[i] > 0 )
			{
				final int r1 = entryCounter + 1;
				final long rightSide1 = diffSequence.getRightSum(i, r1)
						- (noOfEdgesBetweenFixedVertices + l[i].getMinimumIntersectionWithFixedVertices(graph, fixedVertices));
				final long leftside1 = degreeCounter + i;

				if ( rightSide1 < leftside1 )
					return false;

				if ( l[i].isFixed() )
				{
					final Set<Integer> g = CompactDegreeSequence.getAllVerticesFromBag(graph, l[i]);
					System.out.println("fix all:" + Arrays.toString(g.toArray()));
					for (final Integer vert : g)
					{
						noOfEdgesBetweenFixedVertices += CompactDegreeSequence.addDegrees(graph, fixedVertices, vert);
						fixedVertices.add(vert);
					}
				}
				else
				{
					noOfEdgesBetweenFixedVertices += l[i].getMaximumIntersectionWithFixedVertices(graph, fixedVertices);
				}

				final int r2 = entryCounter + diffSequence.degrees[i];
				final long rightSide2 = diffSequence.getRightSum(i, r2) - noOfEdgesBetweenFixedVertices;
				final long leftside2 = degreeCounter + diffSequence.degrees[i] * i;

				if ( rightSide2 < leftside2 )
					return false;

				entryCounter += diffSequence.degrees[i];
				degreeCounter += i * diffSequence.degrees[i];
			}
		}
		return true;
	}

	public final int[] degrees;

	private int[] numberOfEntries;

	private int[] degreeSum;

	public CompactDegreeSequence(final CompactDegreeSequence sequence, final int i)
	{
		this.degrees = Arrays.copyOf(sequence.degrees, i);
	}

	public CompactDegreeSequence(final int[] i)
	{
		this.degrees = Arrays.copyOf(i, i.length);
	}

	public CompactDegreeSequence(final Node[] l)
	{
		new DegreeSequence();
		this.degrees = new int[DegreeSequence.getDelta(l) + 1];
		Arrays.fill(this.degrees, 0);
		for (int i = 1; i < l.length; i++)
		{
			this.degrees[l[i].getDegree()]++;
		}
	}

	// public CompactDegreeSequence(Integer[] array)
	// {
	// this.degrees = new int[array.length];
	// for (int i = 0; i < array.length; i++)
	// degrees[i] = array[i];
	// }

	/**
	 * 
	 * Returns the number of trivial degree one increases which are possible in any (optimal)
	 * outcome of the dynamic programming.
	 * 
	 * @param minAddCost
	 * @param k
	 * @return
	 */
	boolean canYouWasteDegreeOneVertices(final int minAddCost, final int k)
	{
		final int f = 2 * k - 1;
		final boolean[] remainingBlocks = new boolean[this.degrees.length];

		for (int i = 0; i <= this.getMaxValue(); i++)
		{
			if ( i == this.getMaxValue() )
			{
				remainingBlocks[i] = this.get(i) >= k;
			}
			else
			{
				remainingBlocks[i] = this.get(i) >= f || this.get(i + 1) >= f && this.get(i) >= k;
			}
		}

		int c = 0;

		for (int i = 0; i < this.getMaxValue(); i++)
		{
			if ( this.get(i) >= f && remainingBlocks[i + 1] )
			{
				c += this.get(i) - f;
			}
			if ( c >= minAddCost )
				return true;
		}
		return false;
	}

	/**
	 * Calculates the result of this degree sequence minus the given one. Entry i gives the number
	 * of vertices the have been increase by i from old to this.
	 * 
	 * @param oldSeq
	 * @return
	 */
	public CompactDegreeSequence difference(final CompactDegreeSequence oldSeq)
	{
		final int[] newSeq = Arrays.copyOf(this.degrees, this.degrees.length);

		final int[] result = new int[this.degrees.length];
		Arrays.fill(result, 0);

		for (int i = 0; i < oldSeq.degrees.length; i++)
		{
			int c = oldSeq.degrees[i];
			for (int j = i; j < this.degrees.length; j++)
			{
				if ( c <= 0 )
				{
					break;
				}
				final int r = Math.min(c, newSeq[j]);
				newSeq[j] -= r;
				c -= r;
				result[j - i] += r;
			}
		}
		return new CompactDegreeSequence(result);
	}

	public DegreesChangeList differenceWithDegrees(final CompactDegreeSequence oldSeq)
	{
		final int[] newSeq = Arrays.copyOf(this.degrees, this.degrees.length);

		final DegreesChangeList result = new DegreesChangeList();

		for (int i = 0; i < oldSeq.degrees.length; i++)
		{
			int c = oldSeq.degrees[i];
			for (int j = i; j < this.degrees.length; j++)
			{
				if ( c <= 0 )
				{
					break;
				}
				final int r = Math.min(c, newSeq[j]);
				newSeq[j] -= r;
				c -= r;
				if ( j - i > 0 && r > 0 )
				{
					result.add(j - i, i, r);
				}
			}
		}
		return result;
	}

	@Override
	public boolean equals(final Object obj)
	{
		if ( this == obj )
			return true;
		if ( obj == null )
			return false;
		if ( this.getClass() != obj.getClass() )
			return false;
		final CompactDegreeSequence other = (CompactDegreeSequence) obj;
		if ( !Arrays.equals(this.degrees, other.degrees) )
			return false;
		return true;
	}

	public int get(final int i)
	{
		return this.degrees[i];
	}

	// public CompactDegreeSequence newDifference(CompactDegreeSequence oldSeq)
	// {
	// int[] newDegrees = new int[this.degrees.length];
	//
	// int c=0;
	// for(int i=0;i<newDegrees.length;i++)
	// {
	// newDegrees[i]=this.getNumberOfVerticesInBags(0, i);
	// }
	// }

	public int[] getCopy(final int lastFullBag, final int nextFullBag)
	{
		return Arrays.copyOfRange(this.degrees, lastFullBag, nextFullBag);
	}

	public int getDegreeSumsInBags(final int fromBagIndex, final int endBagIndex)
	{
		if ( fromBagIndex > endBagIndex || fromBagIndex < 0 || endBagIndex >= this.degrees.length )
			throw new IllegalArgumentException();
		if ( this.degreeSum == null )
		{
			this.degreeSum = new int[this.degrees.length];
			for (int i = this.degrees.length - 1; i >= 0; i--)
			{
				this.degreeSum[i] = i * this.degrees[i] + (i < this.degrees.length - 1 ? this.degreeSum[i + 1] : 0);
			}
		}

		return this.degreeSum[fromBagIndex] - (endBagIndex < this.degrees.length - 1 ? this.degreeSum[endBagIndex + 1] : 0);
	}

	public int getMaxValue()
	{
		return this.degrees.length - 1;
	}

	// private long getRightSum2(int i, final int r2)
	// {
	// long rightSide2 = (r2 * (r2 - 1));
	// if ( r2 >= i - 1 )
	// {
	// rightSide2 += this.getDegreeSumsInBags(0, i - 1);
	// }
	// else
	// {
	// rightSide2 += (this.getNumberOfVerticesInBags(r2, i - 1) * r2);
	// rightSide2 += this.getDegreeSumsInBags(0, r2 - 1);
	// }
	// return rightSide2;
	// }

	/**
	 * Returns the number of elements in the bags from start to end index (values are cached).
	 * 
	 * @param fromBagIndex
	 * @param endBagIndex
	 * @return
	 */
	public int getNumberOfVerticesInBags(final int fromBagIndex, final int endBagIndex)
	{
		if ( fromBagIndex > endBagIndex || fromBagIndex < 0 || endBagIndex >= this.degrees.length )
			throw new IllegalArgumentException();
		if ( this.numberOfEntries == null )
		{
			this.numberOfEntries = new int[this.degrees.length];
			for (int i = this.degrees.length - 1; i >= 0; i--)
			{
				this.numberOfEntries[i] = this.degrees[i] + (i < this.degrees.length - 1 ? this.numberOfEntries[i + 1] : 0);
			}
		}
		return this.numberOfEntries[fromBagIndex] - (endBagIndex < this.degrees.length - 1 ? this.numberOfEntries[endBagIndex + 1] : 0);
	}

	private long getRightSum(final int i, final int r1)
	{
		long rightSide1 = r1 * (r1 - 1) + (this.getNumberOfVerticesInBags(i, this.getMaxValue()) - r1) * Math.min(r1, i);

		if ( r1 >= i - 1 )
		{
			rightSide1 += this.getDegreeSumsInBags(0, i - 1);
		}
		else
		{
			rightSide1 += this.getNumberOfVerticesInBags(r1, i - 1) * r1;
			rightSide1 += this.getDegreeSumsInBags(0, r1 - 1);
		}
		return rightSide1;
	}

	public int getTotalDegreeSum()
	{
		return this.getDegreeSumsInBags(0, this.degrees.length - 1);
	}

	public int getTotalEntrySum()
	{
		return this.getNumberOfVerticesInBags(0, this.degrees.length - 1);
	}

	public int getUpperBoundForKAnon(final int k)
	{
		return this.getUpperBoundWithSolForKAnon(k).getS1();
	}

	public static int getUpperBoundForKAnonAndRealizable(AdjacentGraph graph, final int k)
	{
		CompactDegreeSequence origSeq = new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph));
		Set<CompactDegreeSequence> allSolutions = new DegreeAnonymizer2(graph, origSeq, k, false, true).getAllkAnonRealSequences();
		int solutionDiffToOrig = allSolutions.iterator().next().difference(origSeq).getTotalDegreeSum();

		int minAddCostErdös = Integer.MAX_VALUE;

		int minSol = Integer.MAX_VALUE;
		for (CompactDegreeSequence sol : allSolutions)
		{
			int minAddSol = sol.difference(origSeq).noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();
			if ( minAddSol <= minAddCostErdös )
			{
				minAddCostErdös = minAddSol;
				boolean foundNewMinSol = false;
				for (int addCost = minAddSol; addCost + solutionDiffToOrig < minSol; addCost += 2)
				{
					final CostWaster waster = new CostWaster(origSeq, sol, addCost, k);
					for (CompactDegreeSequence realSol : waster.getAllRealizableSequences())
					{
						try
						{
							final int c = addCost + solutionDiffToOrig;
							if ( new GraphRealizer(graph, origSeq, realSol, c < 1000 ? 100 : 10, c < 100 ? 25 : c < 1000 ? 5 : c < 10000 ? 2 : 1,
									c < 100 ? 100 : c < 1000 ? 10 : c < 5000 ? 2 : 1).constructGraph() != null )
							{
								minSol = Math.min(minSol, realSol.difference(origSeq).getTotalDegreeSum());
								foundNewMinSol = true;
							}
						} catch (CannotRealize e)
						{
						}
					}
					if ( foundNewMinSol )
						break;
				}
			}
		}
		return minSol;
	}

	public int getUpperBoundForKAnonAndSimpleErdösRealizable(final int k)
	{
		final Store<Integer, CompactDegreeSequence> s = this.getUpperBoundWithSolForKAnon(k);
		final int minAddCost = s.getS2().difference(this).noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();
		if ( minAddCost > 0 )
		{
			for (int addCost = minAddCost; addCost <= k * minAddCost; addCost += 2)
			{
				final CostWaster waster = new CostWaster(this, s.getS2(), addCost, k);
				if ( waster.isThereARealizableSequences() )
					return addCost + s.getS2().difference(this).getTotalDegreeSum();
			}
		}
		else
			return s.getS2().difference(this).getTotalDegreeSum();

		return Integer.MAX_VALUE;
	}

	// int canYouWasteDegreeOneVertices(int minAddCost, int k)
	// {
	// int c = 0;
	// int f = (2 * k) - 1;
	//
	// int kplus1 = 0;
	// int kplus0 = 0;
	//
	// for (int i = 0; i <= getMaxValue(); i++)
	// {
	// if ( i == this.getMaxValue() && this.get(i) - k >= k )
	// {
	// int diff = this.get(i) - k;
	// while (diff >= k)
	// {
	// if ( diff > k && ((kplus1 < kplus0) || (kplus0 + kplus1 == 0 && (c + k +
	// 1) % 2 == minAddCost
	// % 2)) )
	// {
	// kplus1++;
	// diff -= (k + 1);
	// }
	// else
	// {
	// kplus0++;
	// diff -= k;
	// }
	// }
	// }
	// else if ( this.get(i) >= f )
	// {
	// if ( c < minAddCost )
	// {
	// final int n = Math.max(0, Math.min(this.get(i) - f, minAddCost - c));
	// if ( n >= k || (i < this.getMaxValue() && this.get(i + 1) >= f) )
	// c += n;
	// }
	// else
	// {
	// if ( i < this.getMaxValue() && this.get(i + 1) >= f && this.get(i) > f )
	// {
	// c++;
	// }
	// else if ( this.get(i) - f >= k )
	// {
	// int rest = this.get(i) - f;
	// while (rest >= k)
	// {
	// if ( kplus1 < kplus0 && rest > k )
	// {
	// rest -= (k + 1);
	// kplus1++;
	// }
	// else
	// {
	// rest -= k;
	// kplus0++;
	// }
	//
	// }
	// }
	// }
	//
	// if ( c >= minAddCost && (c % 2 == minAddCost % 2) )
	// return c;
	//
	// }
	// }
	//
	// while (c < minAddCost && kplus0 + kplus1 > 0)
	// {
	// int diff = minAddCost - c;
	// if ( (kplus1 == 0 || c % 2 == k % 2) && kplus0 > 0 )
	// {
	//
	// int fac = Math.min(kplus0, Math.max(1, (int) Math.ceil(diff / k)));
	// c += fac * k;
	// kplus0 -= fac;
	// }
	// else
	// {
	// int fac = Math.min(kplus1, Math.max(1, (int) Math.ceil(diff / (k + 1))));
	// c += (fac * (k + 1));
	// kplus1 -= fac;
	// }
	// }
	//
	// if ( c >= minAddCost )
	// {
	// if ( c % 2 == minAddCost % 2 )
	// return c;
	//
	// if ( (c % 2) != (k % 2) )
	// {
	// if ( kplus0 > 0 )
	// return c + k;
	// }
	// else if ( kplus1 > 0 )
	// return c + k + 1;
	//
	// }
	// return Integer.MAX_VALUE;
	// }

	public int getLevelOfAnonymity()
	{
		int t = Integer.MAX_VALUE;
		for (int i = 0; i < this.degrees.length; i++)
			if ( degrees[i] > 0 )
				t = Math.min(t, degrees[i]);
		return t;
	}

	Store<Integer, CompactDegreeSequence> getUpperBoundWithSolForKAnon(final int k)
	{
		final int arr[] = new int[this.degrees.length];

		int lastFullBag = -1;
		int currentBag = this.degrees.length - 1;
		while (currentBag >= 0)
		{
			if ( this.get(currentBag) > 0 )
			{
				int l = currentBag;
				while (l >= 0 && this.getNumberOfVerticesInBags(l, currentBag) < k)
				{
					l--;
				}
				if ( l < 0 )
				{
					if ( lastFullBag < 0 )
						throw new IllegalArgumentException();
					arr[lastFullBag] += this.getNumberOfVerticesInBags(0, currentBag);
					currentBag = -1;
				}
				else if ( l < currentBag )
				{
					final int c = this.degrees[currentBag] * (lastFullBag - currentBag);
					final int neededVertices = k - this.getNumberOfVerticesInBags(l + 1, currentBag);
					int c2 = this.getNumberOfVerticesInBags(l + 1, currentBag) * currentBag - this.getDegreeSumsInBags(l + 1, currentBag);
					c2 += neededVertices * (currentBag - l);
					if ( lastFullBag >= 0 && 2 * c < c2 )// 2 is completely
															// randomely chosen
					{
						arr[lastFullBag] += this.degrees[currentBag];
						currentBag--;
					}
					else if ( this.degrees[l] - neededVertices < k )
					{
						arr[currentBag] = this.getNumberOfVerticesInBags(l, currentBag);
						lastFullBag = currentBag;
						currentBag = l - 1;
					}
					else
					{
						arr[currentBag] = this.getNumberOfVerticesInBags(l + 1, currentBag) + neededVertices;
						arr[l] = this.get(l) - neededVertices;
						lastFullBag = l;
						currentBag = l - 1;
					}

				}
				else
				// l=currentbag
				{
					arr[currentBag] = this.get(currentBag);
					lastFullBag = currentBag;
					currentBag--;
				}
			}
			else
			{
				currentBag--;
			}
		}
		final CompactDegreeSequence compactDegreeSequence = new CompactDegreeSequence(arr);
		return new Store<Integer, CompactDegreeSequence>(compactDegreeSequence.difference(this).getTotalDegreeSum(), compactDegreeSequence);
	}

	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(this.degrees);
		return result;
	}

	public boolean isRealizable()
	{
		if ( this.getTotalDegreeSum() % 2 == 1 )
			return false;

		int entryCounter = 0;
		int degreeCounter = 0;

		for (int i = this.degrees.length - 1; i >= 1; i--)
		{
			if ( this.degrees[i] > 0 )
			{
				final int r1 = entryCounter + 1;
				final int r2 = entryCounter + this.degrees[i];

				final long rightSide1 = this.getRightSum(i, r1);
				final long rightSide2 = this.getRightSum(i, r2);

				final long leftside1 = degreeCounter + i;
				final long leftside2 = degreeCounter + this.degrees[i] * i;
				if ( rightSide1 < leftside1 || rightSide2 < leftside2 )
					return false;

				if ( r1 >= i )
					return true;

				entryCounter += this.degrees[i];
				degreeCounter += i * this.degrees[i];
			}
		}

		return true;
	}

	public int noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös()
	{
		int neededVertices = 0;

		int entryCounter = 0;
		int degreeCounter = 0;

		for (int i = this.degrees.length - 1; i >= 1; i--)
		{
			if ( this.degrees[i] > 0 )
			{
				final int r1 = entryCounter + 1;
				final int r2 = entryCounter + this.degrees[i];

				final long rightSide1 = this.getRightSum(i, r1);
				final long rightSide2 = this.getRightSum(i, r2);

				final long leftside1 = degreeCounter + i;
				final long leftside2 = degreeCounter + this.degrees[i] * i;

				if ( rightSide1 < leftside1 )
				{
					neededVertices = (int) Math.max(leftside1 - rightSide1, neededVertices);
				}

				if ( rightSide2 < leftside2 )
				{
					neededVertices = (int) Math.max(leftside2 - rightSide2, neededVertices);
				}

				if ( r1 >= i )
				{
					break;
				}

				entryCounter += this.degrees[i];
				degreeCounter += i * this.degrees[i];
			}
		}

		if ( (this.getTotalDegreeSum() % 2 + neededVertices % 2) % 2 == 0 )
			return neededVertices;
		return neededVertices + 1;
	}

	/**
	 * checks whether the degree sequence fulfills the criteria of Erdös-Gallai. A degree sequence
	 * fulfills this criteria if and only if the degree sequence is realizable/graphical.
	 * 
	 * @return
	 */
	public boolean OldisRealizable()
	{
		long sumFrom_i_plus1 = 0;
		long r = 0;
		for (int i = this.degrees.length - 1; i >= 0; i--)
		{
			for (int j = 1; j <= this.degrees[i]; j++)
			{
				r++;
				final long leftside = sumFrom_i_plus1 + j * i;

				if ( 2 * r >= i && r * (r - 1) >= leftside && this.getTotalDegreeSum() % 2 == 0 )
					return true;

				long rightside = r * (r - 1) + (this.degrees[i] - j) * Math.min(r, i);
				for (int i2 = i - 1; i2 >= 0; i2--)
				{
					rightside += this.degrees[i2] * Math.min(i2, r);
				}
				if ( leftside > rightside )
					return false;
			}
			sumFrom_i_plus1 += this.degrees[i] * i;
		}
		if ( sumFrom_i_plus1 % 2 == 0 )
			return true;
		return false;
	}

	public int OldnoOfAddDegreeOneVerticesForBeingRealizable()
	{
		int neededVertices = 0;
		long sumFrom_i_plus1 = 0;
		long r = 0;
		for (int i = this.degrees.length - 1; i >= 0; i--)
		{
			for (int j = 1; j <= this.degrees[i]; j++)
			{
				r++;
				final long leftside = sumFrom_i_plus1 + j * i;

				if ( 2 * r >= i && r * (r - 1) >= leftside )
					return (this.getTotalDegreeSum() % 2 + neededVertices % 2) % 2 == 0 ? neededVertices : neededVertices + 1;

				long rightside = r * (r - 1) + (this.degrees[i] - j) * Math.min(r, i);
				if ( leftside > rightside )
				{
					for (int i2 = i - 1; i2 >= 0; i2--)
					{
						rightside += this.degrees[i2] * Math.min(i2, r);
					}
				}
				if ( leftside > rightside )
				{
					neededVertices = Math.max(neededVertices, (int) (leftside - rightside));
				}
			}
			sumFrom_i_plus1 += this.degrees[i] * i;
		}
		if ( (sumFrom_i_plus1 % 2 + neededVertices % 2) % 2 == 0 )
			return neededVertices;
		else
			return neededVertices + 1;
	}

	@Override
	public String toString()
	{
		String s = "[";
		for (int i = 0; i < this.degrees.length; i++)
		{
			if ( this.degrees[i] > 0 )
			{
				s += "(" + i + ", #:" + this.degrees[i] + ")";
			}
		}
		s += "]";
		final String s2 = Arrays.toString(this.degrees);
		return s;
		// return s2.length() < s.length() ? s2 : s;
	}

	public String toStringFull()
	{
		return Arrays.toString(this.degrees);

	}

}
