package seppImpl;

import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Set;

import benchmarks.MainTest;

public class CostWaster
{

	private final int k;

	private final int cost;

	private final CompactDegreeSequence originalSequence;

	private CompactDegreeSequence anonSequence;

	private final int[] solIncrease;

	private int maxDegree;

	private BitSet[][] table;

	private boolean foundRealizable = false;

	private Set<CompactDegreeSequence> solutions = null;

	private final boolean findRealizableSequence;

	// public CostWaster(CompactDegreeSequence originalSequence, Triple[]
	// triples,
	// CompactDegreeSequence anonSequence, int additionalCosts, int k)
	// {
	// this.k = k;
	// this.originalSequence = originalSequence;
	// this.anonSequence = anonSequence;
	// this.triples = triples;
	// this.cost = additionalCosts;
	// this.maxDegree = anonSequence.getMaxValue();
	// }

	public CostWaster(final CompactDegreeSequence compactDegreeSequence, final CompactDegreeSequence anonSequence, final int addCost, final int k)
	{
		this(compactDegreeSequence, anonSequence, addCost, k, true);
	}

	public CostWaster(final CompactDegreeSequence compactDegreeSequence, final CompactDegreeSequence anonSequence, final int addCost, final int k,
			final boolean findRealizableSequence)
	{
		this.k = k;
		this.originalSequence = compactDegreeSequence;
		this.anonSequence = anonSequence;
		this.maxDegree = anonSequence.getMaxValue();
		this.solIncrease = new int[this.maxDegree];
		this.solIncrease[0] = compactDegreeSequence.get(0) - anonSequence.get(0);
		for (int i = 1; i < this.maxDegree; i++)
		{
			this.solIncrease[i] = compactDegreeSequence.get(i) - anonSequence.get(i) + this.solIncrease[i - 1];
		}

		// this.triples = triples;
		this.cost = addCost;
		this.findRealizableSequence = findRealizableSequence;
		this.solutions = new HashSet<CompactDegreeSequence>();
	}

	/**
	 * checks whether vertex degrees can be increased by a total of <code>cost</code>. Produces all
	 * solutions, not just optimal ones.
	 * 
	 * @param i
	 * @param remainingCost
	 * @param wasteTriples
	 * @return
	 */
	private boolean determineIncreases(final int i, final int remainingCost, final Triple[] wasteTriples)
	{
		final int currentBlockSize = this.anonSequence.get(i) + wasteTriples[i - 1].getT();

		if ( i <= 0 || i > this.maxDegree )
			throw new IllegalAccessError("This case should never occur!! Line " + DebugHelper.getLineNumber() + " in class "
					+ DebugHelper.getClassName());
		// return false;
		else if ( remainingCost == 0 )
		{
			if ( currentBlockSize < this.k && currentBlockSize > 0 )
				return false;
			for (int j = i; j <= this.maxDegree; j++)
			{
				wasteTriples[j] = new Triple(0, 0, 0);
			}
			this.examineSequence(wasteTriples);
			return true;
		}
		else if ( i == this.maxDegree )
			return false;
		else
		{
			boolean solFound = false;
			final int max = Math.min(remainingCost, currentBlockSize);

			wasteTriples[i] = new Triple(0, 0, 0);
			for (int j = 0; j <= max; j++)
			{
				if ( currentBlockSize - j == 0 || currentBlockSize - j >= this.k )
				{
					wasteTriples[i].setT(j);
					wasteTriples[i].setC(j);
					solFound = this.determineIncreases(i + 1, remainingCost - j, wasteTriples) || solFound;
				}
			}
			return solFound;
		}
	}

	/**
	 * Dynamic Programming Procedure to check whether <code>cost</code> vertex-degrees can be
	 * increased by exactly one. If such an increase exist, then it will be found.
	 * 
	 * Runs in O(MaxDegree * cost^2) time
	 * 
	 * @return
	 */
	private boolean determineOneIncreasesDynProg()
	{
		this.table = new BitSet[this.maxDegree + 1][];
		for (int i = 0; i <= this.maxDegree; i++)
		{
			this.table[i] = new BitSet[this.cost + 1];
			for (int j = 0; j <= this.cost; j++)
			{
				this.table[i][j] = new BitSet(this.cost + 1);
			}
		}

		int mint = 0;
		int maxt = 0;

		for (int i = 0; i <= this.maxDegree; i++)
		{

			for (int t = 0; t <= Math.min(this.cost, this.originalSequence.get(i)); t++)
			{

				for (int c = 0; c <= this.cost; c++)
				{
					if ( i == 0 )
					{
						if ( c == 0 && (this.anonSequence.get(i) >= this.k + t || this.anonSequence.get(i) == t) )
						{
							this.table[i][t].set(c);
						}
						else
						{
							this.table[i][t].clear(c);
						}
					}
					else
					{
						mint = 0;
						if ( this.anonSequence.get(i - 1) == 0 )
						{
							maxt = 0;
						}
						else
						{
							maxt = Math.min(this.originalSequence.get(i - 1) - this.solIncrease[i - 1], c);
						}

						this.table[i][t].clear(c);
						for (int tt = mint; tt <= maxt; tt++)
						{
							if ( this.anonSequence.get(i) - t + tt == 0 || this.anonSequence.get(i) - t + tt >= this.k )
							{
								if ( this.table[i - 1][tt].get(c - tt) )
								{
									this.table[i][t].set(c);
								}
							}
						}
					}
				}
			}
			if ( MainTest.logLevel > 8 )
			{
				System.out.println(Arrays.toString(this.table[i]));
			}
		}
		final int maxI = Math.min(this.originalSequence.get(this.maxDegree), this.cost);
		boolean foundSolution = false;

		for (int i = 0; i <= maxI; i++)
		{
			if ( (i == 0 || i == this.anonSequence.get(this.maxDegree) || this.anonSequence.get(this.maxDegree) - i >= this.k && i >= this.k)
					&& this.table[this.maxDegree][i].get(this.cost - i) )
			{
				foundSolution = true;
				final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
				wasteTriples[this.maxDegree] = new Triple(this.maxDegree, i, this.cost - i);
				this.recursivelyBuildAllSolutionSequences(this.maxDegree, wasteTriples);
			}
		}
		return foundSolution;
	}

	public int determineOneIncreasesHeuristic()
	{
		return this.determineOneIncreasesHeuristic(false);
	}

	/**
	 * Counts the number of untouched vertices in each block that might be increased by one without
	 * destroying the block. If this number is larger than <code>cost</code>, then a solution
	 * increasing exactly cost vertex-degrees is build. This is a heuristic. It might fail to find
	 * an existing solution!
	 * 
	 * Runs in O(maxDegree) time.
	 * 
	 * @return
	 */
	private int determineOneIncreasesHeuristic(final boolean constructSolution)
	{
		final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
		wasteTriples[this.maxDegree] = new Triple(this.maxDegree, 0, 0);
		int singleIncreases = 0; // number of vertices that might be increased
									// by one without
									// destroying one block
		int maxRemainingIncrease = 0;
		for (int i = 0; i < this.maxDegree; i++)
		{
			wasteTriples[i] = new Triple(i, 0, 0);
			maxRemainingIncrease = Math.max(0, this.cost - singleIncreases);
			if ( this.anonSequence.get(i) > this.k && this.anonSequence.get(i + 1) > 0 )
			{
				singleIncreases += this.anonSequence.get(i) - Math.max(this.solIncrease[i], this.k);
				wasteTriples[i].setT(Math.min(this.anonSequence.get(i) - Math.max(this.solIncrease[i], this.k), maxRemainingIncrease));
			}
		}

		if ( singleIncreases >= this.cost && constructSolution )
		{
			// System.out.println("Solution found; Single increases: " +
			// singleIncreases);
			this.examineSequence(wasteTriples);
		}
		return singleIncreases;
	}

	private void examineSequence(final Triple[] wasteTriples)
	{
		if ( MainTest.logLevel > 3 )
		{
			System.out.println("===");
			System.out.println("seq:\t" + this.anonSequence);
			System.out.println("trip:\t" + Arrays.toString(wasteTriples));
		}

		final int[] newsequence = new int[this.maxDegree + 2];
		Arrays.fill(newsequence, 0);

		newsequence[this.maxDegree + 1] = wasteTriples[this.maxDegree].getT();
		for (int i = 0; i <= this.maxDegree; i++)
		{
			newsequence[i] += this.anonSequence.get(i) - wasteTriples[i].getT() + (i > 0 ? wasteTriples[i - 1].getT() : 0);
		}
		final CompactDegreeSequence seq = new CompactDegreeSequence(newsequence);
		final CompactDegreeSequence differenceToOrignal = seq.difference(this.originalSequence);
		this.foundRealizable = this.foundRealizable || differenceToOrignal.isRealizable();

		if ( differenceToOrignal.isRealizable() || !this.findRealizableSequence )
		{
			this.solutions.add(seq);
			if ( MainTest.logLevel > 3 )
			{
				System.out.println("Final Seq:\t" + Arrays.toString(newsequence));
				System.out.println("===");
			}
		}
	}

	public Set<CompactDegreeSequence> getAllRealizableSequences()
	{
		this.solutions = new HashSet<CompactDegreeSequence>();

		if ( MainTest.logLevel > 3 )
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.originalSequence + "; " + Arrays.toString(this.solIncrease));
			System.out.println("anonSeq: " + this.anonSequence + "; cost: " + this.cost);
		}

		this.iterateOverAllDegreeSequences();

		if ( MainTest.logLevel > 3 )
		{
			System.out.println("Realizable Solution found: " + this.foundRealizable);
			System.out.println("============End CostWaster==============");
		}
		return this.solutions;
	}

	public boolean isThereARealizableSequences()
	{
		if ( MainTest.logLevel > 3 )
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.originalSequence + "; " + Arrays.toString(this.solIncrease));
			System.out.println("anonSeq: " + this.anonSequence + "; cost: " + this.cost);
		}

		this.iterateOverAllDegreeSequences();

		if ( MainTest.logLevel > 3 )
		{
			System.out.println("Realizable Solution found: " + this.foundRealizable);
			System.out.println("============End CostWaster==============");
		}
		return this.foundRealizable;
	}

	/**
	 * Given a k-anonymous sequence <code>anonSequence</code> obtained from
	 * <code>orignalSequence</code> (<code>triples</code> gives the trace in the dynamic programming
	 * table), this method computes all (Erdös-Gallai optimal) k-anon sequences obtainable from
	 * <code>anonSequence</code> by the cost of <code>additionalCosts</code>.
	 * 
	 * @param orignalSequence
	 * @param triples
	 * @param anonSequence
	 * @param additionalCosts
	 */
	public boolean iterateOverAllDegreeSequences()
	{
		// first try to increase the degree of additionalCosts many vertices by
		// exactly one
		// (Erdös-Gallai optimum)
		boolean solFound = false;
		final int increases = this.determineOneIncreasesHeuristic(true);
		if ( increases > this.cost )
		{
			solFound = true;
		}

		if ( !solFound )
		{
			// try to find increase by one solution via dynamic programming
			solFound = this.determineOneIncreasesDynProg();
		}

		if ( !solFound )
		{
			// no solution with just degree-one increases; todo...

			this.anonSequence = new CompactDegreeSequence(this.anonSequence, this.anonSequence.getMaxValue() + 1 + this.cost / this.k);
			this.maxDegree = this.anonSequence.getMaxValue();
			final int max = Math.min(this.cost, this.anonSequence.get(0));
			final Triple[] wasteTriples = new Triple[this.maxDegree + 1];
			boolean tmp = false;

			wasteTriples[0] = new Triple(0, 0, 0);
			for (int i = 0; i <= max; i++)
			{
				if ( this.anonSequence.get(0) - i == 0 || this.anonSequence.get(0) - i >= this.k )
				{
					wasteTriples[0].setT(i);
					wasteTriples[0].setC(i);
					tmp = this.determineIncreases(1, this.cost - i, wasteTriples);
					if ( tmp )
					{
						solFound = true;
					}
				}
			}
		}

		return solFound;
	}

	private void recursivelyBuildAllSolutionSequences(final int i, final Triple[] wasteTriples)
	{
		if ( i <= 0 )
		{
			this.examineSequence(wasteTriples);
		}
		else
		{
			final int t = wasteTriples[i].getT();
			final int c = wasteTriples[i].getC();

			int mint = 0;
			int maxt = 0;

			mint = 0;
			if ( this.anonSequence.get(i - 1) == 0 )
			{
				maxt = 0;
			}
			else
			{
				maxt = Math.min(this.originalSequence.get(i - 1) - this.solIncrease[i - 1], c);
			}

			for (int tt = mint; tt <= maxt; tt++)
			{
				if ( this.anonSequence.get(i) - t + tt == 0 || this.anonSequence.get(i) - t + tt >= this.k )
				{
					if ( this.table[i - 1][tt].get(c - tt) )
					{
						wasteTriples[i - 1] = new Triple(i - 1, tt, c - tt);
						this.recursivelyBuildAllSolutionSequences(i - 1, wasteTriples);
						wasteTriples[i - 1] = null;
					}
				}
			}
		}
	}
}
