package seppImpl;

import java.util.Arrays;
import java.util.BitSet;
import java.util.Iterator;
import java.util.Set;

import seppImpl.oldDynProgramming.DegreeAnonymizer;
import benchmarks.MainTest;
import dataStructures.Store;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;

public class DegreeAnonymizer2
{
	public static int logDR = 0;

	/**
	 * Calculates the cost of increasing the first t entries from the right border of the sequence
	 * formed by the first i blocks up to block j.
	 * 
	 * @param degreeSequence
	 * @param i
	 * @param t
	 * @param j
	 * @return
	 */
	// public static int cost(CompactDegreeSequence degreeSequence, int i, int
	// t, int j)
	// {
	// int count = 0;
	// for (int l = i; l >= 0; l--)
	// {
	// count += ((Math.min(t, degreeSequence.get(l))) * (j - l));
	// t -= Math.min(t, degreeSequence.get(l));
	// if ( t == 0 )
	// return count;
	// }
	// return Integer.MAX_VALUE;
	// }

	public static int cost(final CompactDegreeSequence degreeSequence, final int i, final int t, final int j)
	{
		if ( t == 0 )
			return 0;

		int uB = i;
		int lB = 0;
		int l = Math.max(0, i - 1);

		int noOfVerticesLtillI = degreeSequence.getNumberOfVerticesInBags(l, i);
		while (noOfVerticesLtillI < t || noOfVerticesLtillI - degreeSequence.get(l) >= t)
		{
			if ( noOfVerticesLtillI < t )
			{
				uB = l - 1;
			}
			else
			{
				lB = l + 1;
			}
			if ( lB > uB )
			{
				break;
			}
			l = lB + (uB - lB) / 2;
			noOfVerticesLtillI = degreeSequence.getNumberOfVerticesInBags(l, i);
		}

		final int no = l < i ? degreeSequence.getNumberOfVerticesInBags(l + 1, i) : 0;
		if ( no + Math.min(degreeSequence.get(l), t - no) < t )
			return Integer.MAX_VALUE;

		final int degreeSum = l < i ? no * j - degreeSequence.getDegreeSumsInBags(l + 1, i) : 0;
		return degreeSum + (t - no) * (j - l);

	}

	private final CompactDegreeSequence degreeSequence;

	private final int k;

	private int opt = Integer.MAX_VALUE;

	private final CompactDegreeSequence orginalSequence;

	private final boolean findRealizableSequence;

	private final boolean applyDR;

	private final AdjacentGraph graph;

	public SolutionDegreeIterator2 solutionDegreeIterator;

	public final int preprocessCost;

	public int lastExaminedCost;

	public DegreeAnonymizer2(final AdjacentGraph graph, final CompactDegreeSequence degreeSequence, final int k)
	{
		this(graph, degreeSequence, k, true, true);
	}

	public DegreeAnonymizer2(final AdjacentGraph graph, final CompactDegreeSequence degreeSequence, final int k,
			final boolean findRealizableSolution, final boolean applyDataReduction)
	{
		if ( findRealizableSolution && !degreeSequence.isRealizable() )
			throw new IllegalArgumentException("input sequence is not realizable!");

		this.graph = graph;
		this.orginalSequence = degreeSequence;
		this.k = k;
		this.applyDR = applyDataReduction;
		if ( applyDataReduction )
		{
			this.degreeSequence = this.iniDataReduction(degreeSequence);
			preprocessCost = this.degreeSequence.difference(this.orginalSequence).getTotalDegreeSum();
			if ( DegreeAnonymizer2.logDR > 0 )
			{

				System.out.print(",\tDR #mods:" + preprocessCost);
			}
		}
		else
		{
			this.degreeSequence = degreeSequence;
			preprocessCost = 0;
		}

		this.findRealizableSequence = findRealizableSolution;
	}

	public DegreeAnonymizer2(final AdjacentGraph graph, final int k)
	{
		this(graph, new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph)), k);
	}

	public DegreeAnonymizer2(final AdjacentGraph graph, final int k, final boolean findRealizableSolution, final boolean applyDataReduction)
	{
		this(graph, new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph)), k, findRealizableSolution, applyDataReduction);
	}

	public Set<CompactDegreeSequence> getAllkAnonRealSequences()
	{
		final BitSet[][] table = new BitSet[this.degreeSequence.getMaxValue() + 1][];
		this.solutionDegreeIterator = new SolutionDegreeIterator2(this.graph, table, this.orginalSequence, this.degreeSequence, this.k,
				this.findRealizableSequence, !this.applyDR);

		{// ini table
			for (int i = 0; i <= this.degreeSequence.getMaxValue(); i++)
			{
				final BitSet[] e = new BitSet[this.k];
				table[i] = e;
				// table.add(e);
				for (int t = 0; t < this.k; t++)
				{
					e[t] = new BitSet();
				}
			}
		}

		for (int c = 0;; c++)
		{
			for (int i = 0; i <= this.degreeSequence.getMaxValue(); i++)
			{
				final BitSet[] subTable = table[i];

				for (int t = 0; t < this.k; t++)
				{
					final BitSet bitSet = subTable[t];
					if ( i == 0 )
					{
						if ( c == 0 && (this.degreeSequence.get(i) - t >= this.k || this.degreeSequence.get(i) - t == 0) )
						{
							bitSet.set(c);
						}
					}
					else if ( this.degreeSequence.get(i) > t )
					{
						if ( this.degreeSequence.get(i) - t >= this.k )
						{
							if ( table[i - 1][0].get(c) )
							{
								bitSet.set(c);
							}
							else
							{
								int j = i - 1;
								int counter = 0;
								while (j >= 0 && (counter += this.degreeSequence.get(j)) < this.k
										&& DegreeAnonymizer2.cost(this.degreeSequence, i - 1, counter, i) <= c)
								{
									if ( DegreeAnonymizer.getTableEntry(this.degreeSequence, table, this.k, j, this.degreeSequence.get(j), c
											- DegreeAnonymizer2.cost(this.degreeSequence, i - 1, counter, i)) )
									{
										bitSet.set(c);
										break;
									}
									j--;
								}
							}
						}
						// 0 < degreeSequence.get(i) - t < k
						else if ( DegreeAnonymizer2.cost(this.degreeSequence, i - 1, this.k - (this.degreeSequence.get(i) - t), i) <= c )
						{
							int j = i - 1;
							int counter = 0;
							while (j >= 0 && counter + this.degreeSequence.get(j) < this.k - (this.degreeSequence.get(i) - t))
							{
								counter += this.degreeSequence.get(j);
								j--;
							}
							if ( j >= 0 )
							{
								final int remainingFromJ = this.k - (counter + this.degreeSequence.get(i) - t);
								if ( DegreeAnonymizer.getTableEntry(this.degreeSequence, table, this.k, j, remainingFromJ,
										c - DegreeAnonymizer2.cost(this.degreeSequence, i - 1, this.k - (this.degreeSequence.get(i) - t), i)) )
								{
									bitSet.set(c);
								}
								else
								{
									int u = this.degreeSequence.get(j) - remainingFromJ;
									while (u < this.k
											&& DegreeAnonymizer2.cost(this.degreeSequence, i - 1, u + this.k - (this.degreeSequence.get(i) - t), i) <= c)
									{
										if ( DegreeAnonymizer.getTableEntry(
												this.degreeSequence,
												table,
												this.k,
												j,
												this.degreeSequence.get(j),
												c
														- DegreeAnonymizer2.cost(this.degreeSequence, i - 1, u + this.k
																- (this.degreeSequence.get(i) - t), i)) )
										{
											bitSet.set(c);
											break;
										}

										j--;
										if ( j >= 0 )
										{
											u += this.degreeSequence.get(j);
										}
										else
										{
											break;
										}
									}
								}
							}
						}
					}
					else
					{
						bitSet.set(c, table[i - 1][t - this.degreeSequence.get(i)].get(c));
					}
				}
			}
			if ( table[this.degreeSequence.getMaxValue()][0].get(c) )
			{
				if ( MainTest.logLevel > 0 )
				{
					System.out.println("found dyn prog solution for c:" + (c + preprocessCost) + "\tcurrentUpperBound:"
							+ this.solutionDegreeIterator.getCurrentUpperbound() + "\t timeSoFar:"
							+ (System.currentTimeMillis() - MainTest.timeStamp) / 1000d + "sec");
				}
				this.opt = Math.min(c, this.opt);
				this.lastExaminedCost = c;
			}
			if ( this.opt != Integer.MAX_VALUE )
			{
				Set<CompactDegreeSequence> s;
				if ( (s = this.solutionDegreeIterator.getAllRealizableSequences(c)) != null )
					return s;
				if ( MainTest.logLevel > 3 )
				{
					System.out.println("test solution of size:" + c + "(recInstances:"
							+ this.solutionDegreeIterator.getNumberOfRecConsideredInstances() + ")");
				}
			}
		}
	}

	public SolutionDegreeIterator2 getLastSolutionDegreeIterator()

	{
		return this.solutionDegreeIterator;
	}

	/**
	 * Returns the set of k-anon sequences with the smallest difference c to the input sequence such
	 * that the "increase" from the given to the resulting instance is graphically-realizable
	 * (Erdös-Galii).
	 * 
	 * @return
	 */
	public int getSizeOfMinimumSolution()
	{
		final Set<CompactDegreeSequence> s = this.getAllkAnonRealSequences();
		if ( s != null && s.size() > 0 )
			return s.iterator().next().difference(this.orginalSequence).getTotalDegreeSum();
		return -1;
	}

	private CompactDegreeSequence iniDataReduction(final CompactDegreeSequence degreeSequence2)
	{
		final int[] newSequence = degreeSequence2.getCopy(0, degreeSequence2.getMaxValue() + 1);

		final CompactDegreeSequence sol = degreeSequence2.getUpperBoundWithSolForKAnon(k).getS2();
		final int minAddCost = sol.difference(degreeSequence2).noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();

		if ( degreeSequence2.canYouWasteDegreeOneVertices(minAddCost, this.k) )
		{
			int lastFullBag = 0;

			for (int i = 0; i <= degreeSequence2.getMaxValue(); i++)
			{

				if ( degreeSequence2.get(i) >= this.k )
				{
					lastFullBag = i;
				}
				else if ( degreeSequence2.get(i) < this.k && degreeSequence2.get(i) > 0 )
				{
					if ( i < degreeSequence2.getMaxValue() )
					{
						int nextFullBag = getNextFullBag(i + 1, degreeSequence2);

						if ( nextFullBag > 0 && (lastFullBag > 0 || nextFullBag < degreeSequence2.getMaxValue()) )
						{

							// System.out.println("test:" + lastFullBag + "(" +
							// degreeSequence2.get(lastFullBag) + ")-" + nextFullBag + "("
							// + degreeSequence2.get(nextFullBag) + ")");
							final CompactDegreeSequence parSequence = new CompactDegreeSequence(Arrays.copyOfRange(newSequence, lastFullBag,
									nextFullBag + 1));
							final Set<CompactDegreeSequence> t = new DegreeAnonymizer2(this.graph, parSequence, this.k, false, false)
									.getAllkAnonRealSequences();
							final Iterator<CompactDegreeSequence> iter = t.iterator();

							while (iter.hasNext())
							{
								final CompactDegreeSequence parSolution = iter.next();
								final CompactDegreeSequence difference = parSolution.difference(parSequence);

								if ( (lastFullBag == 0 || parSolution.get(0) >= k)
										&& (difference.getMaxValue() == 1 || difference.getNumberOfVerticesInBags(2, difference.getMaxValue()) <= 0) )
								{
									if ( MainTest.logLevel > 1 )
									{
										System.out.println("DR:" + parSequence + ">\n" + parSolution);
									}
									for (int r = 0; r <= parSolution.getMaxValue(); r++)
									{
										newSequence[lastFullBag + r] = parSolution.get(r);
									}

									break;
								}
							}

							lastFullBag = nextFullBag;
							i = nextFullBag;
						}
					}
					else
					{
						final CompactDegreeSequence t = new CompactDegreeSequence(newSequence);
						int index = t.getMaxValue() - 1;
						while (t.getNumberOfVerticesInBags(index, t.getMaxValue()) < this.k)
						{
							newSequence[index] = 0;
							index--;
						}
						newSequence[t.getMaxValue()] = this.k;
						newSequence[index] = newSequence[index] - (this.k - t.getNumberOfVerticesInBags(index + 1, t.getMaxValue()));
					}
				}
			}

		}

		// return degreeSequence2;
		return new CompactDegreeSequence(newSequence);

	}

	private int getNextFullBag(int i, CompactDegreeSequence degreeSequence2)
	{
		int c = 0;
		int cIndex = -1;
		for (int t = i; t <= degreeSequence2.getMaxValue(); t++)
		{
			if ( degreeSequence2.get(t) >= k )
			{
				if ( cIndex == -1 )
					cIndex = t;

				c += degreeSequence2.get(t) - k;
			}
			else
			{
				cIndex = -1;
				c = 0;
			}
			if ( c >= k - 1 )
				return cIndex;
		}
		return -1;
	}
}
