package seppImpl;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.set.TIntSet;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.TreeMap;
import java.util.Vector;

import sun.reflect.generics.reflectiveObjects.NotImplementedException;
import dataStructures.Store;
import dataStructures.utils.IntStore;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.data.DegreeSequence;
import de.vibee.kanon.data.IntTupel;

public class GraphRealizer
{
	class DegreeIncrease implements Cloneable

	{
		private int degreeIncrease;

		private final int originalDegree;

		private int vertex = Integer.MAX_VALUE;

		public DegreeIncrease(final int originalDegree)
		{
			this.originalDegree = originalDegree;
			this.degreeIncrease = 0;
		}

		@Override
		protected DegreeIncrease clone()
		{
			final DegreeIncrease s = new DegreeIncrease(this.originalDegree);
			s.degreeIncrease = this.degreeIncrease;
			s.vertex = this.vertex;
			return s;
		}

		public int getDegreeIncrease()
		{
			return this.degreeIncrease;
		}

		public int getOriginalDegree()
		{
			return this.originalDegree;
		}

		public void setDegreeIncrease(final int increase)
		{
			this.degreeIncrease = increase;
		}

		public void setVertex(final int vertex)
		{
			this.vertex = vertex;
		}

		@Override
		public String toString()
		{
			return "[" + this.originalDegree + "->" + (this.degreeIncrease + this.originalDegree)
					+ (this.vertex != Integer.MAX_VALUE ? ", vert:" + this.vertex : "") + "]";
		}

	}

	class DegreeIncreaseCollection implements Cloneable
	{
		private final Vector<DegreeIncrease> finalDegreeIncreases;

		private final Vector<DegreeIncrease> opendedDegreeIncreases;

		public DegreeIncreaseCollection()
		{
			this.opendedDegreeIncreases = new Vector<DegreeIncrease>();
			this.finalDegreeIncreases = new Vector<DegreeIncrease>();
		}

		@Override
		protected DegreeIncreaseCollection clone()
		{
			final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();
			for (final DegreeIncrease in : this.opendedDegreeIncreases)
			{
				coll.opendedDegreeIncreases.add(in.clone());
			}
			for (final DegreeIncrease in : this.finalDegreeIncreases)
			{
				coll.finalDegreeIncreases.add(in.clone());
			}
			return coll;
		}

		public void finalize(final int degree, final int numberVerticesToFinalize)
		{
			final Iterator<DegreeIncrease> iterator = this.opendedDegreeIncreases.iterator();
			int i = 1;
			while (iterator.hasNext() && i++ <= numberVerticesToFinalize)
			{
				final DegreeIncrease in = iterator.next();
				iterator.remove();
				in.setDegreeIncrease(degree - in.getOriginalDegree());
				this.finalDegreeIncreases.add(in);
			}
		}

		public Vector<DegreeIncrease> getFinializedDegrees()
		{
			return this.finalDegreeIncreases;
		}

		public int getNoOpenedVertices()
		{
			return this.opendedDegreeIncreases.size();
		}

		public void openVertices(final int degree, final int noOfVerticesToOpen)
		{
			for (int i = 0; i < noOfVerticesToOpen; i++)
			{
				this.opendedDegreeIncreases.add(new DegreeIncrease(degree));
			}
		}

		@Override
		public String toString()
		{
			return "DegreeIncreaseCollection [vec=" + Arrays.toString(this.opendedDegreeIncreases.toArray()) + ", finalDegreeIncreases="
					+ Arrays.toString(this.finalDegreeIncreases.toArray()) + "]";
		}

	}

	public final int innerIteration;

	public static int log = 1;

	public final int outerIter;

	private static final Random rnd = new Random();

	public static int shuffleMapping = 0;

	public static int sortVerticesRealize = 2;

	private static void disturbOrderALittleBit(Vector finalDegreeIncreases)
	{
		final int size = finalDegreeIncreases.size();
		if ( size > 2 )
		{
			final int range = Math.max((int) (size * 0.1f), 1);
			for (int i = 0; i < size - range; i++)
				if ( GraphRealizer.rnd.nextBoolean() )
				{
					Collections.swap(finalDegreeIncreases, i, i + GraphRealizer.rnd.nextInt(range + 1));
				}
		}
	}

	private static ArrayList<Integer> getShuffeldIndexRange(final int j)
	{
		final ArrayList<Integer> indices = new ArrayList<Integer>();
		{
			for (int i = 0; i < j; i++)
			{
				indices.add(i);
			}
			Collections.shuffle(indices);
		}
		return indices;
	}

	private Integer[] degreesWithPossibleJumps;

	private final AdjacentGraph graph;

	private final CompactDegreeSequence initialSeq;

	private int[] jumpDegreePosition;

	private Integer[] maxAmountOfJumps;

	// private DegreeIncreaseCollection
	// getAdaptiveDegreeIncreaseList(CompactDegreeSequence
	// solutionSeq)
	// {
	//
	// final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();
	//
	// for (int i = 0; i <= solutionSeq.getMaxValue(); i++)
	// {
	// int verticesToFinalize = Math.min(solutionSeq.get(i),
	// coll.getNoOpenedVertices());
	// int noOfVerticesToOpen = (i <= this.initialSeq.getMaxValue() ?
	// this.initialSeq.get(i) : 0) -
	// (solutionSeq.get(i) - verticesToFinalize);
	//
	// if ( verticesToFinalize > 0 && noOfVerticesToOpen > 0 &&
	// this.conflictMap.get(i) != null &&
	// this.conflictMap.get(i).size() > 0 )// adaptive
	// // part
	// {
	// if ( noOfVerticesToOpen == this.initialSeq.get(i) ||
	// this.conflictMap.get(i).size() ==
	// this.initialSeq.get(i) )
	// {
	// float prob = 0.4f;
	// while (verticesToFinalize > 0 && noOfVerticesToOpen > 0 && Math.random()
	// >= prob)
	// {
	// verticesToFinalize--;
	// noOfVerticesToOpen--;
	// prob += ((1f - prob) * 0.5f * Math.random());
	// }
	// }
	// }
	//
	// coll.finalize(i, verticesToFinalize);
	// coll.openVertices(i, noOfVerticesToOpen);
	// }
	// if ( GraphRealizer.log > 1 )
	// {
	// System.out.println("new coll:" + coll);
	// }
	// return coll;
	// }

	private int[] numberOfJumps;

	private final CompactDegreeSequence solutionSeq;

	//
	// private void getDegreeVertexMapping(DegreeIncreaseCollection coll)
	// {
	// final TreeMap<Integer, List<Integer>> degreeToPossibleVertices = this
	// .getPossibleDegreeToVertexMapping(coll);
	//
	// final int[] indexCounter = new int[this.initialSeq.getMaxValue() + 1];
	// Arrays.fill(indexCounter, 0);
	//
	// final int[] neededVertices = this.getNeededVerticesPerDegree(coll);
	//
	// final int[] possibleJumps = new int[this.initialSeq.getMaxValue() + 1];
	// Arrays.fill(possibleJumps, 0);
	// for (int i = 0; i < possibleJumps.length; i++)
	// possibleJumps[i] = this.initialSeq.get(i) - neededVertices[i];
	//
	// System.out.println("possible :"+Arrays.toString(possibleJumps));
	// int[] usedJumps = new int[this.initialSeq.getMaxValue() + 1];
	// Arrays.fill(usedJumps, 0);
	//
	// Vector<Integer> assignedVertices = new Vector<>();
	// for (final DegreeIncrease deg : coll.getFinializedDegrees())
	// {
	// List<Integer> degreeList = degreeToPossibleVertices
	// .get(deg.originalDegree);
	// if (usedJumps[deg.originalDegree] < possibleJumps[deg.originalDegree]
	// && numberOfEdges(assignedVertices,
	// degreeList.get(indexCounter[deg.originalDegree])) > numberOfEdges(
	// assignedVertices,
	// degreeList
	// .get(indexCounter[deg.originalDegree + 1])))
	// {
	// indexCounter[deg.originalDegree]++;
	// usedJumps[deg.originalDegree]++;
	// System.err.println("jump");
	// }
	//
	// deg.setVertex(degreeList.get(indexCounter[deg.originalDegree]++));
	// assignedVertices.add(deg.vertex);
	// }
	//
	// }

	public GraphRealizer(final AdjacentGraph graph, final CompactDegreeSequence iniSeq, final CompactDegreeSequence solutionSeq, int maxReorg,
			int innerIter, int outerIter)
	{
		this.graph = graph;
		this.initialSeq = iniSeq;
		this.solutionSeq = solutionSeq;
		this.maxReorg = maxReorg;
		this.innerIteration = innerIter;
		this.outerIter = outerIter;
	}

	public GraphRealizer(final AdjacentGraph graph, final CompactDegreeSequence iniSeq, final CompactDegreeSequence solutionSeq, int maxReorg)
	{
		this(graph, iniSeq, solutionSeq, maxReorg, 25, 100);
	}

	final int maxReorg;

	public GraphRealizer(final AdjacentGraph graph, final CompactDegreeSequence iniSeq, final CompactDegreeSequence solutionSeq)
	{
		this(graph, iniSeq, solutionSeq, Math.max(100, (int) (0.1 * solutionSeq.difference(iniSeq).getTotalDegreeSum())));
	}

	public GraphRealizer(final AdjacentGraph graph, final CompactDegreeSequence solutionSeq)
	{
		this(graph, new CompactDegreeSequence(DegreeSequence.calcDegreeSequence(graph)), solutionSeq);
	}

	public Set<IntTupel> constructGraph() throws CannotRealize
	{
		DegreeIncreaseCollection iniColl = this.getInitialDegreeIncreaseList(this.initialSeq, this.solutionSeq);

		final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices = this.getPossibleDegreeToVertexMapping(iniColl);

		int trials = 0;

		long time = System.currentTimeMillis();
		while (true)
		{
			trials++;
			int tmp = 0;

			while (tmp++ < innerIteration)
			{
				final DegreeIncreaseCollection coll = iniColl.clone();
				this.getDegreeVertexMapping(coll, degreeToPossibleVertices);

				try
				{
					final Set<IntTupel> realizeMapping = this.realizeMapping(coll.finalDegreeIncreases);

					if ( log > 0 )
					{
						System.out.print("sucess realize:" + trials + " (inner:" + tmp + ")\n");
					}
					return realizeMapping;
				} catch (final CannotRealize e)
				{
					if ( log > 1 && trials % 2 == 0 && tmp % innerIteration == 0 )
					{
						System.out.println("trials:" + trials + ", inner:" + tmp + " (" + (System.currentTimeMillis() - time) / 1000d + "s)");
						time = System.currentTimeMillis();
					}

					if ( trials >= outerIter && tmp == innerIteration )
						throw e;
				}
			}
			// iniColl = this.getAdaptiveDegreeIncreaseList(solutionSeq);
			boolean realizable = false;
			while (!realizable)
			{
				try
				{
					iniColl = this.getAdaptiveDegreeIncreaseList2(this.solutionSeq);
				} catch (final IllegalAccessError e)
				{
					throw new CannotRealize(this, -1, -1);
				}
				realizable = this.isRealizable(iniColl);
			}
			// iniColl = this.getInitialDegreeIncreaseList(this.initialSeq,
			// solutionSeq);
		}
	}

	private DegreeIncreaseCollection getAdaptiveDegreeIncreaseList2(final CompactDegreeSequence solutionSeq)
	{
		if ( this.degreesWithPossibleJumps == null || this.degreesWithPossibleJumps.length == 0 )
			throw new IllegalAccessError();

		if ( this.jumpDegreePosition == null )
		{
			this.jumpDegreePosition = new int[1];
			this.numberOfJumps = new int[1];
			Arrays.fill(this.numberOfJumps, 1);
		}
		else if ( this.increaseInNumberOfJumpsPossible() )
		{
		}
		else
		{
			boolean success = false;
			for (int k = this.jumpDegreePosition.length - 1; k >= 0; k--)
			{
				if ( k == this.jumpDegreePosition.length - 1 && this.jumpDegreePosition[k] < this.degreesWithPossibleJumps.length - 1
						|| k < this.jumpDegreePosition.length - 1 && this.jumpDegreePosition[k] < this.jumpDegreePosition[k + 1] - 1 )
				{
					this.jumpDegreePosition[k]++;
					this.numberOfJumps[k] = 1;
					for (int k2 = k + 1; k2 < this.jumpDegreePosition.length; k2++)
					{
						this.jumpDegreePosition[k2] = this.jumpDegreePosition[k2 - 1] + 1;
						this.numberOfJumps[k2] = 1;
					}
					success = true;
					break;
				}
			}
			if ( !success )
			{
				if ( this.jumpDegreePosition.length < this.degreesWithPossibleJumps.length )
				{
					this.jumpDegreePosition = new int[this.jumpDegreePosition.length + 1];
					for (int l = 0; l < this.jumpDegreePosition.length; l++)
					{
						this.jumpDegreePosition[l] = l;
					}
					this.numberOfJumps = new int[this.jumpDegreePosition.length];
					Arrays.fill(this.numberOfJumps, 1);
				}
				else
					throw new IllegalAccessError();
			}
		}

		int pCount = 0;
		final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();

		for (int deg = 0; deg <= solutionSeq.getMaxValue(); deg++)
		{

			int verticesToFinalize = Math.min(solutionSeq.get(deg), coll.getNoOpenedVertices());
			int noOfVerticesToOpen = (deg <= this.initialSeq.getMaxValue() ? this.initialSeq.get(deg) : 0)
					- (solutionSeq.get(deg) - verticesToFinalize);

			if ( pCount < this.jumpDegreePosition.length && deg == this.degreesWithPossibleJumps[this.jumpDegreePosition[pCount]] )
			{
				verticesToFinalize -= this.numberOfJumps[pCount];
				noOfVerticesToOpen -= this.numberOfJumps[pCount];
				pCount++;
			}
			coll.finalize(deg, verticesToFinalize);
			coll.openVertices(deg, noOfVerticesToOpen);

		}
		return coll;

	}

	private void getDegreeVertexMapping(final DegreeIncreaseCollection coll, final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices)
	{
		HashSet<Integer> assignedVertices = new HashSet<>(coll.finalDegreeIncreases.size());

		final int[] indexCounter = new int[this.initialSeq.getMaxValue() + 1];
		Arrays.fill(indexCounter, 0);
		for (final DegreeIncrease deg : coll.getFinializedDegrees())
		{
			Vector<Integer> s = degreeToPossibleVertices.get(deg.originalDegree);
			if ( indexCounter[deg.originalDegree] == 0 )
			{
				if ( shuffleMapping == 0 )
				{
					Collections.shuffle(s);
				}
				else if ( shuffleMapping == 1 || shuffleMapping == 2 )
				{
					Vector<IntStore<Integer>> sVec = new Vector<>();
					for (int v : s)
						sVec.add(new IntStore<Integer>(getNoVertices(assignedVertices, v), v));
					Collections.sort(sVec);

					Vector<Integer> newVec = new Vector<>();
					for (IntStore<Integer> vec : sVec)
						newVec.add(vec.getS2());
					degreeToPossibleVertices.put(deg.originalDegree, newVec);
					s = newVec;
				}
				if ( shuffleMapping == 2 )
					disturbOrderALittleBit(s);
			}
			Integer vertex = s.get(indexCounter[deg.originalDegree]++);
			assignedVertices.add(vertex);
			deg.setVertex(vertex);
		}
	}

	private Iterable<DegreeIncrease> getIncreaseIter(Vector<IntStore<Store<DegreeIncrease, Vector<DegreeIncrease>>>> degreeIncreases)
	{
		class D implements Iterable<DegreeIncrease>
		{
			private final Iterator<IntStore<Store<DegreeIncrease, Vector<DegreeIncrease>>>> d;

			public D(Vector<IntStore<Store<DegreeIncrease, Vector<DegreeIncrease>>>> degreeIncreases)
			{
				this.d = degreeIncreases.iterator();
			}

			@Override
			public Iterator<DegreeIncrease> iterator()
			{
				class Iter implements Iterator<DegreeIncrease>
				{

					@Override
					public boolean hasNext()
					{
						return D.this.d.hasNext();
					}

					@Override
					public DegreeIncrease next()
					{
						return D.this.d.next().getS2().getS1();
					}

					@Override
					public void remove()
					{
						throw new NotImplementedException();
					}

				}
				;
				return new Iter();
			}

		}
		;

		return new D(degreeIncreases);
	}

	private DegreeIncreaseCollection getInitialDegreeIncreaseList(final CompactDegreeSequence initialSeq, final CompactDegreeSequence solutionSeq)
	{
		final DegreeIncreaseCollection coll = new DegreeIncreaseCollection();

		final ArrayList<Integer> indices = new ArrayList<Integer>();
		final ArrayList<Integer> amoungOfJumpsT = new ArrayList<Integer>();

		for (int deg = 0; deg <= solutionSeq.getMaxValue(); deg++)
		{
			final int verticesToFinalize = Math.min(solutionSeq.get(deg), coll.getNoOpenedVertices());
			coll.finalize(deg, verticesToFinalize);
			final int noOfVerticesToOpen = (deg <= initialSeq.getMaxValue() ? initialSeq.get(deg) : 0) - (solutionSeq.get(deg) - verticesToFinalize);
			coll.openVertices(deg, noOfVerticesToOpen);

			if ( noOfVerticesToOpen > 0 && verticesToFinalize > 0 )
			{
				indices.add(deg);
				amoungOfJumpsT.add(Math.min(noOfVerticesToOpen, verticesToFinalize));
			}
		}

		this.degreesWithPossibleJumps = indices.toArray(new Integer[0]);
		this.maxAmountOfJumps = amoungOfJumpsT.toArray(new Integer[0]);

		if ( GraphRealizer.log > 1 )
		{
			System.out.println("poss:" + this.degreesWithPossibleJumps.length + "_" + Arrays.toString(this.degreesWithPossibleJumps)
					+ "_amount of Jumps:" + Arrays.toString(this.maxAmountOfJumps));
		}
		return coll;
	}

	private int[] getNeededVerticesPerDegree(final DegreeIncreaseCollection coll)
	{
		final int[] neededVertices = new int[this.initialSeq.getMaxValue() + 1];
		Arrays.fill(neededVertices, 0);

		for (final DegreeIncrease deg : coll.getFinializedDegrees())
		{
			neededVertices[deg.originalDegree]++;
		}
		return neededVertices;
	}

	private int getNoVertices(HashSet<Integer> assignedVertices, int vertex)
	{
		int i = 0;
		if ( assignedVertices.size() < this.graph.getDegree(vertex) )
		{
			for (int v : assignedVertices)
			{
				if ( graph.isAdjacent(v, vertex) )
					i++;
			}
		}
		else
		{
			TIntIterator iter = this.graph.getAdjacentNodes(vertex).iterator();
			while (iter.hasNext())
			{
				if ( assignedVertices.contains(iter.next()) )
					i++;
			}
		}
		return i;
	}

	private TreeMap<Integer, Vector<Integer>> getPossibleDegreeToVertexMapping(final DegreeIncreaseCollection coll)
	{
		// final int[] neededVertices = this.getNeededVerticesPerDegree(coll);

		final TreeMap<Integer, Vector<Integer>> degreeToPossibleVertices = new TreeMap<>();

		for (final DegreeIncrease co : coll.finalDegreeIncreases)
		{
			if ( !degreeToPossibleVertices.containsKey(co.originalDegree) )
			{
				final Vector<Integer> possibleVertices = new Vector<>();
				final TIntIterator iter = this.graph.getAllNodesWithDegree(co.originalDegree).iterator();
				while (iter.hasNext())
				{
					possibleVertices.add(iter.next());
				}
				degreeToPossibleVertices.put(co.originalDegree, possibleVertices);
			}
		}

		// final int maxValue = this.initialSeq.getMaxValue();
		// for (int deg = 0; deg <= maxValue; deg++)
		// {
		// final Vector<Integer> possibleVertices = new Vector<>();
		//
		// // if (possibleVertices.size() < neededVertices[deg])
		// // throw new IllegalArgumentException("not enough vertices");
		// // Collections.shuffle(possibleVertices);
		// degreeToPossibleVertices.put(deg, possibleVertices);
		// }
		return degreeToPossibleVertices;
	}

	private boolean increaseInNumberOfJumpsPossible()
	{
		for (int i = 0; i < this.numberOfJumps.length; i++)
		{
			if ( this.numberOfJumps[i] < this.maxAmountOfJumps[this.jumpDegreePosition[i]] )
			{
				this.numberOfJumps[i]++;
				return true;
			}
		}
		return false;
	}

	// private Set<IntTupel> realizeMapping(DegreeIncreaseCollection coll)
	// throws CannotRealize
	// {
	// final Set<IntTupel> newEdges = new HashSet<>();
	// final Vector<DegreeIncrease> degreeIncreases = new Vector<>();
	// degreeIncreases.addAll(coll.getFinializedDegrees());
	// Collections.shuffle(degreeIncreases);
	//
	// final ArrayList<Integer> tmpIndices =
	// GraphRealizer.getShuffeldIndexRange(degreeIncreases.size());
	//
	// boolean stillEdgesMissing = true;
	// while (stillEdgesMissing)
	// {
	// stillEdgesMissing = false;
	// for (final DegreeIncrease e1 : degreeIncreases)
	// {
	// while (e1.degreeIncrease > 0)
	// {
	// stillEdgesMissing = true;
	// boolean found = false;
	//
	// Collections.shuffle(tmpIndices);
	// for (final int j : tmpIndices)
	// {
	// final DegreeIncrease e2 = degreeIncreases.get(j);
	// if ( this.possibleToInsertEdge(newEdges, e1, e2) )
	// {
	// newEdges.add(new IntTupel(e2.vertex, e1.vertex));
	// e2.degreeIncrease--;
	// e1.degreeIncrease--;
	// found = true;
	// break;
	// }
	// }
	// if ( !found )
	// {
	// this.reorganizeSolution(e1, degreeIncreases, newEdges);
	// }
	// }
	// }
	// }
	// return newEdges;
	// }

	private boolean isRealizable(final DegreeIncreaseCollection iniColl)
	{
		int[] intArray = new int[10];
		for (final DegreeIncrease deg : iniColl.finalDegreeIncreases)
		{
			if ( deg.degreeIncrease >= intArray.length )
			{
				intArray = Arrays.copyOf(intArray, deg.degreeIncrease + 1);
			}
			intArray[deg.degreeIncrease]++;
		}
		return new CompactDegreeSequence(intArray).isRealizable();
	}

	private Vector<DegreeIncrease> possibleNewNeighbors(final Vector<DegreeIncrease> verticesThatNeedNeighbors, final int vertex)
	{
		final Vector<DegreeIncrease> p = new Vector<>();
		for (final DegreeIncrease v : verticesThatNeedNeighbors)
		{
			if ( vertex != v.originalDegree && !this.graph.isAdjacent(vertex, v.originalDegree) )
			{
				p.add(v);
			}
		}
		return p;
	}

	private boolean possibleToInsertEdge(final Set<IntTupel> newEdges, final int vertex1, final int vertex2)
	{
		return vertex1 != vertex2 && !newEdges.contains(new IntTupel(vertex1, vertex2)) && !this.graph.isAdjacent(vertex1, vertex2);
	}

	private Set<IntTupel> realizeMapping(final Vector<DegreeIncrease> finalDegreeIncreases) throws CannotRealize
	{
		final Set<IntTupel> newEdges = new HashSet<>();
		final Comparator<? super DegreeIncrease> d = new Comparator<DegreeIncrease>() {

			@Override
			public int compare(DegreeIncrease o1, DegreeIncrease o2)
			{
				return o2.degreeIncrease - o1.degreeIncrease;
			}

		};
		if ( sortVerticesRealize == 0 )
		{
			Collections.sort(finalDegreeIncreases, d);
			this.disturbOrderALittleBit(finalDegreeIncreases);
		}
		else if ( sortVerticesRealize == 1 )
		{
			Collections.sort(finalDegreeIncreases, d);
			// this.disturbOrderALittleBit(finalDegreeIncreases);
		}
		else
			Collections.shuffle(finalDegreeIncreases);

		int reorg = 0;
		final int numberOfIncreaseVertices = finalDegreeIncreases.size();
		for (int i = 0; i < numberOfIncreaseVertices; i++)
		{
			final DegreeIncrease e1 = finalDegreeIncreases.get(i);
			if ( e1.degreeIncrease > 0 )
			{
				for (int j = i + 1; j < numberOfIncreaseVertices; j++)
				{
					final DegreeIncrease e2 = finalDegreeIncreases.get(j);
					if ( e2.degreeIncrease > 0 && this.possibleToInsertEdge(newEdges, e1.vertex, e2.vertex) )
					{
						e1.degreeIncrease--;
						e2.degreeIncrease--;
						newEdges.add(new IntTupel(e1.vertex, e2.vertex));
						if ( e1.degreeIncrease == 0 )
						{
							break;
						}
					}
				}
				if ( e1.degreeIncrease > 0 )
					reorg++;

				while (e1.degreeIncrease > 0 && reorg < maxReorg && this.reorganizeSolution(e1, finalDegreeIncreases, newEdges))
				{

				}
				if ( e1.degreeIncrease > 0 )
					throw new CannotRealize();
			}
		}

		return newEdges;

		// // final HashSet<Integer> assignedVertices = new
		// // HashSet<>(coll.finalDegreeIncreases.size());
		// // for (final DegreeIncrease deg : coll.finalDegreeIncreases)
		// // {
		// // assignedVertices.add(deg.vertex);
		// // }
		//
		// final Vector<IntStore<Store<DegreeIncrease, Vector<DegreeIncrease>>>>
		// degreeIncreases =
		// new Vector<>();
		// for (final DegreeIncrease deg : coll.finalDegreeIncreases)
		// {
		// IntStore<Store<DegreeIncrease, Vector<DegreeIncrease>>> degSt = new
		// IntStore<>();
		// Vector<DegreeIncrease> s =
		// this.possibleNewNeighbors(coll.finalDegreeIncreases,
		// deg.vertex);
		// Collections.shuffle(s);
		// // GraphRealizer.shuffle(s);
		//
		// Store<DegreeIncrease, Vector<DegreeIncrease>> t = new
		// Store<DegreeIncrease,
		// Vector<DegreeIncrease>>();
		// t.setS1(deg);
		// t.setS2(s);
		// degSt.setS2(t);
		// degSt.setS1(coll.finalDegreeIncreases.size() - s.size());
		// degreeIncreases.add(degSt);
		//
		// if ( deg.degreeIncrease > s.size() )
		// throw new CannotRealize(null, -1, -1);
		// }
		// Collections.sort(degreeIncreases, Collections.reverseOrder());
		//
		// final Set<IntTupel> newEdges = new HashSet<>();
		//
		// for (int i = 0; i < degreeIncreases.size(); i++)
		// {
		// final DegreeIncrease e1 = degreeIncreases.get(i).getS2().getS1();
		// if ( e1.degreeIncrease > 0 )
		// {
		// final Vector<DegreeIncrease> possibleNeighborsForE1 =
		// degreeIncreases.get(i).getS2().getS2();
		// if ( possibleNeighborsForE1.size() < e1.degreeIncrease )
		// throw new CannotRealize();
		// for (DegreeIncrease possibleNeigh : possibleNeighborsForE1)
		// {
		// final IntTupel tup = new IntTupel(possibleNeigh.vertex, e1.vertex);
		// if ( possibleNeigh.degreeIncrease > 0 && !newEdges.contains(tup) )
		// {
		// newEdges.add(tup);
		// possibleNeigh.degreeIncrease--;
		// e1.degreeIncrease--;
		// if ( e1.degreeIncrease == 0 )
		// break;
		// }
		// }
		//
		// while (e1.degreeIncrease > 0 &&
		// this.reorganizeSolution(getIncreaseIter(degreeIncreases),
		// degreeIncreases.get(i).getS2(), newEdges))
		// {
		// }
		// if ( e1.degreeIncrease > 0 )
		// throw new CannotRealize(this, e1.vertex, e1.originalDegree);
		// }
		// }
		// return newEdges;

	}

	private boolean reorganizeSolution(final DegreeIncrease e1, final Vector<DegreeIncrease> degreeIncreases, final Set<IntTupel> newEdges)
			throws CannotRealize
	{
		if ( e1.degreeIncrease >= 2 )
		{
			for (final IntTupel tup : newEdges)
			{
				if ( this.possibleToInsertEdge(newEdges, e1.vertex, tup.getFirst())
						&& this.possibleToInsertEdge(newEdges, e1.vertex, tup.getSecond()) )
				{
					newEdges.remove(tup);
					newEdges.add(new IntTupel(e1.vertex, tup.getFirst()));
					newEdges.add(new IntTupel(e1.vertex, tup.getSecond()));
					e1.degreeIncrease -= 2;
					return true;
				}
			}
		}

		for (final DegreeIncrease deg2 : degreeIncreases)
		{

			if ( deg2 != e1 && deg2.degreeIncrease > 0 )
			{
				for (final IntTupel tup : newEdges)
				{
					if ( this.possibleToInsertEdge(newEdges, e1.vertex, tup.getFirst())
							&& this.possibleToInsertEdge(newEdges, deg2.vertex, tup.getSecond()) )
					{
						newEdges.remove(tup);
						newEdges.add(new IntTupel(e1.vertex, tup.getFirst()));
						newEdges.add(new IntTupel(deg2.vertex, tup.getSecond()));
						e1.degreeIncrease--;
						deg2.degreeIncrease--;
						// System.err.println("fix2");
						return true;
					}
					else if ( this.possibleToInsertEdge(newEdges, e1.vertex, tup.getSecond())
							&& this.possibleToInsertEdge(newEdges, deg2.vertex, tup.getFirst()) )
					{
						newEdges.remove(tup);
						newEdges.add(new IntTupel(e1.vertex, tup.getSecond()));
						newEdges.add(new IntTupel(deg2.vertex, tup.getFirst()));
						e1.degreeIncrease--;
						deg2.degreeIncrease--;
						// System.err.println("fix3");
						return true;
					}
				}
			}
		}

		return false;
	}

}
