package seppImpl;

import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.Set;

import seppImpl.oldDynProgramming.DegreeAnonymizer;
import benchmarks.MainTest;
import dataStructures.Store;
import de.vibee.kanon.data.AdjacentGraph;
import de.vibee.kanon.tools.FastErdosGallai;

public class SolutionDegreeIterator2
{
	private final BitSet[][] table;

	private final CompactDegreeSequence preProcessedSequence;

	private final int k;

	private static final int noOfStoredSolutions = 500;

	private int reachedSolutionsLimit = 0;

	private int alreadyConsiderCosts;

	private Store<Integer, Set<CompactDegreeSequence>> upperBound;

	private long numberOfRecursivelyConsideredInstances = 0;

	private final boolean findRealizableSequence;

	private final AdjacentGraph graph;

	private final CompactDegreeSequence originalSequence;

	private final boolean fixDegreeOneVerticesInErdösGallai;

	private long noOfAEGHopeDestroyed = 0;

	private long noOfAEGnegatives = 0;

	private long noOfAEGHopeTests = 0;

	private boolean lastStepSuccessful = false;

	// private final ErdosGallaiInterface erdosGallai;
	private final FastErdosGallai erdosGallai;

	private int constantHowOftenHopeForIsApplied = Integer.MAX_VALUE;

	private final int preProcessCosts;

	public SolutionDegreeIterator2(final AdjacentGraph graph, final BitSet[][] table2, final CompactDegreeSequence orginalSequence,
			final CompactDegreeSequence preProcessedSequence, final int k, final boolean findRealizableSequence, final boolean fixDegreeOn)
	{
		this.graph = graph;
		this.table = table2;
		this.originalSequence = orginalSequence;
		this.preProcessedSequence = preProcessedSequence;
		this.preProcessCosts = this.preProcessedSequence.difference(orginalSequence).getTotalDegreeSum();
		this.k = k;
		this.alreadyConsiderCosts = -1;
		this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(Integer.MAX_VALUE, null);
		this.reachedSolutionsLimit = 0;
		this.findRealizableSequence = findRealizableSequence;
		this.fixDegreeOneVerticesInErdösGallai = fixDegreeOn;
		this.erdosGallai = new FastErdosGallai(graph, orginalSequence);
	}

	private int[] constructSequenceFromTriples(final Triple[] triples)
	{
		final int[] newsequence = new int[this.preProcessedSequence.getMaxValue() + 1];
		Arrays.fill(newsequence, 0);

		int c = 0;
		for (int i = 0; i <= this.preProcessedSequence.getMaxValue(); i++)
		{
			if ( triples[i] != null )
			{
				if ( triples[i].getT() >= this.preProcessedSequence.get(i) )
				{
					c += this.preProcessedSequence.get(i);
				}
				else
				{
					newsequence[i] = this.preProcessedSequence.get(i) - triples[i].getT() + c;
					c = triples[i].getT();
				}
			}
			else
			{
				if ( c > 0 )
					throw new IllegalArgumentException("illegal triple");
				newsequence[i] = this.preProcessedSequence.get(i);
			}
		}
		return newsequence;
	}

	/**
	 * Given a list of triples (a "chain" in the dynamic programming table pointing to a certain raw
	 * solution), computes all solution degree sequences.
	 * 
	 * @param triples
	 */
	private void examineSequence(final Triple[] triples)
	{
		this.numberOfRecursivelyConsideredInstances++;

		final CompactDegreeSequence newCompactSequence = new CompactDegreeSequence(this.constructSequenceFromTriples(triples));
		final CompactDegreeSequence differenceSeq = newCompactSequence.difference(this.originalSequence);
		final int degreeSum = differenceSeq.getTotalDegreeSum();
		int minAddCost;

		if ( MainTest.logLevel > 5 )
		{
			System.out.println("============Begin CostWaster==============");
			System.out.println("OrigSeq: " + this.preProcessedSequence);
			System.out.println(Arrays.toString(triples));
			System.out.println("anonSeq:\t" + newCompactSequence);
			System.out.println("cost so far:\t" + degreeSum);
			System.out.println("upper bound:\t" + this.upperBound);
		}

		if ( !this.findRealizableSequence || differenceSeq.isRealizable()
				&& (this.graph == null || this.erdosGallai.computeFeasibility(newCompactSequence, this.fixDegreeOneVerticesInErdösGallai) <= 0) )
		{
			if ( degreeSum < this.upperBound.getS1() )
			{
				final HashSet<CompactDegreeSequence> s2 = new HashSet<CompactDegreeSequence>();
				this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(degreeSum, s2);
				this.reachedSolutionsLimit = 0;
				s2.add(newCompactSequence);
			}
			else if ( degreeSum == this.upperBound.getS1() && this.noOfStoredSolutions < this.upperBound.getS2().size() )
			{
				this.upperBound.getS2().add(newCompactSequence);
				if ( this.upperBound.getS2().size() >= this.noOfStoredSolutions )
				{ // limit reached; don't store further solutions
					this.reachedSolutionsLimit = 1;
				}
			}

		}
		else if ( degreeSum + (minAddCost = differenceSeq.noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös()) <= this.upperBound.getS1()
				- this.reachedSolutionsLimit )
		{
			for (int addCost = minAddCost; addCost + degreeSum <= this.upperBound.getS1() - this.reachedSolutionsLimit; addCost++)
			{
				// System.out.println("addcost:"+addCost);
				final CostWaster cw = new CostWaster(this.originalSequence, newCompactSequence, addCost, this.k);
				Set<CompactDegreeSequence> t;
				if ( (t = cw.getAllRealizableSequences()) != null && t.size() > 0 )
				{
					boolean found = false;
					for (final CompactDegreeSequence sol : t)
					{
						// System.out.println(differenceSeq.getTotalDegreeSum()
						// + addCost);
						long erdosGallaiAdvanced = 1;
						if ( this.graph == null
								|| (erdosGallaiAdvanced = this.erdosGallai.computeFeasibility(sol, this.fixDegreeOneVerticesInErdösGallai)) == 0 )
						{
							found = true;
							if ( this.upperBound.getS1() > differenceSeq.getTotalDegreeSum() + addCost )
							{
								final HashSet<CompactDegreeSequence> s = new HashSet<CompactDegreeSequence>();
								s.add(sol);
								this.upperBound = new Store<Integer, Set<CompactDegreeSequence>>(differenceSeq.getTotalDegreeSum() + addCost, s);
								this.reachedSolutionsLimit = 0;
							}
							else if ( this.upperBound.getS2().size() <= noOfStoredSolutions )
							{
								this.upperBound.getS2().add(sol);
								if ( this.upperBound.getS2().size() >= noOfStoredSolutions )
								{
									this.reachedSolutionsLimit = 1;
								}
							}
						}
						else
						{
							this.noOfAEGnegatives++;
							addCost += erdosGallaiAdvanced - 1; // -1 due to
																// increase of
																// addCost in
																// next run
																// through the
																// loop
						}

					}
					if ( found )
					{
						break;
					}
				}
			}
		}
	}

	public Set<CompactDegreeSequence> getAllRealizableSequences(final int maximumCosts)
	{
		if ( MainTest.logLevel > 1 )
		{
			final DecimalFormat myFormatter = new DecimalFormat("###,###,###.###");
			System.out.println("maxCost:" + maximumCosts + "\talreadyKnownUpperBound:" + this.upperBound.getS1() + "\talreadyConsideredCosts:"
					+ this.alreadyConsiderCosts + "\tinstances considered: " + myFormatter.format(this.numberOfRecursivelyConsideredInstances)
					+ "\thope tests: " + myFormatter.format(this.noOfAEGHopeTests) + "\tdropped: " + myFormatter.format(this.noOfAEGHopeDestroyed)
					+ "\tAEG negatives: " + myFormatter.format(this.noOfAEGnegatives));
			System.out.println("AEG-Calls: " + myFormatter.format(this.erdosGallai.getNoOfEGCalls()) + "\tTime1: ("
					+ myFormatter.format(this.erdosGallai.getTime1() / 1000d) + "sec) \tTime2: ("
					+ myFormatter.format(this.erdosGallai.getTime2() / 1000d) + "sec) \tTime3: ("
					+ myFormatter.format(this.erdosGallai.getTime3() / 1000d) + "sec) \tHashMapSize: "
					+ myFormatter.format(this.erdosGallai.getHashMapSize()) + "\tlookups: " + myFormatter.format(this.erdosGallai.getNoOfLookUps())
					+ "\toverall: " + myFormatter.format(this.erdosGallai.getEdgesBetweenCalls()));
			System.out.println("SolDepComps: " + myFormatter.format(this.erdosGallai.getNoOfSolDependingComputations()) + "\tSolDepLookUps: "
					+ myFormatter.format(this.erdosGallai.getNoOfSolDependingLookUps()) + "\tSolDepAll: "
					+ myFormatter.format(this.erdosGallai.getSolDependingChecks()) + "\tSolDepMaxSize: "
					+ myFormatter.format(this.erdosGallai.getMaxSolDepHashMapSize()));
			System.out.println("");
		}
		if ( this.alreadyConsiderCosts >= maximumCosts && this.upperBound.getS1() - this.preProcessCosts <= maximumCosts )
			return this.upperBound.getS2();
		if ( this.alreadyConsiderCosts >= maximumCosts )
			throw new IllegalAccessError();

		while (this.alreadyConsiderCosts < maximumCosts)
		{
			this.alreadyConsiderCosts++;
			int cost;
			if ( (cost = this.table[this.preProcessedSequence.getMaxValue()][0].nextSetBit(this.alreadyConsiderCosts)) >= 0 )
			{
				if ( cost > maximumCosts )
				{
					this.alreadyConsiderCosts = maximumCosts;
					break;
				}
				else
				{
					this.alreadyConsiderCosts = cost;
					final Triple[] triples = new Triple[this.preProcessedSequence.getMaxValue() + 1];
					this.constantHowOftenHopeForIsApplied = (triples.length + 4) / 3;
					triples[this.preProcessedSequence.getMaxValue()] = new Triple(this.preProcessedSequence.getMaxValue(), 0,
							this.alreadyConsiderCosts);
					this.recursivelyBuildAndExamineAllSolutionSequences(this.preProcessedSequence.getMaxValue(), triples);
				}
			}
		}
		if ( this.upperBound.getS1()-this.preProcessCosts <= maximumCosts )
			return this.upperBound.getS2();
		return null;
	}

	public int getCurrentUpperbound()
	{
		return this.upperBound.getS1();
	}

	public long getNoOfAEGHopeDestroyed()
	{
		return this.noOfAEGHopeDestroyed;
	}

	public long getNoOfAEGHopeTests()
	{
		return this.noOfAEGHopeTests;
	}

	public long getNumberOfRecConsideredInstances()
	{
		return this.numberOfRecursivelyConsideredInstances;
	}

	private boolean hopeForDecreasingUpperBound(final Triple[] triples, final int depth)
	{
		if ( !this.findRealizableSequence || !this.lastStepSuccessful && depth % this.constantHowOftenHopeForIsApplied != 0 )
			return true;
		// if ( !this.findRealizableSequence || !this.lastStepSuccessful &&
		// Integer.numberOfLeadingZeros(depth) + Integer.numberOfTrailingZeros(depth) > 30 )
		// return true;

		final CompactDegreeSequence newCompactSequence = new CompactDegreeSequence(this.constructSequenceFromTriples(triples));
		final CompactDegreeSequence differenceSeq = newCompactSequence.difference(this.originalSequence);

		final int degreeSum = differenceSeq.getTotalDegreeSum();
		int t = degreeSum + differenceSeq.noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös();
		boolean hopeForImprovingUpperbound = t <= this.upperBound.getS1() - this.reachedSolutionsLimit;

		this.noOfAEGHopeDestroyed += !hopeForImprovingUpperbound ? 1 : 0;
		this.noOfAEGHopeTests++;

		this.lastStepSuccessful = !hopeForImprovingUpperbound;

		if ( !hopeForImprovingUpperbound || triples[depth].getT() != 0 )
			return hopeForImprovingUpperbound;

		t = degreeSum + this.erdosGallai.computeFeasibility(newCompactSequence, this.fixDegreeOneVerticesInErdösGallai);
		hopeForImprovingUpperbound = t <= this.upperBound.getS1() - this.reachedSolutionsLimit;
		this.noOfAEGHopeDestroyed += !hopeForImprovingUpperbound ? 1 : 0;

		this.lastStepSuccessful = !hopeForImprovingUpperbound;

		return hopeForImprovingUpperbound;

	}

	private void recursivelyBuildAndExamineAllSolutionSequences(final int i, final Triple[] triples)
	{
		// for (int p = 0; p <= i; p++)
		// System.out.print(" ");
		// System.out.println(i);
		if ( i <= 0 )
		{
			this.examineSequence(triples);
		}
		else if ( this.hopeForDecreasingUpperBound(triples, i) )
		{
			final int t = triples[i].getT();
			final int c = triples[i].getC();

			if ( this.preProcessedSequence.get(i) > t )
			{
				if ( this.preProcessedSequence.get(i) - t >= this.k )
				{
					if ( this.table[i - 1][0].get(c) )
					{
						triples[i - 1] = new Triple(i - 1, 0, c);
						this.recursivelyBuildAndExamineAllSolutionSequences(i - 1, triples);
						triples[i - 1] = null;
					}

					int j = i - 1;
					int counter = 0;
					int oldCounter = 0;
					while (j >= 0 && (counter += this.preProcessedSequence.get(j)) < this.k
							&& DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, counter, i) <= c)
					{
						final int remCost = c - DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, counter, i);
						triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
						if ( counter > oldCounter
								&& DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j, this.preProcessedSequence.get(j),
										remCost) )
						{
							if ( j > 0 )
							{
								triples[j - 1] = new Triple(j - 1, 0, remCost);
							}
							this.recursivelyBuildAndExamineAllSolutionSequences(j - 1, triples);
						}
						j--;
						oldCounter = counter;
					}
					Arrays.fill(triples, 0, i, null);
				}
				// 0 < sequence.get(i) - t < k
				else if ( DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, this.k - (this.preProcessedSequence.get(i) - t), i) <= c )
				{
					int j = i - 1;
					final int iniJ = j;
					int counter = 0;
					final int noElementsToFillUpBag_i = this.k - (this.preProcessedSequence.get(i) - t);
					while (j >= 0 && counter + this.preProcessedSequence.get(j) < noElementsToFillUpBag_i)
					{
						counter += this.preProcessedSequence.get(j);
						triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
						j--;
					}
					if ( j >= 0 )
					{
						final int neededElementsFromBag_j = noElementsToFillUpBag_i - counter;
						if ( DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j, neededElementsFromBag_j, c
								- DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, noElementsToFillUpBag_i, i)) )
						{
							triples[j] = new Triple(j, neededElementsFromBag_j, c
									- DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, this.k - (this.preProcessedSequence.get(i) - t), i));
							this.recursivelyBuildAndExamineAllSolutionSequences(j, triples);
							triples[j] = null;

						}

						int u = this.preProcessedSequence.get(j) - neededElementsFromBag_j;
						int oldU = 0;
						while (u < this.k && DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i) <= c)
						{
							triples[j] = new Triple(j, this.preProcessedSequence.get(j), -1);
							if ( u > oldU
									&& DegreeAnonymizer.getTableEntry(this.preProcessedSequence, this.table, this.k, j,
											this.preProcessedSequence.get(j),
											c - DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i)) )
							{
								if ( j > 0 )
								{
									triples[j - 1] = new Triple(j - 1, 0, c
											- DegreeAnonymizer2.cost(this.preProcessedSequence, i - 1, u + neededElementsFromBag_j + counter, i));
								}
								this.recursivelyBuildAndExamineAllSolutionSequences(j - 1, triples);
							}

							j--;
							if ( j >= 0 )
							{
								oldU = u;
								u += this.preProcessedSequence.get(j);
							}
							else
							{
								break;
							}
						}
						Arrays.fill(triples, 0, iniJ, null);
					}
				}
			}
			// sequence.get(i) - t <= 0
			else
			{
				triples[i - 1] = new Triple(i - 1, t - this.preProcessedSequence.get(i), c);
				this.recursivelyBuildAndExamineAllSolutionSequences(i - 1, triples);
				triples[i - 1] = null;
			}

		}
	}
}
