package seppImpl.unitTests;

import java.util.Arrays;
import java.util.Random;

import junit.framework.Assert;

import org.junit.Test;

import seppImpl.CompactDegreeSequence;
import seppImpl.DegreeAnonymizer2;
import seppImpl.oldDynProgramming.DegreeAnonymizer;
import agape.generators.RandGenerator;
import agape.tutorials.UndirectedGraphFactoryForStringInteger;
import edu.uci.ics.jung.graph.Graph;

public class TestDegreeSeq
{

	private int[] dis(final int[] arr, int distortion)
	{
		while (distortion > 0)
		{
			int i;
			do
			{
				i = (int) Math.round(Math.random() * (arr.length - 1));
			} while (arr[i] == 0 || i == 0);
			final int d = Math.min(arr[i], Math.max((int) Math.round(Math.random() * 0.5f * distortion), 1));
			distortion -= d;
			arr[i] -= d;
			arr[i - 1] += d;
		}
		return arr;
	}

	@Test
	public void testAllSmallSequences()
	{
		long i = 0;
		final long t = System.currentTimeMillis();
		for (int n = 3; n <= 11; n++)
		{
			System.out.println("n=" + n);
			for (int k = 2; k <= Math.max(3, n / 2); k++)
			{
				System.out.println("\tk:" + k);
				final SmallInstanceIterator iter = new SmallInstanceIterator(n, k);
				while (iter.hasNext())
				{

					final int[] next = iter.next();
					i++;
					if ( n < 7 || i % Math.pow(n, k) == 0 )
					{
						final CompactDegreeSequence initialDegreeSequence = new CompactDegreeSequence(next);
						// System.out.println(initialDegreeSequence);
						if ( initialDegreeSequence.getTotalEntrySum() >= k && initialDegreeSequence.isRealizable() )
						{
							final int degreeAnon = DegreeAnonymizer.degreeAnon(initialDegreeSequence, k);
							final int degreeAnon2 = new DegreeAnonymizer2(null, initialDegreeSequence, k).getSizeOfMinimumSolution();
							if ( degreeAnon != degreeAnon2 )
							{
								System.err.println("AHHH: k:" + k + "\tSeq: " + initialDegreeSequence);
								System.err.println("degreeAnon1:" + degreeAnon + "\tdegreeAnon2: " + degreeAnon2);
							}
							Assert.assertEquals(degreeAnon, degreeAnon2);
						}
					}
				}
			}
		}
		System.out.println("all instance test in " + (System.currentTimeMillis() - t) / 1000d + "s");
	}

	// @Test
	// public void testCost()
	// {
	//
	// final int n = 9;
	// final int k = 2;
	// SmallInstanceIterator iter = new SmallInstanceIterator(n, k);
	//
	// long time1 = 0;
	// long time2 = 0;
	// while (iter.hasNext())
	// {
	// int[] arr = iter.next();
	// final CompactDegreeSequence degreeSequence = new
	// CompactDegreeSequence(arr);
	// int ijkafdj = degreeSequence.getDegreeSumsInBags(0, 2);
	// int in = degreeSequence.getNumberOfVerticesInBags(0, 2);
	//
	// final int i = 1 + (int) Math.round(Math.random() * (n - 2));
	// final int t = 1 + (int) Math.round(Math.random() * n * 2f * k);
	// final int j = Math.max(i, 1 + (int) Math.round(Math.random() * (n - 2)));
	//
	// // System.out.println(Arrays.toString(arr));
	// // System.out.println(i + "\t" + t + "\t" + j);
	//
	// long timeStamp = System.currentTimeMillis();
	// final int cost = DegreeAnonymizer2.cost(degreeSequence, i, t, j);
	// time1 += (System.currentTimeMillis() - timeStamp);
	// timeStamp = System.currentTimeMillis();
	//
	// final int cost2 = DegreeAnonymizer2.newCost(degreeSequence, i, t, j);
	// time2 += (System.currentTimeMillis() - timeStamp);
	//
	// if ( cost != cost2 )
	// {
	// System.out.println(degreeSequence);
	// System.out.println(i + "\t" + t + "\t" + j);
	// System.out.println(cost + "\t" + cost2);
	// }
	// Assert.assertEquals(cost, cost2);
	// }
	// System.out.println("1:" + (time1 / 1000d));
	// System.out.println("2:" + (time2 / 1000d));
	//
	// }

	@Test
	public void testIsRealizable()
	{

		for (int k = 2; k <= 3; k++)
		{
			final SmallInstanceIterator iter = new SmallInstanceIterator(8, k);
			while (iter.hasNext())
			{
				final int[] arr = iter.next();

				final CompactDegreeSequence c = new CompactDegreeSequence(arr);
				final boolean realizable = c.OldisRealizable();
				final boolean newRealizable = c.isRealizable();

				if ( realizable != newRealizable )
				{
					System.out.println(c);
					c.isRealizable();
					c.OldisRealizable();
				}
				Assert.assertEquals(newRealizable, realizable);
				Assert.assertEquals(c.OldnoOfAddDegreeOneVerticesForBeingRealizable(), c.noOfAddDegreeOneVerticesForBeingRealizableWRSimpleErdös());
			}
		}
	}

	@Test
	public void testOnBarabasiAlbertGraphs() throws Exception
	{
		for (int t = 0; t < 1000; t++)
		{
			final UndirectedGraphFactoryForStringInteger factory = new UndirectedGraphFactoryForStringInteger();

			final int n = 100;
			final Graph<String, Integer> g5 = RandGenerator.generateBarabasiAlbertGraph(factory, factory.vertexFactory, factory.edgeFactory, n, 20,
					10);
			int arr[] = new int[2];
			for (final String v : g5.getVertices())
			{
				final int i1 = g5.getNeighborCount(v);
				if ( i1 > arr.length - 1 )
				{
					arr = Arrays.copyOf(arr, i1 + 1);
				}
				arr[i1]++;
			}

			final CompactDegreeSequence degreeSequence = new CompactDegreeSequence(arr);
			for (int k = 2; k <= 4; k++)
			{

				final int degreeAnon2 = new DegreeAnonymizer2(null, degreeSequence, k).getSizeOfMinimumSolution();
				final int degreeAnon = DegreeAnonymizer.degreeAnon(degreeSequence, k);
				if ( degreeAnon != degreeAnon2 )
				{
					System.err.println("k:" + k + "\t" + Arrays.toString(arr));
					System.err.println(degreeAnon + "\t" + degreeAnon2);
				}
				Assert.assertEquals(degreeAnon, degreeAnon2);
			}
		}
	}

	// @Test
	// public void testDynamicProgramming() throws Exception
	// {
	// UndirectedGraphFactoryForStringInteger factory = new
	// UndirectedGraphFactoryForStringInteger();
	//
	// for (int i = 0; i < 1000000; i++)
	// {
	// final int k = 2 + (int) Math.round((5f * Math.random()));
	// final int n = 2 + (int) Math.round(20f * Math.random());
	//
	// Graph<String, Integer> g5 =
	// RandGenerator.generateBarabasiAlbertGraph(factory,
	// factory.vertexFactory, factory.edgeFactory, 5, 5, 7);
	// int arr2[] = new int[2];
	// for (String v : g5.getVertices())
	// {
	// int i1 = g5.getNeighborCount(v);
	// if ( i1 > (arr2.length - 1) )
	// arr2 = Arrays.copyOf(arr2, i1 + 1);
	// arr2[i1]++;
	// }
	// System.out.println(Arrays.toString(arr2));
	// int[] arr = SeppsMain.getRandomkAnonDegreeSequence(n, k);
	// if ( SeppsMain.sum(arr) >= k && new
	// CompactDegreeSequence(arr).isRealizable() )
	// {
	// System.out.println(new CompactDegreeSequence(arr));
	// int distortion = (int) (Math.random() * SeppsMain.sum(arr));
	// arr = dis(arr, distortion);
	//
	// final CompactDegreeSequence seq = new CompactDegreeSequence(arr);
	// System.out.println(seq + "\tk:" + k + "\td:" + distortion);
	// final int sol = new DegreeAnonymizer2(seq, k).degreeAnon2();
	// final boolean b = sol <= distortion;
	// if ( !b )
	// {
	// System.out.println("\tsol:" + sol);
	// }
	// Assert.assertTrue(b);
	// }
	// }
	// }

	@Test
	public void testRandomInstances()
	{
		for (int i = 0; i < 5000; i++)
		{
			final int[] input = new int[50];
			int sum = 0;
			for (int j = 0; j < input.length; j++)
			{
				final Random generator = new Random();
				input[j] = generator.nextInt(40);
				sum += input[j];
			}
			if ( sum % 2 != 0 )
			{
				input[1]++;
				sum++;
			}

			for (int k = 2; k < 15; k += 3)
			{
				final CompactDegreeSequence initialDegreeSequence = new CompactDegreeSequence(input);
				if ( sum >= k && initialDegreeSequence.isRealizable() )
				{
					// if ( Math.random() < 0.0001f )
					// System.out.println("ins:\t\t" + initialDegreeSequence);
					// System.out.println("k:\t\t" + k);
					final int degreeAnon = DegreeAnonymizer.degreeAnon(initialDegreeSequence, k);
					// System.out.println("Nr 2");
					final int degreeAnon2 = new DegreeAnonymizer2(null, initialDegreeSequence, k).getSizeOfMinimumSolution();
					if ( degreeAnon != degreeAnon2 )
					{
						System.out.println("AHHH: k:" + k + "\tSeq: " + initialDegreeSequence);
						System.out.println("degreeAnon1:" + degreeAnon + "\tdegreeAnon2: " + degreeAnon2);
					}
					Assert.assertEquals(degreeAnon, degreeAnon2);
				}
			}
		}
	}

}
